%% als je gewoon de kleinste wil:
%macro(w(Surf,Lex,Cat),
%      word(Surf) x [word(Lex), ':', Cat]
%     ).

%% minimal dictionary: simply use t_minimize around dict.

%% in order to implement `guessing' automata, it appear to make sense to
%% construct an `aligned' cross-product for the surface and lexical word
%% form.

rx(dict_align_tr(File),Fa) :-
    fsa_util:time(fsa_dict:get_file(File,Lines0)),
    fsa_util:time(fsa_regex_aux:lines_to_aligned_mapping(Lines0,Lines)),
    fsa_util:time(fsa_dict:dict(Lines,Fa0)),
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,r(M),t(M,M)).

rx(dict_tr(File),Fa) :-
    fsa_util:time(fsa_dict:get_file(File,Lines0)),
    fsa_util:time(fsa_regex_aux:lines_to_mapping(Lines0,Lines)),
    fsa_util:time(fsa_dict:dict(Lines,Fa0)),
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,r(M),t(M,M)).

lines_to_mapping([],[]).
lines_to_mapping([H|T],[NH|NT]) :-
    line_to_mapping(H,NH),!,
    lines_to_mapping(T,NT).

line_to_mapping(Line,Pairs) :-
    lists:append(Surface,[' '|Rest],Line),
    lists:append(Lexical,[' '|Cat0],Rest),
    fsa_util:concat_all(Cat0,Category),    
    simple_align(Surface,Lexical,Pairs,[[]/':',[]/Category]).

simple_align([],Sec,T0,T) :-
    complete_sec(Sec,T0,T).

simple_align([Ha|Ta],Sec,T0,T) :-
    simple_align_ne(Sec,Ha,Ta,T0,T).

simple_align_ne([],Ha,Ta,[Ha/[]|T1],T) :-
    simple_align(Ta,[],T1,T).
simple_align_ne([B|Bs],A,As,[A/B|T0],T) :-
    simple_align(As,Bs,T0,T).

complete_sec([],T,T).
complete_sec([H|Hs],[[]/H|T0],T) :-
    complete_sec(Hs,T0,T).



lines_to_aligned_mapping([],[]).
lines_to_aligned_mapping([H|T],[NH|NT]) :-
    line_to_aligned_mapping(H,NH),!,
    lines_to_aligned_mapping(T,NT).

line_to_aligned_mapping(Line,Pairs) :-
    lists:append(Surface,[' '|Rest],Line),
    lists:append(Lexical,[' '|Cat0],Rest),
    fsa_util:concat_all(Cat0,Category),    
    align(Surface,Lexical,Pairs,[[]/':',[]/Category]).

align(L,R,Pairs0,Pairs) :-
    fsa_hash:fsa_hash_new(Done),
    approx_trans([p(0,0,[],L,R)],Items,Done),
    sols(Items,Sols0,Done),
    approx_closure(Sols0,Items,Sols,Done),
    lists:member(His,Sols),
    reverse(His,Pairs,Pairs0).

sols([],[],_).
sols([p(PosL,PosR,Rec,L,R)|Tail],Sols,Done):-
    sols0(L,R,PosL,PosR,Rec,Sols0,Sols,Done),
    sols(Tail,Sols0,Done).

sols0(L,R,PosL,PosR,Rec,Sols0,Sols,Done):-
    (	L == [],
	R == []
    ->	Sols=[Rec|Sols0]
    ;   Sols=Sols0
    ),
    fsa_hash:fsa_hash_access(PosL/PosR,[],Done).

%% approx_closure(+Sols0,+Items,-Sols,?Done)
%% Sols:  list of solutions at current penalty level (if it's
%%        non-empty, we're finished)
%% Items: list of triples p(PosL,PosR,His) indicating the possibilities
%%        for the current penalty level,
%% Sols:  list of solutions
%% Done:  hash table which maintains all PosL/PosR pairs which have
%%        been encountered at previous penalty levels (so these must
%%        not be generated at this or later levels)
approx_closure([H|T],_,[H|T],_).
approx_closure([],Items0,Sols,Done):-
    % extend each item in all possible ways, by applying penalty
    % transformation (ins,del,subs,transp)
    add_one(Items0,Items1),
    % compute transitive closure of all items (i.e. taking transitions)
    approx_trans(Items1,Items,Done),
    % extract list of items which are solutions (State final, Pos final)
    sols(Items,Sols0,Done),
    % recurse
    approx_closure(Sols0,Items,Sols,Done).

add_one([],[]).
add_one([p(Pa,Pb,His,A,B)|T],Items):-
    findall(Item,pen_step(Pa,Pb,His,Item,A,B),Items,Items0),
    add_one(T,Items0).

pen_step(Pa0,Pb,His,p(Pa,Pb,[  Sym/[]  |His],A,B),[Sym |A],      B) :-
    Pa is +(Pa0,1).

pen_step(Pa,Pb0,His,p(Pa,Pb,[   []/Sym |His],A,B),     A,  [Sym |B]) :-
    Pb is +(Pb0,1).

pen_step(Pa0,Pb0,His,p(Pa,Pb,[SymA/SymB|His],A,B),[SymA|A],[SymB|B]) :-
    SymA \== SymB,
    Pa is +(Pa0,1),
    Pb is +(Pb0,1).

approx_trans([],[],_).
approx_trans([p(Pa,Pb,Rec,A,B)|T],I,Done):-
    transition_seq(Pa,Pb,Rec,A,B,I,I0,Done),
    approx_trans(T,I0,Done).

%% extend each item in a greedy way, i.e. as far as possible without
%% penalty; don't continue as soon as you enter a state which is in
%% Done
transition_seq(Pa0,Pb0,Rec0,A,B,Items0,Items,Done):-
    fsa_hash:fsa_hash_access(Pa0/Pb0,Var,Done),
    (	nonvar(Var)  % this pair has already been
    ->  Items0=Items % found with better distance
    ;	A = [Sym|A1],
	B = [Sym|B1]
    ->	Pa1 is +(Pa0,1),
	Pb1 is +(Pb0,1),
	transition_seq(Pa1,Pb1,[Sym/Sym|Rec0],A1,B1,Items0,Items,Done)
    ;	Items0=[p(Pa0,Pb0,Rec0,A,B)|Items]
    ).

reverse([],R,R).
reverse([H|T],SoFar,R):-
    reverse(T,[H|SoFar],R).

