%% (c) 1998-2000 Gertjan van Noord, Dale Gerdemann; available under GPL
%%
%% this implements the syllabification examples from
%% Lauri Karttunen, 'The Proper Treatment of Optimality in
%% Computational Phonology'. in: FSMNLP 98. Also available
%% through CMP-LG. It's somewhat different in places to make it
%% comparable with our alternative treatment (cf matching.pl);
%% it's more general too in that it allows counting of other
%% constraints in addition to the parse constraint as in Karttunen's
%% paper.

:- op(400,yfx,(^>)).
:- op(403,yfx,(lc)).

:- multifile macro/2.
:- discontiguous macro/2.

:- ensure_loaded(utils).

%% As in Karttunen, the (A ^> N) operator means more than N iterations
%% of A.
macro(A ^> N,[AN,A+]) :-
   make_list(N,A,AN).

%% make_list(+Length,+ListMember,?List)
%% Construct a list full of ListMember
make_list(N,A,List) :-
    (	N=:=0
    ->	List=[]
    ;	List=[A|List0],
	N1 is N-1,
	make_list(N1,A,List0)
    ).

%%macro(priority_union(Q,S), {Q, ~domain(Q) o S}).
%%macro(R lc C, priority_union(R o C,R)).

macro(R0 lc C0, pragma([R-R0,C-C0], {R o C, ~domain(R o C) o R} )).


macro(Input oo Precision :: Constraint, oo(Input,Precision,Constraint)).
macro(Input oo              Constraint, oo(Input,0        ,Constraint)).

%% apply Constraint to Input with Precision
macro(oo(Input,Precision,Constraint),  Body o  { @ : [], ? - @ }* ) :-
    get_constraint(Precision,
		   Input o mark_ot_constraint(Constraint,@), Body).

get_constraint(Precision,B0,B) :-
    (	Precision =< 0
    ->	B = (B0 lc ~ ($ @))
    ;   B = (B1 lc ~ ($ @ ^> Precision)),
	Precision1 is Precision-1,
	get_constraint(Precision1,B0,B1)
    ).

