%% Levelt and van de Vijver, Syllable types in cross-linguistic and
%% development grammar. ROA. (3) Factorial Typology. 

:- multifile macro/2.

:- ensure_loaded(utils).

macro(onset,     [o_br, cons+  , r_br]).
macro(nucleus,   [n_br, vowel , r_br]).
macro(coda,      [d_br, cons+  , r_br]).
macro(unparsed,  [x_br, letter , r_br]).

macro(parse,
      { [[]:o_br,cons+, []:r_br],
        [[]:d_br,cons+, []:r_br],
	[[]:n_br,vowel, []:r_br],
	[[]:x_br,letter,[]:r_br]
      }* ).

macro(syllable_structure,
             ignore([onset^,nucleus,coda^],unparsed)*).

macro(gen, input* o parse o syllable_structure ).

%% Faith
macro(mark_ot_constraint(faith,Mark),
      %% dots({x_br,[n_br, r_br],[o_br, r_br]},[],Mark) ).
      dots(x_br,[],Mark) ).

%% Syllables must not have codas
macro(mark_ot_constraint(no_coda,Mark),
      dots(d_br,        [],Mark) ).

macro(mark_ot_constraint(complex_ons,Mark),
      dots([o_br, letter, letter+, r_br],[],Mark)).

macro(mark_ot_constraint(complex_coda,Mark),
      dots([d_br, letter, letter+, r_br],[],Mark)).

%% Syllables must have onsets; i.e. every nucleus
%% must immediately be preceded by an onset
%% see below for a variant that is easier to understand; this one is faster
%% and doesn't require the replacement stuff to be loaded.
macro(mark_ot_constraint(have_ons,Mark),
      {?, [[] x Mark, n_br]}*  % optional marker before a nucleus
              o
      ~ $ [onset,Mark]         % but not if there is a preceding onset
              o                % and obligatory if there is no onset
           ~ [ ( ~[? *,onset] & ~[? *, Mark] ),  n_br,? *]
     ).

%% Gertjan suggests the following alternative:
%%:- ensure_loaded('../GerdemannVannoord99/replace').
%%macro(mark_ot_constraint(have_ons,Mark),
%%      replace([] x Mark,[],n_br)
%%                 o
%%      replace(Mark x [],onset,[])
%%     ).
%%

%% Dale suggests the following alternative:
%%:- ensure_loaded('../GerdemannVannoord99/replace').
%%macro(add_anchors,[[] x '$',? *, [] x '$']).
%%
%%macro(mark_ot_constraint(have_ons,Mark),
%%      add_anchors
%%          o
%%      replace([] x Mark,(['$',? *] - [? *,onset]),n_br)
%%          o
%%     inverse(add_anchors)).

%% conslist(Lang,ListOfConstraints)
%% for each `language' identified by Prince and Smolensky (page 94),
%% give the corresponding list of constraints (most prominent first)
conslist(hua,      [have_ons,no_coda,complex_ons,complex_coda,faith]).
conslist(cayuvava, [no_coda,complex_ons,complex_coda,faith,have_ons]).
conslist(thargari, [have_ons,complex_ons,complex_coda,faith,no_coda]).
conslist(arabela,  [have_ons,no_coda,complex_coda,faith,complex_ons]).
conslist(sedang,   [have_ons,complex_coda,faith,no_coda,complex_ons]).
conslist(mazateco, [no_coda,complex_coda,faith,have_ons,complex_ons]).
conslist(mokilese, [complex_coda,faith,have_ons,complex_ons,no_coda]).
conslist(klamath,  [have_ons,complex_ons,faith,no_coda,complex_coda]).
conslist(totonac,  [have_ons,faith,no_coda,complex_coda,complex_ons]).
conslist(spanish,  [complex_coda,faith,have_ons,no_coda,complex_ons]).
conslist(finnish,  [complex_ons,faith,have_ons,no_coda,complex_coda]).
conslist(dutch,    [faith,have_ons,no_coda,complex_ons,complex_coda]).

%%conslist(1,        [have_ons,no_coda,complex_ons,complex_coda,faith]).
%%conslist(2,        [no_coda,complex_ons,complex_coda,faith,have_ons]).
%%conslist(3,        [have_ons,complex_ons,complex_coda,faith,no_coda]).
%%conslist(4,        [have_ons,no_coda,complex_coda,faith,complex_ons]).
%%conslist(5,        [have_ons,complex_coda,faith,no_coda,complex_ons]).
%%conslist(6,        [no_coda,complex_coda,faith,have_ons,complex_ons]).
%%conslist(7,        [complex_coda,faith,have_ons,complex_ons,no_coda]).
%%conslist(8,        [have_ons,complex_ons,faith,no_coda,complex_coda]).
%%conslist(9,        [have_ons,faith,no_coda,complex_coda,complex_ons]).
%%conslist(10,       [complex_coda,faith,have_ons,no_coda,complex_ons]).
%%conslist(11,       [complex_ons,faith,have_ons,no_coda,complex_coda]).
%%conslist(12,       [faith,have_ons,no_coda,complex_ons,complex_coda]).

