:- module( bit_preds, []).

:- use_module(bitsets).

:- user:bb_put(bit_preds:current_no,0).

true(not_in(0)).

%% NB this is a funny predicate module in that sets of symbols are
%% represented with bitsets. However, the actual symbols are _not_
%% represented in the output automata but are added to the database
%% and hence are gone when FSA finishes. This is ok, if you output
%% the machine in an external format which first uses
%% 'evaluate_predicate' to get the symbols back in (C, C++, JAVA).

:- dynamic leaf_n/2.
:- dynamic n_leaf/2.

regex_notation_to_predicate(Atom,in(Bitset)) :-
    new_leaf_n(Atom,N),
    bitsets:list_to_bitset([N],Bitset).

new_leaf_n(Atom,N):-
    (	leaf_n(Atom,N)
    ->	true
    ;   user:bb_get(bit_preds:current_no,N),
	N1 is +(N,1),
	user:bb_put(bit_preds:current_no,N1),
	assertz(leaf_n(Atom,N)),
	assertz(n_leaf(N,Atom))
    ).

new_leaf_ns([],[]).
new_leaf_ns([L|T0],[N|T]) :-
    new_leaf_n(L,N),
    new_leaf_ns(T0,T).

ns_leaf([],[]).
ns_leaf([N|T0],[L|T]) :-
    n_leaf(N,L),
    ns_leaf(T0,T).

evaluate_predicate(in(Bitset),Symbol) :-
    bitsets:bitset_to_list(Bitset,List),
    lists:member(No,List),
    n_leaf(No,Symbol).

evaluate_predicate(not_in(Bitset),Symbol) :-
    bitsets:bitset_to_list(Bitset,List),
    dif_list(List,Symbol).

dif_list([],_).
dif_list([N|T],Symbol) :-
    n_leaf(N,H),
    dif(H,Symbol),
    dif_list(T,Symbol).

talks_about(in(Bitset),Symbol) :-
    bitsets:bitset_to_list(Bitset,List),
    lists:member(No,List),
    n_leaf(No,Symbol).

talks_about(not_in(Bitset),Symbol) :-
    talks_about(in(Bitset),Symbol).

disjunction(not_in(A),B,C):-
    disjunction_not_in(B,A,C).

disjunction(in(A),B,C):-
    disjunction_in(B,A,C).

disjunction_not_in(not_in(B),A,not_in(C)) :-
    bitsets:bit_intersection(A,B,C).
disjunction_not_in(in(B),A,not_in(C)):-
    bitsets:bit_subtract(A,B,C).

disjunction_in(in(B),A,in(C)):-
    bitsets:bit_union(A,B,C).
disjunction_in(not_in(A),B,not_in(C)):-
    bitsets:bit_subtract(A,B,C).

conjunction(not_in(A),B,C):-
    conjunction_not_in(B,A,C).

conjunction(in(A),B,C):-
    conjunction_in(B,A,C).

conjunction_not_in(not_in(B),A,not_in(C)):-
    bitsets:bit_union(A,B,C).
conjunction_not_in(in(B),A,in(C)):-
    bitsets:bit_subtract(B,A,C),
    C =\= 0.

conjunction_in(in(B),A,in(C)):-
    bitsets:bit_intersection(A,B,C),
    C =\= 0.
conjunction_in(not_in(A),B,in(C)):-
    bitsets:bit_subtract(B,A,C),
    C =\= 0.

negation(in(P),not_in(P)).
negation(not_in(L),in(L)) :-
    L =\= 0.

display_predicate(in(List0),Conj):-
    bitsets:bitset_to_list(List0,List1),
    ns_leaf(List1,List),
    fsa_visualization:abbreviate_label0(List,[H|T]),
    display_predicate_in(T,H,Conj).

display_predicate(not_in(X),Conj):-
    bitsets:bitset_to_list(X,List1),
    ns_leaf(List1,List2),
    fsa_visualization:abbreviate_label0(List2,List),
    display_predicate_not_in(List,Conj).

display_predicate_in([],H,H).
display_predicate_in([H|T],A,{Conj}):-
    list_to_conj([H|T],A,Conj).

display_predicate_not_in([],?).
display_predicate_not_in([H|T],'`'(Conj)):-
    display_predicate_in(T,H,Conj).

list_to_conj([],A,A).
list_to_conj([H|T],A,(A,B)):-
    list_to_conj(T,H,B).

prepare_complement_of_set(_fa,_).

complement_of_set(Preds,_,Comps):-
    true(True),
    (	conjoin_negated(Preds,True,Comp)
    ->	Comps=[Comp]
    ;   Comps=[]
    ).

conjoin_negated([],Pred,Pred).
conjoin_negated([P0|Preds],P1,Pred) :-
    negation(P0,NotP0),
    conjunction(P1,NotP0,P2),
    conjoin_negated(Preds,P2,Pred).

determinize_preds(U0,U):-
    fsa_determinizer:determinize_preds(U0,U,bit_preds).

t_determinize_preds(U0,U):-
    fsa_t_determinizer:t_determinize_preds(U0,U,bit_preds).

identity(Pred,'$@'(Pred)).

class_to_pred(List,[in(Pred)]):-
    new_leaf_ns(List,Ns),
    bitsets:list_to_bitset(Ns,Pred).

negated_class_to_pred(List,[not_in(Pred)]):-
    new_leaf_ns(List,Ns),
    bitsets:list_to_bitset(Ns,Pred).

cleanup([],[]).
cleanup([H|T],[Pred]) :-
    cleanup(T,H,Pred).

cleanup([],P,P).
cleanup([H|T],Pred0,Pred) :-
    disjunction(H,Pred0,Pred1),
    cleanup(T,Pred1,Pred).
