:- module(bitsets, [
	bitset_to_list/2,
	list_to_bitset/2,
	bit_add_element/3,
	bit_del_element/3,
	bit_disjoint/2,
	bit_intersect/2,
	bit_intersection/2,
	bit_intersection/3,
	bit_member/2,
	bit_memberchk/2,
	bit_seteq/2,
	bit_subtract/3,
	bit_symdiff/3,
	bit_union/2,
	bit_union/3
   ]).

%% only practicle for smallish sets
%% msb/1 below is not available in SWI Prolog

bit_add_element(M, BitNo, M1) :-
    integer(BitNo),
    BitNo >= 0,
    M1 is (1 << BitNo) \/ M.

bit_del_element(M, BitNo, M1) :-
    integer(BitNo),
    BitNo >= 0,
    M1 is \(1 << BitNo) /\ M.

bit_disjoint(M2, M1) :-
    M1 /\ M2 =:= 0.

bit_intersect(Set1, Set2) :-
    \+ bit_disjoint(Set1, Set2).

bit_intersection(M1, M2, M3) :-
    M3 is M1 /\ M2.

bit_intersection([BitSet|BitSets], Intersection) :-
    bit_intersection1(BitSets, BitSet, Intersection).

bit_intersection1([], Intersection, Intersection).
bit_intersection1([BitSet|BitSets], Intersection0, Intersection) :-
    bit_intersection(BitSet, Intersection0, Intersection1),
    bit_intersection1(BitSets, Intersection1, Intersection).

bit_seteq(BitSet, BitSet).

bit_subset(M1, M2) :-
    \(M2)/\M1 =:= 0.

bit_subtract(M1,M2, M3) :-
    M3 is \(M2) /\ M1.

bit_symdiff(M1,M2,M3) :-
    M3 is (\(M2) /\ M1) \/ (\(M1) /\ M2).

bit_union(M1,M2,M3):-
    M3 is M1 \/ M2.

bit_union(BitSets, Union) :-
    bit_union1(BitSets, 0, Union).

bit_union1([], Union, Union).
bit_union1([BitSet|BitSets], Union0, Union) :-
    bit_union(BitSet, Union0, Union1),
    bit_union1(BitSets, Union1, Union).

list_to_bitset(Numbers, BitSet) :-
    list_to_bitset(Numbers, 0, BitSet).

list_to_bitset([],B,B).
list_to_bitset([Number|Numbers],M0,M):-
    M1 is (1 << Number) \/ M0,
    list_to_bitset(Numbers,M1,M).


bitset_to_list(BitSet, Numbers) :-
    bitset_to_list(BitSet, 0, Numbers, []).

bitset_to_list(M, Shift,P0,P) :-
    (	M =:= 0
    ->	P0=P
    ;	Y is \(M-1) /\ M,
	M1 is M-Y,
	N is msb(Y),   % not in SWI Prolog
	Bit is N<<Shift,
	P0=[Bit|P1],
	bitset_to_list(M1, Shift,P1,P)
    ).

bit_memberchk(BitNo,M) :-
    (	M >> BitNo) /\ 1 =\= 0.

bit_member(BitNo, BitSet) :-
    (	var(BitNo)
    ->	bitset_to_list(BitSet,List),
	lists:member(BitNo,List)
    ;	integer(BitNo), BitNo >= 0,
	bit_memberchk(BitNo, BitSet)
    ).
