%% Dale:
%% With all this talk of booleans, I got inspired to define
%% IsFunctional. But, of course, I didn't acutally finish. I just managed
%% to do UNAMBIGUOUS, which I called remove_ambiguity. A student of mine
%% did this a while back and sent you the code. But since you didn't
%% approve of the code, I thought I would do it again myself from
%% scratch. I'm reasonably confident that the code here is okay. It could
%% possibly be optimized somewhat, but unlike my student's code, it does
%% do first argument indexing (as one would expect).

:- ensure_loaded('../Booleans/booleans').

:- multifile macro/2.
:- multifile rx/2.

% for stand-alone binaries these should not be re-loaded - they're in
% the system already anyway...
% :- use_module(library(ordsets)). 
% :- use_module(library(lists)).



%% The example from R&S p. 29 and p. 34
test(Q2,I2,F2,E2) :-
   Q = [0,1,2,3],
   I = 0,        % only one initial state, need extension to allow multiple
   F = [2],      % set of finals is sorted
   E = [trans(0,a/[b],1),
        trans(0,a/[b,c],3),
        trans(0,a/[b,c,d],2),
        trans(1,b/[c,d],2),
        trans(2,b/[],2),
        trans(3,b/[d],2)],
   remove_ambiguity(Q,I,F,E,Q2,I2,F2,E2).
   

remove_ambiguity(Q,I,F,E,Q2,0,F2,E2) :-
   do_loop([(I,[])-_N],[],0,Q,F,E,[],Q2,[],E2,F2).


%% do_loop(...)   
%% This all follows R&S very closely. The main difference is that I've
%% replaced the array C, with a queue of "Items". Each Item is of the
%% form (X1,S)-N, where N indicates the position of the Item in the
%% queue. In order to reduce the number of variables I have to keep
%% track of, I don't actually instantiate N until the item is taken
%% off the queue. At this point, N can be instantiated to the loop
%% variable SmallQ, i.e., the variable q in R&S.

do_loop([(X1,S)-SmallQ|Items0],Assoc0,SmallQ,Q,F,E,Q20,Q2,E20,E2,F2) :-
   Q21 = [SmallQ|Q20],
   Assoc1 = [(X1,S)-SmallQ|Assoc0],  % Assoc is set of already considered items
   writeit(SmallQ,X1,S),
   (   lists:memberchk(X1,F),
       \+ ordsets:ord_intersect(S,F)  % S and F do not intersect
   ->  F2 = [SmallQ|RestF2]
   ;   F2 = RestF2
   ),
   foreach1(E,E,X1,Assoc1,SmallQ,S,Q21,E20,E21,Items0,Items1), % 1st foreach in R&S
   SmallQ1 is +(SmallQ,1),
   do_loop(Items1,Assoc1,SmallQ1,Q,F,E,Q21,Q2,E21,E2,RestF2).
do_loop([],_,_,_,_,_,Q2,Q2,E2,E2,[]).

%% A little debug info to see what pair each output state corresponds to
writeit(SmallQ,X1,S) :- write(SmallQ),write(' = '),write((X1,S)),nl.

foreach1([],_E,_,_,_SQ,_S,_Q2,E2,E2,Items,Items).
foreach1([trans(X1,A/W,Y1)|T],E,X1,Assoc,SQ,S,Q2,E20,E2,Items0,Items) :-
   !,
   foreach2(E,X1,A,W,[],S_prime1),    % the second foreach loop in R&S
   foreach3(S,E,A,S_prime1,S_prime2), % the third foreach loop in R&S
   sort(S_prime2,S_prime),
   (  lists:member(Y1,S_prime)
   -> Items1 = Items0,
      E21 = E20
   ;  (  (lists:member(((Y1,S_prime)-SmallE),Assoc) ;     %% addSet
          lists:member(((Y1,S_prime)-SmallE),Items0))
      -> Items1 = Items0,
         E21 = [trans(SQ,A/W,SmallE)|E20]
      ;  lists:append(Items0,[(Y1,S_prime)-N],Items1),
				% It would be simpler to put the new Item
				% at the beginning, but I want the list of 
				% items to be a queue just like C in R&S
         E21 = [trans(SQ,A/W,N)|E20]
      )
   ),
   foreach1(T,E,X1,Assoc,SQ,S,Q2,E21,E2,Items1,Items).
foreach1([_|T],E,X1,Assoc,SQ,S,Q2,E20,E2,Items0,Items) :-
   foreach1(T,E,X1,Assoc,SQ,S,Q2,E20,E2,Items0,Items).

foreach2([],_,_,_,S_prime,S_prime).
foreach2([trans(X1,A/W_prime,Y1_prime)|E],
                  X1,A,W,S_prime0,S_prime) :-
   W_prime @< W, !,
   foreach2(E,X1,A,W,[Y1_prime|S_prime0],S_prime).
foreach2([_|E],X1,A,W,S_prime0,S_prime) :-
   foreach2(E,X1,A,W,S_prime0,S_prime).

foreach3([],_,_,S_prime,S_prime).
foreach3([X2|X2s],E,A,S_prime0,S_prime) :-
   foreach4(E,X2,A,S_prime0,S_prime1),   % 4th foreach loop in R&S
   foreach3(X2s,E,A,S_prime1,S_prime).

foreach4([],_,_,S_prime,S_prime).
foreach4([trans(X2,A/_,Y2_prime)|T],X2,A,S_prime0,S_prime) :- !,
   foreach4(T,X2,A,[Y2_prime|S_prime0],S_prime).
foreach4([_|T],X2,A,S_prime0,S_prime) :-
   foreach4(T,X2,A,S_prime0,S_prime).


%% Dale:
%% I looked a bit at is_functional. One thing R&S say is that you have to
%% treat the case of input epsilon as special. So it should be defined
%% as:

macro(is_fun(Expr),
   { empty(range([] o Expr)), one_string(range([] o Expr)) }
                              &
    t_equal(remove_ambiguity(input_epsilon_remove(~[] o Expr)),
	    input_epsilon_remove(~[] o Expr))).

%% I assume here that input_epsilon_remove also merges multiple start
%% states into one.
%% What I found interesting here is the difficulty in defining one_string
%% in the calculus. One might think this would be easy since empty(Expr)
%% and total(Expr) are so easy to define. But not so.
%% First I define all_strings_same_length:

macro(all_strings_same_length(Expr),
   empty(length([Expr,? +]) & Expr)).

macro(length(Expr),
   range(Expr o (? x ?)*)).

%% So one_string(Expr) can now be defined as:

macro(one_string(Expr),
   all_stings_same_length(Expr)
              &
   no_different_strings_of_same_length(Expr)).


%% Suppose we have a same-length transducer called change, which takes
%% any input string and transduces it to some at least minimally
%% different output string. If we have such a transducer, then we can
%% define: 

macro(no_different_strings_of_same_length(Expr),
   empty(range(Expr o change o Expr))).


%% For any given alphabet, it's easy to define change, but I see no
%% general way of defining it. 

%% Again this whole exercise leaves me wondering what the general
%% principles are. How can you tell when something is going to be
%% definable in the calculus and when not?


%% t_equal implements the core part of IS_FUNCTIONAL from R&S p. 36.

t_equal(I1,F1,E1,I2,F2,E2,Result) :-
   write((I1,I2,plus,[])-0),nl,
   do_loop([(I1,I2,plus,[])-0],[],0,F1,F2,E1,E2,[],Result0),
   (  Result0 == no
   -> Result = no
   ;  Result = yes
   ).



do_loop([],_Assoc0,_SmallQ,_F1,_F2,_E1,_E2,_E3,_Result).
do_loop([(X1,X2,Sign,U)-SmallQ|Items0],Assoc0,SmallQ,
        F1,F2,E1,E2,E30,Result) :-

   Assoc1 = [(X1,X2,Sign,U)-SmallQ|Assoc0],
   SmallQ1 is +(SmallQ,1),

   (  ( Sign == out ;  U \== [] ),
      lists:member(X1,F1),
      lists:member(X2,F2)
   -> Result = no
   ;  Sign == out
   -> foreach3(E1,E2,X1,X2,SmallQ,Assoc1,Items0,Items1,E30,E31),
      do_loop(Items1,Assoc1,SmallQ1,F1,F2,E1,E2,E31,Result)
   ;  foreach1(E1,E2,X1,X2,Sign,U,SmallQ,Assoc1,E30,E31,Items0,Items1,Result),
      (  Result \== no
      -> do_loop(Items1,Assoc1,SmallQ1,F1,F2,E1,E2,E31,Result)
      ;  true
      )
   ).



foreach1([],_E2,_X1,_X2,_Sign,_U,_SQ,_Assoc,E3,E3,Items,Items,_Result).
foreach1([trans(X1,A/W,Y1)|RestE1],E2,X1,X2,Sign,U,SQ,Assoc,
         E30,E3,Items0,Items,Result) :- !,
   foreach2(E2,X2,A,W,Y1,Sign,U,SQ,Assoc,E30,E31,Items0,Items1,Result),
   (  Result == no
   -> true
   ;  foreach1(RestE1,E2,X1,X2,Sign,U,SQ,Assoc,E31,E3,Items1,Items,Result)
   ).
foreach1([_|RestE1],E2,X1,X2,Sign,U,SQ,Assoc,E30,E3,Items0,Items,Result) :-
   foreach1(RestE1,E2,X1,X2,Sign,U,SQ,Assoc,E30,E3,Items0,Items,Result).


foreach2([],_X2,_A,_W,_Y1,_USign,_U,_SQ,_Assoc,E3,E3,Items,Items,_Result).
foreach2([trans(X2,A/W_prime,Y2)|RestE2],X2,A,W,Y1,USign,U,SQ,Assoc,
         E30,E3,Items0,Items,Result) :- !,
   (  USign == plus
   -> V1 = W,
      lists:append(U,W_prime,V2)
   ;  lists:append(U,W,V1),
      V2 = W_prime
   ),
   (  lists:append(V2,Remainder,V1)   % |V1| > |V2|
   -> VSign = minus,
      V = Remainder
   ;  lists:append(V1,Remainder,V2)
   -> VSign = plus,
      V = Remainder
   ;  VSign = out
   ),
   (  ( lists:member((Y1,Y2,VSign_prime,V_prime)-P,Assoc)
      ; lists:member((Y1,Y2,VSign_prime,V_prime)-P,Items0)
      ),
      ( VSign_prime \== VSign ; V_prime \== V ),
      d_hat3(P,SQ,E30,[P])
   -> Result = no
   ;  (  ( lists:member((Y1,Y2,VSign,V)-SmallE,Assoc)
         ; lists:member((Y1,Y2,VSign,V)-SmallE,Items0)
         )
      -> Items1 = Items0,
         E31 = [trans(SQ,A/W,SmallE)|E30]
      ;  lists:append(Items0,[(Y1,Y2,VSign,V)-SmallE],Items1),
         write((Y1,Y2,VSign,V)-SmallE),nl,
         E31 = [trans(SQ,A/W,SmallE)|E30]
      )
   ),
   foreach2(RestE2,X2,A,W,Y1,USign,U,SQ,Assoc,E31,E3,Items1,Items,Result).
foreach2([_|RestE2],X2,A,W,Y1,USign,U,SQ,Assoc,E30,E3,Items0,Items,Result) :-
   foreach2(RestE2,X2,A,W,Y1,USign,U,SQ,Assoc,E30,E3,Items0,Items,Result).



foreach3([],_E2,_X1,_X2,_SQ,_Assoc,E3,E3,Items,Items).
foreach3([trans(X1,A/W,Y1)|RestE1],E2,X1,X2,SQ,Assoc,E30,E3,Items0,Items) :- !,
   foreach4(E2,X2,A,W,Y1,SQ,Assoc,E30,E31,Items0,Items1),
   foreach3(RestE1,E2,X1,X2,SQ,Assoc,E31,E3,Items1,Items).
foreach3([_|RestE1],E2,X1,X2,SQ,Assoc,E30,E3,Items0,Items) :-
   foreach3(RestE1,E2,X1,X2,SQ,Assoc,E30,E3,Items0,Items).


foreach4([],_X2,_A,_W,_Y1,_SQ,_Assoc,E3,E3,Items,Items).
foreach4([trans(X2,A/_W_prime,Y2)|RestE2],X2,A,W,Y1,SQ,Assoc,
         E30,E3,Items0,Items) :- !,
   (  (lists:member((Y1,Y2,out,_)-SmallE,Assoc) ;
       lists:member((Y1,Y2,out,_)-SmallE,Items0))
   -> Items1 = Items0,
      E31 = [trans(SQ,A/W,SmallE)|E30]
   ;  lists:append(Items0,[(Y1,Y2,out,_)-SmallE],Items1),
      write((Y1,Y2,out)-SmallE),nl,
      E31 = [trans(SQ,A/W,SmallE)|E30]
   ),
   foreach4(RestE2,X2,A,W,Y1,SQ,Assoc,E31,E3,Items1,Items).
foreach4([_|RestE2],X2,A,W,Y1,SQ,Assoc,E30,E3,Items0,Items) :-
   foreach4(RestE2,X2,A,W,Y1,SQ,Assoc,E30,E3,Items0,Items).



%% d_hat3(FromState,ToState,Transitions)
%% Find a non-looping path from FromState to ToState
%% The definition is a bit ackward because of my decision to have late
%% instantiation of State numbers.

d_hat3(FromState,ToState,Transitions) :-
   d_hat3(FromState,ToState,Transitions,[]).

d_hat3(P,P,_E30,_Visited).
d_hat3(P,Q,E30,Visited) :-
   lists:member(trans(P1,_,R),E30),
   P == P1,
   \+ ( lists:member(R1,Visited), R == R1 ),
   d_hat3(R,Q,E30,[R|Visited]).





%% Fig 1.26 & 1.29
test1(Result) :-
   _Q1 = [0,1,2,3],
   I1 = 0,        % only one initial state, need extension to allow multiple
   F1 = [2],      % set of finals is sorted
   E1 = [trans(0,a/[b],1),
         trans(0,a/[b,c],3),
         trans(0,a/[b,c,d],2),
         trans(1,b/[c,d],2),
         trans(2,b/[],2),
         trans(3,b/[d],2)],

   _Q2 = [0,1,2,3,4],
   I2 = 0,
   F2 = [3,4],
   E2 = [trans(0,a/[b],1),
         trans(0,a/[b,c],2),
         trans(0,a/[b,c,d],3),
         trans(1,b/[c,d],4),
         trans(4,b/[],4)],

   t_equal(I1,F1,E1,I2,F2,E2,Result).


%% Fig 1.24
test2(Result) :-
   _Q1 = [0,1],
   I1 = 0,
   F1 = [1],
   E1 = [trans(0,a/[c],1),
         trans(0,a/[b],1)],

   _Q2 = [0,1],
   I2 = 0,
   F2 = [1],
   E2 = [trans(0,a/[b],1)],

   t_equal(I1,F1,E1,I2,F2,E2,Result).


%% Fig 1.31 & 1.33
test3(Result) :-
   _Q1 = [0,1,2,3],
   I1 = 0,
   F1 = [3],
   E1 = [trans(0,b/[z],1),
         trans(1,a/[x,x],1),
         trans(1,b/[z],3),
         trans(1,a/[x],2),
         trans(2,a/[],1)],

   _Q2 = [0,1,2,3,4,5],
   I2 = 0,
   F2 = [3],
   E2 = [trans(4,a/[],1),
         trans(2,a/[x],5),
         trans(2,b/[z],3),
         trans(1,a/[x],4),
         trans(1,b/[z],3),
         trans(1,a/[x,x],2),
         trans(0,b/[z],1)],

   t_equal(I1,F1,E1,I2,F2,E2,Result).




