%% test the spelling and lexical info modules in generation mode

:- multifile macro/2.

:- fsa_globals:fsa_global_set(regex_cache,on).

:- bb_put(fsa_rx_spy:generate,on).
:- bb_put(fsa_rx_spy:insert_diacritics,on).
:- bb_put(fsa_rx_spy:apply_stem_changes,on).
:- bb_put(fsa_rx_spy:dictionary_lookup,on).
:- bb_put(fsa_rx_spy:ge_infixation,on).
:- bb_put(fsa_rx_spy:ge_realization,on).
:- bb_put(fsa_rx_spy:spelling,on).

%%%%%%%%%%%%%%%%%% 1. Auxiliary files & macros  %%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% 1a. spelling rules
macro(spelling,file('spelling.fsa')).

%% 1b. dictionary information

%% all stems analyzable as prefix+stem are removed from list:
macro(stems, [ verbal_prefix^, file('verbal_stems.fsa') ] ).

%% all stems from celex:
%% macro(stems, file('stems.fsa')).

%% 1c. replace operator
:- ensure_loaded('../../GerdemannVannoord99/replace'). 


%% 1d misc macros
macro(prefix, [g,e,+] ).
macro(suffix, [+,{['T',e^,n^],t, [e,n]}] ).

macro(letter, a..z).
macro(ltrs, (a..z)*).
macro(vowel, {a,e,i,o,u}).
macro(cons, letter - vowel).


%% translate Xerox-style morphological features into corresponding suffixes
macro(morph_features, ['+Verb':[], { ['+Inf']:[+,e,n]
				   , ['+PresInd+Pl']:[+,e,n]
				   , ['+PresInd+1P+Sg']:[+,#]
				   , ['+Impv+Sg']:[+,#]
				   , ['+PresInd+2P+Sg']:[+,t]
				   , ['+PresInd+3P+Sg']:[+,t]
%% u/jullie(?) leest vaak          , ['+PresInd+2P+Pl']:[+,t]    
				   , ['+PastInd+Sg']:[+,'T',e]
%%	                           , ['+PastInd+2P+Pl']:[+,'T',e]
				   , ['+PastInd+Pl']:[+,'T',e,n]
				   , ['+PastPart']:[+,'T']
				   } ]).


%%%%%%%%%%%%%%%%%%%%%% 2. Generate/Analyze %%%%%%%%%%%%%%%%%%%%%%%%%

%% Generate input is:  (abstract) (PREFIX+)STEM(+SUFFIX)
%% Generate output is: inflected surface form

macro(generate,
       insert_diacritics
     o dictionary_lookup
     o rules
     ).

macro(rules,
       apply_stem_changes
     o ge_infixation
     o ge_realization
     o spelling
     ).

%% Analysis as reverse of generate:

macro(analyze,  inverse(generate)).

macro(analyze_morph,
      t_determinize(inverse( [ ([] x prefix)^, ltrs, morph_features ]
			   o generate
			   )
		   )
     ). 
		
%%%%%%%%%%%%%%%%%%%%%% 3. Insert diacritics  %%%%%%%%%%%%%%%%%%%

%% #    = begin/end of word marker, used in spelling
%% 0    = irregular
%% 1    = don't trigger cons doubling, i.e. reken+en = rekenen, used in spelling
%% 2..9 = strong verbs

macro(strong_verb_marker, {'2','3','4','5','6','7','8','9'} ).

macro(insert_diacritics,
      [[]: # , prefix^ , ltrs
      , { [ []:strong_verb_marker, vowel, letter, cons^ , cons^]
	, [ []:'0', irreg_stems ]
	, [ []:'1',cons ]
	}^
      , suffix^, []: #
      ]
     ).

%%%%%%%%%%%%%%%%%%%%%%%% 4. Dictionary lookup %%%%%%%%%%%%%%%%%%%%%%%%

macro(dictionary_lookup, [ #, prefix^ , stems , suffix^, #] ).

%%%%%%%%%%%%%% 5. Apply alternations, ie do stem changes %%%%%%%%%%%%%

macro(apply_stem_changes,
      [ #, { regular_stem_no_change
	   , strong_verb_stem_change
	   , irregular_verb_stem_change
	   },
	# ]
     ).

%% be specific about prefix/suffix combinations to support reversibility
%% ie lopen -/-> loop+T
%% n.b. this seems to increase the size of the resulting automata considerably
%% over a version which just has prefix optional at the beginning
macro(regular_stem_no_change,
      { [ ltrs, ['1', cons]^, [+,{t,[e,n],['T',e,n^]}]^ ]
      , [ prefix, ltrs, ['1', cons]^, +, 'T']
      }
     ).

macro(strong_verb_stem_change,
	{ [         ltrs, strong_verb_marker:[], ltrs, [+, {t, [e,n]}]^ ]
	, [ prefix, ltrs, strong_verb_participle_stem, +, 'T' x [e,n] ]
	, [         ltrs, strong_verb_past_tense_stem, +, { ['T',e] x []
					  , ['T',e,n] x [e,n]
					  }
	  ]						      
	}
     ).

macro(irregular_verb_stem_change,
      { [ ltrs, '0':[], {   irreg(1)
			     , [ irreg(3), +, t:[] ]
			     , [ irreg(4), +, [e,n] x [] ]
			     , [ irreg(5), +, ['T',e] x [] ]
			     , [ irreg(6), +, ['T',e,n] x [] ]
			     }
	]
      , [ prefix, ltrs, '0':[], irreg(7), +, 'T':[] ]
      }
     ).

macro(strong_verb_participle_stem,
     { ['2':[],[i,j]:[e,e], cons]       %% blijv-gebleven (ANS, p 87, group 1a)
     , ['3':[],{i, [e,e]}:[e,e], cons ] %% leez -gelezen, (2d) bid - gebeden (3b)
     , ['4':[],{i,e}:o, cons+ ]         %% bind -gebonden, glim-geglommen,
                                        %% schenk-geschonken, (1f,1g),
                                        %% vecht-gevochten, zwem-gezwommen
     , ['5':[],{[i,e],[u,i],[e,e],[i,j]} x [o,o], cons ]
                                        %% bied-geboden, sluit-gesloten, (1c,d,e,q)
                                        %% scheer-geschoren, tijg-getogen ?
     , ['6':[],a,a,cons]                %% begraven - begroef -begraven (1h)
     , ['6':[],[e,e]:[o,o], cons]       %% bezweer -bezwoer - bezworen (2c)
     , ['7':[],{a,[a,a],[o,o],[o,e]},cons]
                                        %% blaaz - geblazen, val - gevallen,
                                        %% lopen - gelopen, roep - geroepen (1i,j,l,n) 
     , ['8':[],{[a,n,g],[o,r,d]}]       %% hang -gehangen, vang -gevangen (1k)
                                        %% worden - -geworden (1m)
     , ['9':[],e:o,{l,r},{p,v}]         %% sterv - gestorven, help - geholpen 
                                        %% irregulars with part in -en
     }
     ).
     
macro(strong_verb_past_tense_stem,
     { ['2':[],[i,j]:[e,e], cons ]      %% blijv-bleev (ANS, p 87, group 1a)   
     , ['3':[],{i, [e,e]}:[a,'1'], cons]%% leez -laz - lazen, bid -bad - baden (2d, 3b)
     , ['4':[],{i,e}:o, cons+ ]         %% vecht-vocht, zwem-zwom (1f,1g)
                                        %% bind -bond, glim-glom, schenk-schonk,    
     , ['5':[],{[i,e],[u,i],[e,e],[i,j]} x [o,o], cons ]
                                        %% bied-bood, sluit-sloot, scheer-schoor,
                                        %% tijg-toog ? (1c,d,e,q)
     , ['6':[],{[a,a],[e,e]} x [o,e], cons]  %% begraven - begroef (1h)
                                        %% bezweer -bezwoer
     , ['7':[],{a,[a,a],[o,o],[o,e]} x [i,e],cons]
                                        %% blaaz -bliez, val-viel,
                                        %% lopen -liep, roep -riep (1i,j,l,n), 
     , ['8':[],{[a:i,n,g],[o:e,r,d]}]   %% hang -hing, vang -ving (1k)
                                        %% worden -werd  (1m)
     , ['9':[],e:[i,e],{l,r},{p,v}]     %% sterv -stierv, help -hielp
     }
     ).

%%%%%%%%%%%%%%%%%%%%  6. Spell out ge-prefix %%%%%%%%%%%%%%%%%%%%%%%%%%%%

macro(ge_infixation,
      replace([ [g,e,+] x [],seperable_prefix,[] x [+,g,e,+] ], #, [])
     ).

macro(ge_realization,
      %% for the moment, see ANS 8.3.2.7
      replace([g,e] x [],{+,#},[+,unstressed_prefix])
     ).

macro(seperable_prefix,
      %% long list..., diacritic needed?
      %% these are all prefixes of komen - {be, ont}
      words([aan,achteraan,achterop,af,bijeen,bij,binnen,
	     boven,buiten,door,gelijk,gereed,in,klaar,langs,
	     los,mee,mis,nabij,na,neer,omhoog,om,onder,op,
	     overeen,over,recht,rond,samen,tegen,terecht,
	     terug,thuis,toe,uit,voorbij,voor,
	     voort,vooruit,vrij,weder,weer,weerom,weg]
	   )
     ).

macro(unstressed_prefix,
      words([be,er,ge,her,ont,ver])
     ). 


%%%%%%%%%%%%%%%%%%%%%  X. Define irregular verbs %%%%%%%%%%%%%%%%%

macro(irreg(N),set(Irregulars)) :-
    findall(word(Stem):word(Inflected),
	    (	paradigm(Stem,Sg1,Sg2,Sg3,Pl,PSg,PPl,Part)
	    ,   arg(N,p(Sg1,Sg2,Sg3,Pl,PSg,PPl,Part),Inflected)
	    ),
	    Irregulars
	   ).

macro(irreg_stems, set(Stems)) :-
    findall(word(Stem),paradigm(Stem,_,_,_,_,_,_,_),Stems).

%% combien:combinen,spie:spieen, gym:gymmen,roekoe:roekoeen shampoo:shampoot,tea:teat
%% paradigm(Root,Sg1,Sg3,Plu,PSg,PPlu,Part,Prefixes).
paradigm(kan,kan,kunt,kan,kunnen,kon,konden,gekund).
paradigm(zijn,ben,bent,is,zijn,was,waren,geweest).
paradigm(breng,breng,brengt,brengt,brengen,bracht,brachten,gebracht).
paradigm(denk,denk,denkt,denkt,denken,dacht,dachten,gedacht).
paradigm(doe,doe,doet,doet,doen,deed,deden,gedaan).
paradigm(ga,ga,gaat,gaat,gaan,ging,gingen,gegaan).
paradigm(heb,heb,hebt,heeft,hebben,had,hadden,gehad).
paradigm(hef,hef,heft,heft,heffen,hief,hieven,geheven).
paradigm(houd,houd,houdt,houdt,houden,hield,hielden,gehouden).
paradigm(kom,kom,komt,komt,komen,kwam,kwamen,gekomen).
paradigm(koop,koop,koopt,koopt,kopen,kocht,kochten,gekocht).
paradigm(mag,mag,mag,mag,mogen,mocht,mochten,gemogen).
paradigm(moet,moet,moet,moet,moeten,moest,moesten,gemoeten).
paradigm(sla,sla,slaat,slaat,slaan,sloeg,sloegen,geslagen).
paradigm(sta,sta,staat,staat,staan,stond,stonden,gestaan).
paradigm(verliez,verlies,verliest,verliest,verliezen,verloor,verloren,verloren).
paradigm(vriez,vries,vriest,vriest,vriezen,vroor,vroren,gevroren).
paradigm(weet,weet,weet,weet,weten,wist,wisten,geweten).
paradigm(zal,zal,zult,zal,zullen,zou,zouden,gezuld).
paradigm(zie,zie,ziet,ziet,zien,zag,zagen,gezien).
paradigm(zoek,zoek,zoekt,zoekt,zoeken,zocht,zochten,gezocht).
paradigm(cruise,cruis,cruist,cruist,cruisen,cruisde,cruisden,gecruisd).
paradigm(lease,leas,least,least,leasen,leasde,leasden,geleasd).



%% most frequent prefixes, used in combination with simple_stems.pl/fsa 
macro(verbal_prefix,
      words([ver,
af,
be,
uit,
aan,
op,
in,
om,
door,
ont,
over,
weg,
toe,
na,
terug,
voor,
rond,
onder,
bij,
her,
neer,
samen,
mee,
los,
mis,
voort,
vast,
open,
dood,
binnen,
aaneen,
re,
dicht,
de,
tegen,
vol,
leeg,
bijeen,
weer,
omver,
vrij,
plat,
gelijk,
achteruit,
vooruit,
uiteen,
omhoog,
achter,
stuk,
schoon,
droog,
dooreen,
mede,
weder,
voorbij,
thuis,
ineen,
heen,
glad,
fijn,
goed
	    ]
	   )
     ).

