%%% French plurals from Karttunen, Kaplan, and Zaenen (1992), as implemented 
%%% in appendix 2  of Karttunen (1993)
%%%
%%%  R. Malouf (malouf@let.rug.nl), 20 Apr 2000
%%%
%%% See:
%%%   http://www.rxrc.xerox.com/research/mltt/fssoft/docs/lexc-93/lexc93.html

:- multifile macro/2.

%% Lexical transducer

macro(root, ['#', nouns]).

macro(nouns, { [words([chef,cheval,dieu,htel,nez,travail]), nm],
               [word(dmocrate), nmf],
               [word(eau), nf],
               [word(social), adj],
               [word(travail), nms],
               [words([prorata,'porte-plume']), nmi], 
               [words(['cheval-vapeur','htel-dieu']), nn1m],
               [word('plain-chant'), nn2m],
               [word('eau-de-vie'), nn1f],
               [word('porte-fentre'), nn2f],
               [word('social-dmocrate'), an2mf] } ).

macro(nmi, [[]:'^', nm]).
macro(nm, ['+N', masc]).
macro(nf, ['+N', fem]).
macro(nmf, ['+N', gender]).
macro(nms, [[]:'^s', nm]).
macro(adj, ['+Adj', gender]).
macro(gender, {masc, fem}).
macro(masc, ['+Masc', number]).
macro(fem, ['+Fem', number]).
macro(number, [{'+Sg','+Pl'},#]).

macro(nn1m, [[]:'^n', nm]).
macro(nn1f, [[]:'^n', nf]).
macro(nn2m, [[]:'^nn', nm]).
macro(nn2f, [[]:'^nn', nf]).
macro(an2mf, [[]:'^an', {nm, nf}]).

%% First rule transducer

macro(defaults, 
  {a,,,,b,c,,d,e,,,,,f,g,h,i,,,j,k,l,m,n,,o,,,p,q,r,s,t,
   u,,,v,w,x,y,z,'-','+Sg':0,'+Pl':s,'+Masc':0,'+Fem':0,'+N':0,'+Adj':0,
   '^':0,'^s':0,'#':0}).
macro(other,{'+Pl':0,'+Pl':x,'+Fem':e,l:u,i:0}).
macro(pairs1, {defaults, other}).

macro(sibilant, {s,z,x}).

macro(cat, {'+N', '+Adj'}).
macro(gend, {'+Masc', '+Fem'}).

macro(xcontext, ~['+Adj','+Fem'] o [cat o pairs1, gend o pairs1]).

% No plural marker after a sibilant or diacritic ^

macro(zero_plural, 
  comp(pairs1, '+Pl':0, [{sibilant, '^':0}, cat o pairs1, gend o pairs1], [])).

% Plural is realized as x in words ending in -al, -ail, -au, -eu
%  (except if ^s intervenes).

macro(plural_x,
  comp(pairs1, '+Pl':x, [{a, e}, (i o pairs1)^, pairs1 o u, xcontext], [])).

% Final l is realized as u in the plural of words ending in -al and
%  -ail (unless blocked by ^s). 

macro(l_to_u, 
  comp(pairs1, l:u, [a, (i o pairs1)^], [xcontext, '+Pl' o pairs1])).

% i is deleted in the plural of words ending in -ail (unless blocked by ^s).

macro(i_deletion, 
  comp(pairs1, i:0, a, [l o pairs1, xcontext, '+Pl' o pairs1])).

% Feminine is realized as e in adjectives.

macro(feminine_e,
  comp(pairs1, '+Fem':e, '+Adj' o pairs1, [])).

macro(rules1, 
  twolevel(pairs1, [zero_plural, plural_x, l_to_u, i_deletion, feminine_e])).

%% Second rule transducer

macro(defs2, 
  {a,,,,b,c,,d,e,,,,,f,g,h,i,,,j,k,l,m,n,,o,,,p,q,r,s,t,
   u,,,v,w,x,y,z,'-','+Sg','+Sg':0,'+Pl','+Pl':0,'+Masc':0,'+Masc',
   '+Fem','+Fem':0,'+N':0,'+Adj':0,'+N','+Adj','^','#':0,'^s','^n':0,
   '^nn':0,'^an':0}).
macro(other2,{0:'+N',0:'+Adj',0:'+Masc',0:'+Fem',0:'+Sg',0:'+Pl'}).
macro(pairs2, {defs2, other2}).

macro(letter, 
  {a,,,,b,c,,d,e,,,,,f,g,h,i,,,j,k,l,m,n,,o,,,p,q,r,s,t,u,,,
   v,w,x,y,z,'-'}).

macro(copynmark, { '^n', '^nn' }).
macro(copymark, { copynmark, '^an' }).
macro(deletemark, '^n').
macro(num, {'+Sg', '+Pl'}).

macro(firstword, [ '#' o pairs2, ((letter - '-') o pairs2)+ ]).
macro(lastword, ['-', letter+]).

macro(insertcontext, [ (0 x gend)^, (0 x num)^, lastword]).
macro(copycontext, [ lastword, copymark x 0]).
macro(deletecontext, [deletemark:0, (cat o pairs2)^, (gend o pairs2)^]).

% Insert +N

macro(insert_n, 
  comp(pairs2, 0:'+N', firstword, [insertcontext, copynmark o pairs2])).

% Insert +Adj

macro(insert_adj, 
  comp(pairs2, 0:'+Adj', firstword, [insertcontext, '^an' o pairs2])).

% Copy gender

macro(copy_gender,
  var(Gend, gend,
      comp(pairs2, 0:Gend, [firstword, pairs2 o cat],
                           [(pairs2 o num)^, copycontext, 
                            cat o pairs2, Gend o pairs2]))).

% Copy number

macro(copy_number,
  var(Num, num,
      comp(pairs2, 0:Num, [firstword, pairs2 o cat, pairs2 o gend],
                        [copycontext, cat o pairs2, gend o pairs2,
                         Num o pairs2]))).

% Delete category

macro(del_cat, comp(pairs2, cat:0, deletecontext, [])).

% Delete gender

macro(del_gend, comp(pairs2, gend:0, deletecontext, [])).

% Delete number

macro(del_num, comp(pairs2, num:0, deletecontext, [])).

macro(rules2, 
  twolevel(pairs2, 
     [insert_n,insert_adj,copy_gender,copy_number,del_cat,del_gend,del_num])).

%% Whole thing

macro(french, root o rules2 o rules1).
