%% Optional rule: Trans => LC __ RC
%% Trans is allowed only in context LC_RC.
%% GvN, may 2000.
%% This implements the case of multiple (disjunctive) contexts
%% using the approach of Kaplan and Kay, Regular Models of Phonological
%% Rule Systems. Computational Linguistics 20 (3), 1994.
:- op(398,xfx,(~~)).
:- op(401,yfx,(&&)).
:- multifile macro/2.

:- ensure_loaded(same_length).

macro(E ~~ X,sl_complement(E,X)).
macro(P && Q,sl_intersection(P,Q)).

macro(intro(L), { ? - L , [] x L}*).

macro(if_p_then_s(P,L1,L2), 
      P~~[L1,P~~L2]).
macro(if_s_then_p(P,L1,L2), 
      P~~[P~~L1,L2]).
macro(p_iff_s(P,L1,L2), 
      if_p_then_s(P,L1,L2) 
            &&
      if_s_then_p(P,L1,L2)).
macro(restrict(P,T,L,R),
      if_p_then_s(P,[P*,T],[R,P*])
            &&
      if_s_then_p(P,[P*,L],[T,P*])
     ).

macro(right_context(P,Markers,R,M),
      reverse(left_context(P,Markers,reverse(R),M))).

macro(left_context(P,Markers,L,M),
      if_s_then_p(P,[P*,ignore0(L,Markers)],[M,P*])).

%% introduce a marker after every instance of Trans
macro(opt(P,Trans,LC1,RC1,LC2,RC2),
          intro(M)
              o
     (  { restrict({P,M},Trans,l1,r1),
	  restrict({P,M},Trans,l2,r2) }*
                     &&
          left_context({P,M},M,LC1,l1)
                     &&
	  right_context({P,M},M,RC1,r1)
	             &&
          left_context({P,M},M,LC2,l2)
	             &&
          right_context({P,M},M,RC2,r2)
      )                              
              o
      inverse(intro(M))
     ) :- M={l1,l2,r1,r2}.

%% introduce a marker after every instance of Trans
macro(opt(P,Trans,LC1,RC1,LC2,RC2,LC3,RC3),
          intro(M)
              o
     (  { restrict({P,M},Trans,l1,r1),
	  restrict({P,M},Trans,l2,r2),
	  restrict({P,M},Trans,l3,r3)}*
                     &&
          left_context({P,M},M,LC1,l1)
	             &&
	  right_context({P,M},M,RC1,r1)
	             &&
          left_context({P,M},M,LC2,l2)
	             &&
	  right_context({P,M},M,RC2,r2)
	             &&
          left_context({P,M},M,LC3,l3)
	             &&
	  right_context({P,M},M,RC3,r3)
      )                              
              o
      inverse(intro(M))
     ) :- M={l1,l2,l3,r1,r2,r3}.
