%% Macros for two-level rules, in the style of PC-KIMMO
%% Rob Malouf, 11-Apr-2000
%% Adapted by Gertjan van Noord, 9-May-2000,
%% in order to use the new same-length macro's.
%% Also fixed a problem with obligatory insertion rules, 13-May-2000
:- multifile macro/2.
:- discontiguous macro/2.

:- op(398,xfx,(~~)).
:- op(401,yfx,(&&)).

:- ensure_loaded(same_length).
:- ensure_loaded('../Booleans/booleans').

macro(E ~~ X,sl_complement(E,X)).
macro(A && B,sl_intersection(A,B)).

%% Some useful operations on transducers
macro(big_and(List),Expr) :-
    sl_intersect_list(List,Expr).

%% to find the composition of a list of expressions, first
%% take the composition of the first half of the list, then the
%% second half, and then intersect
sl_intersect_list(List,Expr) :-
    length(List,Length),
    sl_intersect_list(Length,List,[],Expr).

sl_intersect_list(Length,List0,List,Expr) :-
    (	Length =:= 0
    ->	Expr =  kleene_star(?),
	List0=List
    ;	Length =:= 1
    ->	List0 = [Expr|List]
    ;	Length =:= 2
    ->	List0 = [E1,E2|List], Expr=sl_intersection(E1,E2)
    ;	A is Length>>1,
	Z is Length-A,
	Expr=sl_intersection(E1,E2),
	sl_intersect_list(A,List0,List1,E1),
	sl_intersect_list(Z,List1,List,E2)
    ).



%% Obligatory rule: Trans <= LC __ RC

macro(oblig(P,Trans0,LC,RC), pragma([Trans-Trans0], 
      if(subset(0,domain(Trans)),
	 check_zero(P,LC,RC)
	     &&
         oblig0(P,Trans,LC,RC),
	 oblig0(P,Trans,LC,RC)
	))).

%% the zero case is complex because either of LC, RC might be the
%% empty string. For instance, if RC is the empty string, we must
%% be careful not to rule out all strings containing LC since LC
%% might be followed correctly by 0.
macro(check_zero(P0,LC,RC), pragma([P-P0], 
      {P~~[P*,LC,[RC,P*] && P~~[0 o P,P*]],
       P~~[[P*,LC] && P~~[P*,0 o P],RC,P*]})).
      

%% case 2 (normal)
macro(oblig0(P0,Trans0,LC,RC), pragma([P-P0,Trans-Trans0], 
      P~~[P*,LC,domain(Trans) x {range(P)-range(Trans)},RC,P*])).

%% Optional rule: Trans => LC __ RC
macro(opt(P0,Trans0,LC,RC), pragma([P-P0,Trans-Trans0],
      P~~{[P~~[P*,LC],Trans,P*],
	  [P*,Trans,P~~[RC,P*]]})).

%% Composite rule: Trans <=> LC __ RC

macro(comp(P0,Trans0,LC0,RC0), pragma([P-P0,Trans-Trans0,LC-LC0,RC-RC0],
      opt(P,Trans,LC,RC)
             &&
      oblig(P,Trans,LC,RC))).
    

%% Exclusion rule: Trans /<= LC __ RC

macro(excl(P0,Trans,LC,RC), pragma([P-P0],
      P~~[P*,LC,Trans,RC,P*])).
    

%% Handle insertions and deletions in same-length transducers using a special
%% null character '0'.
macro(twolevel(_,Rules),twolevel(Rules)).

macro(twolevel(Rules),
      {? ,[]:0}*
          o
      big_and(Rules)
          o
      {? - 0,0:[]}*     ).


%% Useful stuff for building lexicons
macro(words(List0), set(List1)) :-
    words(List0,List1).

words([],[]).
words([L:S|T0], [word(L):word(S)|T]) :-
    words(T0,T).
words([H0|T0], [word(H0)|T]) :-
    words(T0,T).

%% Matching variables
macro(var(Var, Set, Expr), big_and(List)) :-
    fsa_regex:rx(Set, Fa),
    findall(Expr, fsa_interpreter:fsa_accepts([Var],Fa),List).



  
