#!/usr/bin/python

## convert an FSA read a list of alternative spellings, and generate
## an FSA module to match texts that contain one of the alternates
## (case sensitive)
## author: Rob Malouf, rmalouf@mail.sdsu.edu

"""

Uses to SWIG to produce and compile wrapper library for fsa-generated
C code as a python module.

For example:

% fsa -aux replace.pl -r "{a,b,c}* o replace(b:a,[c],[])" | fsa -cc c_with_main=off >ex.c
% fsa2py.py ex.c
% python
Python 2.4.1 (#1, May 16 2005, 15:19:29)
[GCC 4.0.0 20050512 (Red Hat 4.0.0-5)] on linux2
Type "help", "copyright", "credits" or "license" for more information.
>>> import ex
>>> ex.t_accepts('bbbbbba')
[1, 'bbbbbba']
>>> ex.t_accepts('bbbcbbba')
[1, 'bbbcabba']
>>> ex.t_accepts('bbbcbbxba')
[0, 'bbbcab']
>>>

"""

import sys,re,os
from distutils.sysconfig import customize_compiler,get_python_inc
from distutils.ccompiler import new_compiler

maxlen = 32768   # maximum length of returned string -- overflows will crash!

# get file names

cfile = sys.argv[1]
module = re.match(r'.*(?=\.)',cfile).group()

# create SWIG interface file

i = open(module+'.i','w')
print >>i,'%module',module
print >>i,'%{'
print >>i,'%}'
print >>i

# scan C code

flag = False

for line in open(cfile,'r'):
    if line.find('wt_accepts') != -1:
        print >>i,'%include "typemaps.i"'
        print >>i,'%include cstring.i'
        print >>i,'%apply int *OUTPUT { int *wout };'
        print >>i,'%%cstring_bounded_output(char *out, %d);'%maxlen
        print >>i,'int wt_accepts(char *in,char *out, int *wout);'
    elif line.find('w_accepts') != -1:
        print >>i,'%include "typemaps.i"'
        print >>i,'%apply int *OUTPUT { int *out };'
        print >>i,'int w_accepts(char *in, int *out);'
        flag = True
    elif line.find('t_accepts') != -1:
        print >>i,'%include cstring.i'
        print >>i,'%%cstring_bounded_output(char *out, %d);'%maxlen
        print >>i,'int t_accepts(char *in,char *out);'
        flag = True
    elif line.find('accepts') != -1:
        print >>i,'int accepts(char *in);'
        flag = True

if not flag:
    print >>sys.stderr, cfile,"doesn't look like an FSA!"
    sys.exit()
    
i.close()

# generate wrapper

os.system('swig -python %s.i'%module)

# compile and link

c = new_compiler()
customize_compiler(c)
obj = c.compile([cfile,module+'_wrap.c'],include_dirs=[get_python_inc()])
c.link_shared_object(obj,'_'+module+'.so')
