:- module( fsa_fl, [ fsa_fl_fa/3,
		     fsa_fl_fa/4,
		     fsa_fl_file/3,
		     fsa_fl_file/4,
		     fsa_fl/4
		   ]).

%% the C++ part as suggested by Dale Gerdemann (although I ignored a
%% few suggestions for improvement ...)

#swi# :- style_check([-atom]).
#yap# :- use_module(library(lists), [ select/3, member/2 ]).
#sic# :- use_module(library(lists), [ select/3 ]).

fsa_fl_file(Lang,Type,Source,Target) :-
    fsa_io:fsa_read_file(Source,Fa),
    fsa_fl_fa(Lang,Type,Fa,Target).

fsa_fl_file(Lang,Source,Target) :-
    fsa_io:fsa_read_file(Source,Fa),
    fsa_fl_fa(Lang,Fa,Target).

fsa_fl_fa(Lang,Fa,Target) :-
    fsa_data:fsa_type(Fa,Type),
    fsa_fl_fa(Lang,Type,Fa,Target).

fsa_fl_fa(Lang,Type,Fa,Target0) :-
    fsa_fl_parse_file(Lang,Type,Target0,Target,Label),
    fsa_util:to_file(fsa_fl:fsa_fl(Lang,Type,Label,Fa),Target).

fsa_fl_parse_file(c,_,Path,Path,Path).
fsa_fl_parse_file(cpp,_,Path,Path,Path).
fsa_fl_parse_file(java,_,Path,File,Label) :-
    (	Path == user
    ->	File=user, Label=user
    ;	(   atom_concat(Path0,'.java',Path)
	->  File=Path
	;   atom_concat(Path,'.java',File),
	    Path0=Path
	),
	core_name(Path0,Label)
    ).
		 
%% core_name
core_name(Path,Name) :-
    atom_codes(Path,Codes),
    core_name_(Codes,T,T,NCodes),
    atom_codes(Name,NCodes).

core_name_([],[],Sol,Sol).
core_name_([H|T],Sol0,Sol,N):-
    (	H =:= 47
    ->	core_name_(T,S,S,N)
    ;   Sol0=[H|Sol1],
	core_name_(T,Sol1,Sol,N)
    ).

fsa_fl(Lang,Type,Label,Fa0) :-
    get_main(Lang,Main),
    get_conversion(Lang,Converse),
    converse_fa(Converse,Type,Fa0,Fa1,Specials),
    select_epsilon_start(Type,Fa1,Fa2,Ini),
    with_final_output(Type,Fa2,Fa,FinalOutput,FinalTable),
    calc_nr_next(Fa,NextTable),
    fsa_data:fsa_start_states(Fa,[Start]),
    fsa_data:fsa_transitions(Fa,T),
    length(T,Length0), Length is Length0+1,
    format_comment,
    format_headers(Lang),
    format_queue(Type,Lang),
    format_init(Lang,Type,Length,Label,Ini,FinalOutput,
		FinalTable,NextTable,Specials),
    format_sink(Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable),
    format_start(Lang,Type,Label,Start,Ini,FinalOutput,FinalTable,NextTable),
    format_transitions(T,Lang,Type,FinalOutput,FinalTable,NextTable,Fa),
    format_transitions_end(Lang),
    format_more(Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable),
    format_main(Main,Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable),
    format_tail(Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable).


get_main(java,Main) :-
    fsa_globals:fsa_global_get(java_with_main,Main).
get_main(c,Main) :-
    fsa_globals:fsa_global_get(c_with_main,Main).
get_main(cpp,Main) :-
    fsa_globals:fsa_global_get(cpp_with_main,Main).

get_conversion(java,Converse) :-
    fsa_globals:fsa_global_get(to_java_conversion,Converse).
get_conversion(c,Converse) :-
    fsa_globals:fsa_global_get(to_c_conversion,Converse).
get_conversion(cpp,Converse) :-
    fsa_globals:fsa_global_get(to_cpp_conversion,Converse).

converse_fa(off,_,Fa,Fa,[]).
converse_fa(on,Type,Fa0,Fa,Syms):-
    symbols_talked_about(Fa0,Syms0),
    unknown_symbol(Special),
    n_expand_predicates_fa(Type,Fa0,Fa1,[Special|Syms0]),
    one_letter_atoms(Syms0,Syms1),
    sort(Syms1,Syms),
    converse_fa(Type,Fa1,Fa).

one_letter_atoms([],[]).
one_letter_atoms([H0|T],Atoms) :-
    ensure_atom(H0,H),
    atom_length(H,L),
    (	L =:= 1
    ->	Atoms=[H|Atoms1]
    ;	Atoms=Atoms1
    ),
%    atom_to_atoms(H,Atoms,Atoms1),
    one_letter_atoms(T,Atoms1).

atom_to_atoms(Atom,Atoms,Tail) :-
    atom_codes(Atom,String),
    string_to_atoms(String,Atoms,Tail).

string_to_atoms([],Tail,Tail).
string_to_atoms([Int|Ints],[Atom|Atoms0],Atoms):-
    atom_codes(Atom,[Int]),
    string_to_atoms(Ints,Atoms0,Atoms).

converse_fa(recognizer,Fa0,Fa):-
    fsa_regex:fsa_regex_compile(determinize(fa(Fa0)),Fa).
converse_fa(w_recognizer,Fa0,Fa):-
    fsa_regex:fsa_regex_compile(w_determinize(fa(Fa0)),Fa).
converse_fa(transducer(_),Fa0,Fa):-
    fsa_regex:fsa_regex_compile(t_determinize(fa(Fa0)),Fa).
converse_fa(w_transducer(_),Fa0,Fa):-
    fsa_regex:fsa_regex_compile(wt_determinize(fa(Fa0)),Fa).

select_epsilon_start(recognizer,Fa,Fa,"\"\"").
select_epsilon_start(transducer(_),Fa0,Fa,IniOut):-
    fsa_data:fsa_copy_except(start_states,Fa0,Fa1,[Start0],[Start]),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    (	select(trans(Start0,[]/Out,Start),Trans0,Trans),
	\+ member(trans(Start0,_,_),Trans)
    ->	true
    ;   Fa0=Fa,
	Out=[]
    ),
    string_symbol(Out,IniOut).

select_epsilon_start(w_recognizer,Fa0,Fa,IniOut):-
    fsa_data:fsa_copy_except(start_states,Fa0,Fa1,[Start0],[Start]),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    (	select(trans(Start0,[]/Out,Start),Trans0,Trans),
	\+ member(trans(Start0,_,_),Trans)
    ->	true
    ;   Fa0=Fa, Out=0
    ),
    number_codes(Out,IniOut).

select_epsilon_start(w_transducer(_),Fa0,Fa,S/W):-
    fsa_data:fsa_copy_except(start_states,Fa0,Fa1,[Start0],[Start]),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans0,Trans),
    (	select(trans(Start0,[]/(Out/OutW),Start),Trans0,Trans),
	\+ member(trans(Start0,_,_),Trans)
    ->	true
    ;   Fa0=Fa, OutW=0, Out=[]
    ),
    number_codes(OutW,W),
    string_symbol(Out,S).


%% for each state, find out the position in the transition list.
calc_nr_next(Fa,Table):-
    fsa_data:fsa_states_number(Fa,Size),
    fsa_array:fsa_array_new(Table,Size),
    fsa_data:fsa_transitions(Fa,Trs),
    calc_nr_next0(Trs,Table).

calc_nr_next0([],_).
calc_nr_next0([trans(P,_,_)|Tail],Table):-
    calc_nr_next(Tail,P,2,2,Table).

calc_nr_next([],Prev,NrSoFar,_TransNr,Table):-
    fsa_array:fsa_array_access(Prev,NrSoFar,Table).
calc_nr_next([trans(P,_,_)|Tail],Prev,NrSoFar,TransNr,Table):-
    TransNr1 is TransNr+1,
    (	P==Prev
    ->	calc_nr_next(Tail,Prev,NrSoFar,TransNr1,Table)
    ;   fsa_array:fsa_array_access(Prev,NrSoFar,Table),
	calc_nr_next(Tail,P,TransNr1,TransNr1,Table)
    ).

with_final_output(recognizer,Fa,Fa,[],Table) :-
    fsa_data:index_finals(Fa,Table).

with_final_output(w_recognizer,Fa,Fa,[],Table) :-
    fsa_data:index_finals(Fa,Table).

with_final_output(transducer(_),Fa0,Fa,FinalOutputs,FTable):-
    fsa_data:fsa_states_number(Fa0,Size),
    fsa_array:fsa_array_new(FinalOutputs,Size),
    fsa_data:fsa_copy_except(transitions,Fa0,Fa1,Trans0,Trans),
    fsa_data:fsa_copy_except(final_states,Fa1,Fa,Finals0,Finals),
    with_final_output0(Trans0,Trans,Finals1,Finals0,FinalOutputs,t),
    sort(Finals1,Finals),
    fsa_data:index_finals(Fa,FTable).

with_final_output(w_transducer(_),Fa0,Fa,FinalOutputs,FTable):-
    fsa_data:fsa_states_number(Fa0,Size),
    fsa_array:fsa_array_new(FinalOutputs,Size),
    fsa_data:fsa_copy_except(transitions,Fa0,Fa1,Trans0,Trans),
    fsa_data:fsa_copy_except(final_states,Fa1,Fa,Finals0,Finals),
    with_final_output0(Trans0,Trans,Finals1,Finals0,FinalOutputs,w),
    sort(Finals1,Finals),
    fsa_data:index_finals(Fa,FTable).

with_final_output1([],[],F,F,[],_).
with_final_output1([trans(P,L/R,Q)|TransT],Trans,F0,F,Outs,TW):-
    (	L == []
    ->	Trans1=Trans,
	F0=[P|F1],
	Outs=[P-R|Outs1]
    ;	F0=F1,
	Trans=[trans(P,L/R,Q)|Trans1],
	Outs=Outs1
    ),
    with_final_output1(TransT,Trans1,F1,F,Outs1,TW).

with_final_output0(Tr0,Tr,F0,F,OutputsArray,TW) :-
    with_final_output1(Tr0,Tr,F0,F,Outputs0,TW),
    sort(Outputs0,Outputs),
    combine_outputs(Outputs,OutputsArray,TW).

combine_outputs([],_,_).
combine_outputs([P-Out|T],Ar,TW) :-
    combine_outputs(T,P,[Out],Ar,TW).

combine_outputs([],P,Outs,Ar,TW) :-
    fsa_array:fsa_array_access(P,Out,Ar),
    combine_output_seq(TW,Outs,Out).
combine_outputs([P1-Out1|T],P,Outs0,Ar,TW) :-
    (	P==P1
    ->	combine_outputs(T,P,[Out1|Outs0],Ar,TW)
    ;	fsa_array:fsa_array_access(P,Out,Ar),
	combine_output_seq(TW,Outs0,Out),
	combine_outputs(T,P1,[Out1],Ar,TW)
    ).

combine_output_seq(t,List,Out) :-
    combine_output_seq_t(List,Out).
combine_output_seq(w,List,Out/Score) :-
    get_scores(List,List1,Score),
    combine_output_seq_t(List1,Out).

get_scores([],[],_).
get_scores([H/Score|T],[H|NT],Score) :-
    get_scores(T,NT,Score).

combine_output_seq_t([H|T],Out):-
    combine_output_seq_t(T,H,Out).

combine_output_seq_t([],H,[Sym]) :-
    fsa_util:concat_all(H,Sym).
combine_output_seq_t([H|T],F,[Sym]):-
    flatten1([F,H|T],List),
    fsa_globals:fsa_global_get(fl_multiple_symbol_start,Start),
    fsa_globals:fsa_global_get(fl_multiple_symbol_separator,Sep),
    fsa_util:concat_all(List,Sym0,Sep),
    atom_concat(Start,Sym0,Sym).

flatten1([],[]).
flatten1([H|T],[Atom|L]):-
    fsa_util:concat_all(H,Atom),
    flatten1(T,L).

get_single_symbol(Module,Pred,L,Sym) :-
    findall(L0,fsa_pred_module:evaluate_predicate(Module,Pred,L0),Ls),
    check_get_single_symbol(Ls,Sym,Module,Pred),
    (	atom(Sym),
	atom_codes(Sym,[L])
    ->	true
    ;	number(Sym),
	number_codes(Sym,[L])
    ).

check_get_single_symbol([],_,Mod,P) :-
    user:throw(fsa_error("~w:~w is not true of a single symbol~n",[Mod,P])).

check_get_single_symbol([H|T],L,Module,Pred) :-
    check_get_single_symbol(T,H,L,Module,Pred).

check_get_single_symbol([],L,L,_,_).
check_get_single_symbol([_|_],_,_,Mod,P) :-
    user:throw(fsa_error("~w:~w is true of more than one symbol~n",[Mod,P])).
    
symbol_mark(recognizer,Fa,Pred,L,[]):-
    fsa_data:fsa_symbols_decl(Fa,r(Module)),
    get_single_symbol(Module,Pred,L,_).

symbol_mark(transducer(_),Fa,'$@'(Pred)/['$@'(Pred)],L,R):-
    fsa_data:fsa_symbols_decl(Fa,t(Module,Module)),
    !,
    get_single_symbol(Module,Pred,L,Sym),
    (	integer(Sym)
    ->  integer_symbol(Sym,R)
    ;   string_symbol([Sym],R)
    ).

symbol_mark(transducer(_),Fa,PredL/PredR,L,R):-
    fsa_data:fsa_symbols_decl(Fa,t(ModuleL,ModuleR)),
    fsa_pred_module:evaluate_predicate(ModuleL,PredL,SymL),
    get_single_symbol(ModuleL,PredL,L,SymL),
    evaluate_predicate_list(PredR,SymR,ModuleR),
    (	integer(SymR)
    ->  integer_symbol(SymR,R)
    ;   string_symbol(SymR,R)
    ).

symbol_mark(w_transducer(_),Fa,'$@'(Pred)/(['$@'(Pred)]/SymR),L,Sr/R):-
    fsa_data:fsa_symbols_decl(Fa,t(Module,Module,fsa_frozen)),
    !,
    get_single_symbol(Module,Pred,L,Sym),
    (	integer(Sym)
    ->  integer_symbol(Sym,Sr)
    ;   string_symbol([Sym],Sr)
    ),
    integer_symbol(SymR,R).


symbol_mark(w_transducer(_),Fa,PredL/(PredR/SymR),L,Sr/R):-
    fsa_data:fsa_symbols_decl(Fa,t(ModuleL,ModuleR,fsa_frozen)),
    fsa_pred_module:evaluate_predicate(ModuleL,PredL,SymL),
    get_single_symbol(ModuleL,PredL,L,SymL),
    evaluate_predicate_list(PredR,NrR,ModuleR),
    (	integer(NrR)
    ->  integer_symbol(NrR,Sr)
    ;   string_symbol(NrR,Sr)
    ),
    integer_symbol(SymR,R).

symbol_mark(w_recognizer,Fa,PredL/SymR,L,R):-
    fsa_data:fsa_symbols_decl(Fa,r(ModuleL,fsa_frozen)),
    get_single_symbol(ModuleL,PredL,L,_SymL),
    integer_symbol(SymR,R).

evaluate_predicate_list([],[],_).
evaluate_predicate_list([Pred|Preds],[Sym|Syms],Module) :-
    fsa_pred_module:evaluate_predicate(Module,Pred,Sym),
    evaluate_predicate_list(Preds,Syms,Module).

integer_symbol(Int,IntChars):-
    number_codes(Int,IntChars).

string_symbol(List,[34|Chars]) :-
    string_symbol0(List,Chars,[34]).

string_symbol0([],S,S).
string_symbol0([H|T],S0,S) :-
    escape_symbol(H,S0,S1),
    string_symbol0(T,S1,S).

escape_symbol(Sym,C0,C) :-
    (	number(Sym)
    ->	number_codes(Sym,[Code|Codes])
    ;   atom(Sym)
    ->	atom_codes(Sym,[Code|Codes])
    ;   format(user_error,"Warning: output symbol is not atomic: ~q~n",[Sym]),
	fsa_util:term_atom(Sym, SymAtom),
	atom_codes(SymAtom,[Code|Codes])
    ),
    escape_symbol1(Code,C0,C1),
    escape_symbol_(Codes,C1,C).

escape_symbol_([],C,C).
escape_symbol_([H|T],C0,C) :-
    escape_symbol1(H,C0,C1),
    escape_symbol_(T,C1,C).

escape_symbol1(Sym,C0,C) :-
    (	Sym =:= 92
    ->	C0=[92,92|C]
    ;   Sym =:= 34
    ->  C0 = [92,34|C]
    ;   Sym =:= 10   % newlines need to be escaped these days too!
    ->  C0 = [92,110|C]
    ;   C0=[Sym|C]
    ).

format_comment :-
    format(
  " /* This file is generated automatically by FSA Utilities.\
 \n  * (c) 1993 - 2000 Gertjan van Noord RUG\
 \n  * More information from http://www.let.rug.nl/~~vannoord/Fsa/\
 \n  * \
 \n  * The representation of a finite-automaton is\
 \n  * similar to the technique explained on page 43 (table\
 \n  * 4.2) of Jan Daciuk's dissertation `Incremental\
 \n  * Construction of Finite-State Automata and Transducers\
 \n  * and their use in the Natural Language Processing'.\
 \n  * Politechnika Gdanska, 1998, except that instead of the number\
 \n  * of transitions we have a boolean flag indicating for each\
 \n  * line whether that line is the last transition for the current\
 \n  * state.\
 \n  */  \
 \n  \
 \n ",[]).

format_headers(c) :-
    format("\
 \n #include <string.h>\
 \n #include <stdlib.h>",[]).
format_headers(cpp) :-
    format("\
 \n #include <stdlib.h>\
 \n #include <iostream>\
 \n #include <string>\
 \n #include <queue>\
 \n using namespace std;",[]). 

format_headers(java) :-
    format("import FSA.*;~n",[]).

format_init(cpp,Type,Length,_L,_I,_FinalOutput,_FinalTable,_NextTable,S) :-
    format_unknown_symbol(cpp,S),
    format_procedure_head_cpp(Type),
    (	Length < 2**22
    ->	format("\
 \n   static const struct transition_struct {\
 \n     unsigned char symbol :  8;      // current symbol\
 \n     bool last            :  1;      // last line of current state?\
 \n     bool final           :  1;      // final state?\
 \n     unsigned int next    : 22;      // line in table to go to next ",[])
    ;   format("\
 \n   static const struct transition_struct {\
 \n     char symbol              :  8;      // current symbol\
 \n     bool last                :  1;      // last line of current state?\
 \n     bool final               :  1;      // final state?\
 \n     unsigned long int next   : 54;      // line in table to go to next ",[])
    ),
    format_header_output_cpp(Type),
    format(" \n  } trans[]= {~n",[]).

format_init(c,Type,Length,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable,S) :-
    format_unknown_symbol(c,S),
    format_procedure_head(Type),
    (	Length < 2**22
    ->	format("\
 \n   static const struct transition_struct {\
 \n     unsigned char symbol :  8;      /* current symbol                     */\
 \n     unsigned int last    :  1;      /* last line of current state?        */\
 \n     unsigned int final   :  1;      /* final state?                       */\
 \n     unsigned int next    : 22;      /* line in table to go to next        */ ",[])
    ;   format("\
 \n   static const struct transition_struct {\
 \n     unsigned char symbol     :  8;      /* current symbol                */\
 \n     unsigned int last        :  1;      /* last line of current state?   */\
 \n     unsigned int      final  :  1;      /* final state?                  */\
 \n     unsigned long int next   : 54;      /* line in table to go to next   */ ",[])
    ),
    format_header_output(Type),
    format(" \n  } trans[]= {~n",[]).



format_init(java,Type,_L,Label,Ini,_FinalOutput,_FinalTable,_NextTable,S) :-
    format_class_head(Type,Label,Ini),
    format_unknown_symbol(java,S),
    format("    private static final Transition tA [] = {~n",[]).

format_header_output(recognizer).
format_header_output(transducer(_)) :-
    format(
" \n   char *out;                     /* output string                      */\
 \n     char *final_out;               /* output string if we terminate here */",[]).
format_header_output(w_recognizer) :-
    format(
" \n   unsigned int out;              /* output weight                      */\
 \n     unsigned int final_out;        /* output weight if we terminate here */",[]).
 
format_header_output(w_transducer(_)) :-
    format(
" \n   char *out;                 /* output string                      */\
 \n     char *final_out;           /* output string if we terminate here */\
 \n     unsigned int wout;         /* output weight                      */\
 \n     unsigned int final_wout;   /* output weight if we terminate here */",[]).
 
format_header_output_cpp(recognizer).
format_header_output_cpp(transducer(_)) :-
    format(
" \n   char *out;                     // output string\
 \n     char *final_out;               // output string if we terminate here",[]).
format_header_output_cpp(w_recognizer) :-
    format(
" \n   unsigned int wout;              // output weight\
 \n     unsigned int final_wout;        // output weight if we terminate here",[]).
 
format_header_output_cpp(w_transducer(_)) :-
    format(
" \n   char *out;                 // output string\
 \n     char *final_out;           // output string if we terminate here\
 \n     unsigned int wout;         // output weight\
 \n     unsigned int final_wout;   // output weight if we terminate here",[]).
 
format_procedure_head(recognizer) :-
    format("\
 \n \n int accepts(char *in) {~n",[]).
format_procedure_head(transducer(_)) :-
    format("\
 \n \n /* int t_accepts(char *in, char *out, int max)\
 \n    in is transduced to out. Note that memory for out should be allocated\
 \n    by the caller. At most max characters are writting into out. \
 \n    The return values:\
 \n \
 \n    0: no transduction possible\
 \n    1: transduction possible\
 \n    2: length of transduction would be > max\
 \n */\
 \n \
 \n int t_accepts(char *in,char *out,int max) {~n",[]).
format_procedure_head(w_recognizer) :-
    format("\
 \n \n int w_accepts(char *in, int *out) {~n",[]).

format_procedure_head(w_transducer(_)) :-
    format("\
 \n \n /* int wt_accepts(char *in, char *out, int *wout, int max)\
 \n    in is transduced to out, with associated weight wout. \
 \n    Note that memory for out should be allocated\
 \n    by the caller. At most max characters are writting into out. \
 \n    The return values:\
 \n \
 \n    0: no transduction possible\
 \n    1: transduction possible\
 \n    2: length of transduction would be > max\
 \n */\
 \n \
 \n int wt_accepts(char *in, char *out, int *wout,int max) {~n",[]).

format_procedure_head_cpp(recognizer) :-
    format("\
 \n \n bool accepts(string::const_iterator input_iter,  \
 \n                 const string::const_iterator end) {~n",[]).

format_procedure_head_cpp(transducer(_)) :-
    format("\
 \n \n bool t_accepts(string::const_iterator input_iter,  \
 \n                 const string::const_iterator end,  \
 \n                 string& out) {~n",[]).

format_procedure_head_cpp(w_recognizer) :-
    format("\
 \n \n bool w_accepts(string::const_iterator input_iter,  \
 \n                 const string::const_iterator end,  \
 \n                 unsigned int& wout) {~n",[]).

format_procedure_head_cpp(w_transducer(_)) :-
    format("\
 \n \n bool wt_accepts(string::const_iterator input_iter,  \
 \n                 const string::const_iterator end,  \
 \n                 string& out, \
 \n                 unsigned int& wout) {~n",[]).


format_class_head(recognizer,Class,_Ini) :-
    format("\
 \n \
 \n public class ~w extends DFAapplet {\
 \n \
 \n     public ~w () {\
 \n         super(new DFA(tA,unknown_string));\
 \n     }",[Class,Class]).

format_class_head(transducer(_),Class,Ini) :-
    format("\
 \n \
 \n public class ~w extends DFAapplet {\
 \n \
 \n     public ~w () {\
 \n         super(new DFA(tA,~s,unknown_string));\
 \n     }",[Class,Class,Ini]).

format_class_head(w_recognizer,Class,Ini) :-
    format("\
 \n \
 \n public class ~w extends DFAapplet {\
 \n \
 \n     public ~w () {\
 \n         super(new DFA(tA,~s,unknown_string));\
 \n     }",[Class,Class,Ini]).


format_class_head(w_transducer(_),Class,IniS/IniW) :-
    format("\
 \n \
 \n public class ~w extends DFAapplet {\
 \n \
 \n     public ~w () {\
 \n         super(new DFA(tA,~s,~s,unknown_string));\
 \n     }",[Class,Class,IniS,IniW]).


format_sink(c,Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable):-
    format("                     /* sink                               */~n",
	   []),
    format_sink_c(Type).
format_sink(cpp,Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable):-
    format("                     // sink~n",
	   []),
    format_sink_cpp(Type).
format_sink(java,Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable):-
    format("        /* sink                               */~n",
	   []),
    format_sink_java(Type).

format_sink_c(recognizer):-
    format('    {0,1,0,0},~n',[]).

format_sink_c(transducer(_)):-
    format('    {0,1,0,0,"",""},~n',[]).

format_sink_c(w_recognizer):-
    format('    {0,1,0,0,0,0},~n',[]).

format_sink_c(w_transducer(_)):-
    format('    {0,1,0,0,"","",0,0},~n',[]).


format_sink_cpp(recognizer):-
    format('    {\'\\0\',1,0,0},~n',[]).

format_sink_cpp(transducer(_)):-
    format('    {\'\\0\',1,0,0,"",""},~n',[]).

format_sink_cpp(w_recognizer):-
    format('    {\'\\0\',1,0,0,0,0},~n',[]).

format_sink_cpp(w_transducer(_)):-
    format('    {\'\\0\',1,0,0,"","",0,0},~n',[]).


format_sink_java(recognizer):-
    format('        new Transition(0,true,false,0),~n',
         []).


format_sink_java(transducer(_)):-
    format('        new Transition(0,true,false,0,"",""),~n',
         []).


format_sink_java(w_recognizer):-
    format('        new Transition(0,true,false,0,0,0),~n',
         []).

format_sink_java(w_transducer(_)):-
    format('        new Transition(0,true,false,0,"","",0,0),~n',
         []).


format_start(Lang,Type,_Label,Start,_Ini,FinalOutput,FinalTable,NextTable):-
    final_mark(Lang,Start,FinalTable,Final),
    next_mark(Start,NextTable,Next),
    final_out_mark(Type,Start,FinalOutput,FinalOut),
    format_start_string(Lang,Type,Final,Next,FinalOut).

format_start_string(java,Type,Final,Next,FinalOut):-
    format("        /* start state                                 */~n",
	   []),
    format_start_string_java(Type,Final,Next,FinalOut).
format_start_string(c,Type,Final,Next,FinalOut):-
    format("                     /* start state                        */~n",
	   []),
    format_start_string_c(Type,Final,Next,FinalOut).
format_start_string(cpp,Type,Final,Next,FinalOut):-
    format("                     // start state~n",
	   []),
    format_start_string_cpp(Type,Final,Next,FinalOut).

format_start_string_c(recognizer,Final,Next,_) :-
    format('    {0,0,~w,~w},~n',[Final,Next]).
format_start_string_c(transducer(_),Final,Next,FinalOut) :-
    format('    {0,0,~w,~w,"",~s},~n',
	   [Final,Next,FinalOut]).
format_start_string_c(w_recognizer,Final,Next,FinalOut) :-
    format('    {0,0,~w,~w,0,~s},~n',
	   [Final,Next,FinalOut]).
format_start_string_c(w_transducer(_),Final,Next,FinalOut/Wout) :-
    format('    {0,0,~w,~w,"",~s,0,~s},~n',
	   [Final,Next,FinalOut,Wout]).

format_start_string_cpp(recognizer,Final,Next,_) :-
    format('    {\'\\0\',0,~w,~w},~n',[Final,Next]).
format_start_string_cpp(transducer(_),Final,Next,FinalOut) :-
    format('    {\'\\0\',0,~w,~w,"",~s},~n',
	   [Final,Next,FinalOut]).
format_start_string_cpp(w_recognizer,Final,Next,FinalOut) :-
    format('    {\'\\0\',0,~w,~w,0,~s},~n',
	   [Final,Next,FinalOut]).
format_start_string_cpp(w_transducer(_),Final,Next,FinalOut/Wout) :-
    format('    {\'\\0\',0,~w,~w,"",~s,0,~s},~n',
	   [Final,Next,FinalOut,Wout]).

format_start_string_java(recognizer,Final,Next,_) :-
    format('        new Transition(0,false,~w,~w),~n',
	   [Final,Next]).
format_start_string_java(transducer(_),Final,Next,FinalOut):-
    format('        new Transition(0,false,~w,~w,"",~s),~n',
	   [Final,Next,FinalOut]).
format_start_string_java(w_recognizer,Final,Next,FinalOut):-
    format('        new Transition(0,false,~w,~w,0,~s),~n',
	   [Final,Next,FinalOut]).
format_start_string_java(w_transducer(_),Final,Next,FinalOut/Wout):-
    format('        new Transition(0,false,~w,~w,"",~s,0,~s),~n',
	   [Final,Next,FinalOut,Wout]).

final_out_mark(Type,State,Outs,Mark) :-
    (	Outs==[]
    ->  nil_out_mark(Type,Mark)
    ;	fsa_array:fsa_array_get(State,Mark0,Outs)
    ->	(   Mark0=Mark1/Mark2
	->  Mark=MarkL/MarkR,
	    mark_it(Mark1,MarkL),
	    mark_it(Mark2,MarkR)
	;   mark_it(Mark0,Mark)
	)
    ;	nil_out_mark(Type,Mark)
    ).

mark_it(Mark0,Mark) :-
    (	integer(Mark0)
    ->	integer_symbol(Mark0,Mark)
    ;	string_symbol(Mark0,Mark)
    ).

final_mark(c,State,Table,Mark) :-
    fsa_array:fsa_array_access(State,V,Table),
    (	nonvar(V)
    ->	Mark=1
    ;   Mark=0
    ).

final_mark(cpp,State,Table,Mark) :-
    fsa_array:fsa_array_access(State,V,Table),
    (	nonvar(V)
    ->	Mark=1
    ;   Mark=0
    ).

final_mark(java,State,Table,Mark) :-
    fsa_array:fsa_array_access(State,V,Table),
    (	nonvar(V)
    ->	Mark=true
    ;   Mark=false
    ).

next_mark(State,Table,Mark):-
    fsa_array:fsa_array_access(State,Mark,0,Table).

format_transitions([],_,_,_,_,_,_).
format_transitions([H|T],Lang,Type,Outs,Finals,Nexts,Fa) :-
    format_transitions1(T,H,Lang,Type,Outs,Finals,Nexts,Fa).

format_transitions1([],trans(_,Sym,Q),Lang,Type,Outs,Finals,Nexts,Fa) :-
    last_mark(Lang,1,Last),
    format_transition(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa).

format_transitions1([trans(P1,Sym1,Q1)|T],trans(P,Sym,Q),Lang,Type,Outs,
		    Finals,Nexts,Fa):-
    (	P1 =:= P
    ->	last_mark(Lang,0,Last)
    ;   last_mark(Lang,1,Last)
    ),
    format_transition(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa),
    format_transitions1(T,trans(P1,Sym1,Q1),Lang,Type,Outs,Finals,Nexts,Fa).

last_mark(c,I,I).
last_mark(cpp,I,I).
last_mark(java,I,Bool) :-
    last_mark_java(I,Bool).

last_mark_java(0,false).
last_mark_java(1,true).

format_transition(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa):-
    \+ \+ format_transition0(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa).

format_transition0(Lang,Type,Last,Sym,Q,Outs,Finals,Nexts,Fa) :-
    symbol_mark(Type,Fa,Sym,SymIn,SymOut),
    final_mark(Lang,Q,Finals,Final),
    final_out_mark(Type,Q,Outs,FinalOut),
    next_mark(Q,Nexts,Next),
    format_transition_string(Lang,Type,SymIn,SymOut,Last,Final,Next,FinalOut).

format_transition_string(c,Type,SymIn,SymOut,Last,Final,Next,FinalOut) :-
    format_transition_string_c(Type,SymIn,SymOut,Last,Final,Next,FinalOut).
format_transition_string(cpp,Type,SymIn,SymOut,Last,Final,Next,FinalOut) :-
    format_transition_string_cpp(Type,SymIn,SymOut,Last,Final,Next,FinalOut).
format_transition_string(java,Type,SymIn,SymOut,Last,Final,Next,FinalOut) :-
    format_transition_string_java(Type,SymIn,SymOut,Last,Final,Next,FinalOut).

format_transition_string_java(recognizer,SymIn,_SymOut,Last,Final,Next,_):-
    format("        new Transition(~w,~w,~w,~w),~n",
	   [SymIn,Last,Final,Next]).
format_transition_string_java(transducer(_),SymIn,SymOut,Last,Final,Next,Out):-
    format("        new Transition(~w,~w,~w,~w,~s,~s),~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_java(w_recognizer,SymIn,SymOut,Last,Final,Next,Out):-
    format("        new Transition(~w,~w,~w,~w,~s,~s),~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_java(w_transducer(_),SymIn,SymOut/W0,Last,Final,Next,
			      Out/W):-
    format("        new Transition(~w,~w,~w,~w,~s,~s,~s,~s),~n",
	   [SymIn,Last,Final,Next,SymOut,Out,W0,W]).

format_transition_string_c(recognizer,SymIn,_SymOut,Last,Final,Next,_):-
    format("    {~w,~w,~w,~w},~n",
	   [SymIn,Last,Final,Next]).
format_transition_string_c(transducer(_),SymIn,SymOut,Last,Final,Next,Out):-
    format("    {~w,~w,~w,~w,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_c(w_recognizer,SymIn,SymOut,Last,Final,Next,Out):-
    format("    {~w,~w,~w,~w,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_c(w_transducer(_),SymIn,SymOut/W0,Last,Final,Next,Out/W):-
    format("    {~w,~w,~w,~w,~s,~s,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out,W0,W]).

format_transition_string_cpp(recognizer,SymIn,_SymOut,Last,Final,Next,_):-
    format("    {~w,~w,~w,~w},~n",
	   [SymIn,Last,Final,Next]).
format_transition_string_cpp(transducer(_),SymIn,SymOut,Last,Final,Next,Out):-
    format("    {~w,~w,~w,~w,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_cpp(w_recognizer,SymIn,SymOut,Last,Final,Next,Out):-
    format("    {~w,~w,~w,~w,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out]).
format_transition_string_cpp(w_transducer(_),SymIn,SymOut/W0,Last,Final,Next,Out/W):-
    format("    {~w,~w,~w,~w,~s,~s,~s,~s},~n",
	   [SymIn,Last,Final,Next,SymOut,Out,W0,W]).

format_transitions_end(c) :-
    format("    };~n",[]).
format_transitions_end(cpp) :-
    format("    };~n",[]).
format_transitions_end(java) :-
    format("    };~n",[]).

format_more(java,_,_,_,_,_,_).
format_more(c,Type,_Label,Ini,_FinalOutput,_FinalTable,_NextTable) :-
    format_ini_out(Type,Ini),
    format_procedure(Type).

format_more(cpp,Type,_Label,Ini,_FinalOutput,_FinalTable,_NextTable) :-
    format_ini_out_cpp(Type,Ini),
    format_procedure_cpp(Type).

format_ini_out(recognizer,_).
format_ini_out(transducer(_),String):-
    format('  \n const char *ini_out=~s;~n',[String]).
format_ini_out(w_recognizer,String):-
    format('  \n const unsigned int ini_out=~s;~n',[String]).
format_ini_out(w_transducer(_),String/W):-
    format('  \n const char *ini_out=~s;~n  const unsigned int ini_wout=~s;~n',[String,W]).

format_ini_out_cpp(recognizer,_).

format_ini_out_cpp(transducer(_),String):-
    format('  \n const string ini_out(~s);',[String]).

format_ini_out_cpp(w_recognizer,W):-
    format('  \n const unsigned int ini_wout(~s);',[W]).

format_ini_out_cpp(w_transducer(_),String/W):-
    format('  \n const string ini_out(~s); \
 \n   const unsigned int ini_wout(~s);',[String,W]).


format_procedure(recognizer) :-
    unknown_symbol(U),
    atom_codes(U,[Code]),
    format("  register unsigned char c;\
 \n   register unsigned int i;\
 \n   register unsigned long int line=1;\
 \n   i=0;\
 \n   c=in[i];\
 \n   while (c != '\\0') {\
 \n     line=trans[line].next;\
 \n     if (trans[line].symbol == ~w && unknown_symbol(c) )\
 \n       ;\
 \n     else {\
 \n       while (!trans[line].last && c > trans[line].symbol) {  \
 \n         line++;\
 \n       }\
 \n       if (c != trans[line].symbol)\
 \n     return 0;\
 \n     }\
 \n     i++;\
 \n     c=in[i];\
 \n   }\
 \n   return trans[line].final;        /* ok if at final state */\
 \n }\
 \n",[Code]).

format_procedure(transducer(_)) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    queue_symbol(Q),
    atom_codes(Q,[Qcode]),
    format("  register unsigned char c;\
 \n   register unsigned int i;\
 \n   register unsigned long int line=1;\
 \n   struct queue q;\
 \n   int outl;\
 \n   qinit(&q);\
 \n   i=0;\
 \n   c=in[i];\
 \n   out[0]='\\0';\
 \n   outl=0;\
 \n   outl+=strlen(ini_out);\
 \n   if (outl > max) return 2;\
 \n   strcat(out,ini_out);\
 \n   while (c != '\\0') {\
 \n     line=trans[line].next;\
 \n     if (trans[line].symbol == ~w && unknown_symbol(c) )\
 \n       ;\
 \n     else {\
 \n       if (trans[line].symbol == ~w && unknown_symbol(c) ) {\
 \n     qinsert(&q,c);\
 \n       }\
 \n       else {\
 \n         while (!trans[line].last && c > trans[line].symbol) {  \
 \n           line++;\
 \n         }\
 \n",[Ucode,Qcode]),
    format("\
 \n       if (c != trans[line].symbol)\
 \n         return 0;\
 \n       }\
 \n     }\
 \n     if (outl+strlen(trans[line].out) > max) return 2;\
 \n     strcat(&out[outl],trans[line].out);\
 \n     replace_from_queue(&q,&out[outl]);\
 \n     outl+=strlen(trans[line].out);\
 \n     i++;\
 \n     c=in[i];\
 \n   }\
 \n   if (outl+strlen(trans[line].final_out) > max) return 2;\
 \n   strcat(&out[outl],trans[line].final_out);\
 \n   replace_from_queue(&q,&out[outl]);\
 \n   return trans[line].final;        /* ok if at final state */\
 \n } \
 \n",[]).

format_procedure(w_recognizer) :-
    unknown_symbol(U),
    atom_codes(U,[Code]),
    format("  register unsigned char c;\
 \n   register unsigned int i;\
 \n   register unsigned long int line=1;\
 \n   i=0;\
 \n   c=in[i];\
 \n   *out=0;\
 \n   *out+=ini_out;\
 \n   while (c != '\\0') {\
 \n     line=trans[line].next;\
 \n     if (trans[line].symbol == ~w && unknown_symbol(c) )\
 \n       ;\
 \n     else {\
 \n       while (!trans[line].last && c > trans[line].symbol) {  \
 \n         line++;\
 \n     }\
 \n",[Code]),
    format(
"      if (c != trans[line].symbol)\
 \n          return 0;\
 \n     }\
 \n     *out+=trans[line].out;\
 \n     i++;\
 \n     c=in[i];\
 \n   }\
 \n     *out+=trans[line].final_out;\
 \n     return trans[line].final;\
 \n     \
 \n }\
 \n",[]).


format_procedure(w_transducer(_)) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    queue_symbol(Q),
    atom_codes(Q,[Qcode]),
    format("  register unsigned char c;\
 \n   register unsigned int i;\
 \n   register unsigned long int line=1;\
 \n   struct queue q;\
 \n   int outl;\
 \n   qinit(&q);\
 \n   i=0;\
 \n   c=in[i];\
 \n   *wout=0;\
 \n   *wout+=ini_wout;\
 \n   out[0]='\\0';\
 \n   outl=0;\
 \n   outl+=strlen(ini_out);\
 \n   outl+=strlen(ini_out);\
 \n   if (outl > max) return 2;\
 \n   strcat(out,ini_out);\
 \n   while (c != '\\0') {\
 \n     line=trans[line].next;\
 \n     if (trans[line].symbol == ~w && unknown_symbol(c) )\
 \n       ;\
 \n     else {\
 \n       if (trans[line].symbol == ~w && unknown_symbol(c) ) {\
 \n     qinsert(&q,c);\
 \n       }\
 \n       else {\
 \n         while (!trans[line].last && c > trans[line].symbol) {  \
 \n           line++;\
 \n         }\
 \n",[Ucode,Qcode]),
    format(
"      if (c != trans[line].symbol)\
 \n          return 0;\
 \n        }\
 \n     }\
 \n     *wout+=trans[line].wout;\
 \n     if (outl+strlen(trans[line].out) > max) return 2;\
 \n     strcat(&out[outl],trans[line].out);\
 \n     replace_from_queue(&q,&out[outl]);\
 \n     outl+=strlen(trans[line].out);\
 \n     i++;\
 \n     c=in[i];\
 \n   }\
 \n   *wout+=trans[line].final_wout;\
 \n   if (outl+strlen(trans[line].final_out) > max) return 2;\
 \n   strcat(&out[outl],trans[line].final_out);\
 \n   replace_from_queue(&q,&out[outl]);\
 \n   return trans[line].final;\
 \n     \
 \n }\
 \n",[]).

format_procedure_cpp(recognizer) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    format("\
 \n   size_t line = 1; \
 \n  \
 \n   for (;input_iter != end; ++input_iter) { \
 \n     line=trans[line].next; \
 \n     if (trans[line].symbol == ~w && unknown_symbol(*input_iter) )\
 \n       ;\
 \n     else {\
 \n       while (!trans[line].last && (unsigned char)*input_iter > trans[line].symbol) {   \
 \n         line++; \
 \n       } \
 \n       if ((unsigned char)*input_iter != trans[line].symbol) line=0;          // sink  \
 \n     }\
 \n   } \
 \n   return trans[line].final; \
 \n }\
 \n",[Ucode]).

format_procedure_cpp(transducer(_)) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    queue_symbol(Q),
    atom_codes(Q,[Qcode]),
    format("\
 \n   out = ini_out; \
 \n  \
 \n   size_t line = 1; \
 \n \
 \n   queue< char > q;\
 \n   \
 \n   for (;input_iter != end; ++input_iter) { \
 \n     line=trans[line].next; \
 \n     if (trans[line].symbol == ~w && unknown_symbol(*input_iter) )\
 \n       ;\
 \n     else {\
 \n       if (trans[line].symbol == ~w && unknown_symbol(*input_iter) ) {\
 \n     q.push(*input_iter);\
 \n       }\
 \n       else {\
 \n         while (!trans[line].last && (unsigned char)*input_iter > trans[line].symbol) {   \
 \n           line++; \
 \n         } \
 \n         if ((unsigned char)*input_iter != trans[line].symbol) line=0;          // sink\
 \n       }\
 \n     }\
 \n     out += replace_from_queue(q,trans[line].out);\
 \n   } \
 \n   out += replace_from_queue(q,trans[line].final_out); \
 \n   return trans[line].final; \
 \n}\
 \n",[Ucode,Qcode]).

format_procedure_cpp(w_recognizer) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    format("\
 \n   wout = ini_wout; \
 \n  \
 \n   size_t line = 1; \
 \n   for (;input_iter != end; ++input_iter) { \
 \n     line=trans[line].next; \
 \n     if (trans[line].symbol == ~w && unknown_symbol(*input_iter) )\
 \n       ;\
 \n     else {\
 \n       while (!trans[line].last && (unsigned char)*input_iter > trans[line].symbol) {   \
 \n         line++; \
 \n       } \
 \n       if ((unsigned char)*input_iter != trans[line].symbol) line=0;          // sink  \
 \n     }\
 \n     wout += trans[line].wout; \
 \n   } \
 \n   wout += trans[line].final_wout; \
 \n   return trans[line].final; \
 \n }\
 \n",[Ucode]).

format_procedure_cpp(w_transducer(_)) :-
    unknown_symbol(U),
    atom_codes(U,[Ucode]),
    queue_symbol(Q),
    atom_codes(Q,[Qcode]),
    format("\
 \n   wout = ini_wout; \
 \n   out = ini_out; \
 \n  \
 \n   size_t line = 1; \
 \n  \
 \n   queue< char > q;\
 \n   \
 \n   for (;input_iter != end; ++input_iter) { \
 \n     line=trans[line].next; \
 \n     if (trans[line].symbol == ~w && unknown_symbol(*input_iter) )\
 \n       ;\
 \n     else {\
 \n       if (trans[line].symbol == ~w && unknown_symbol(*input_iter) ) {\
 \n     q.push(*input_iter);\
 \n       }\
 \n       else {\
 \n         while (!trans[line].last && (unsigned char)*input_iter > trans[line].symbol) {   \
 \n           line++; \
 \n         } \
 \n         if ((unsigned char)*input_iter != trans[line].symbol) line=0;          // sink  \
 \n       }\
 \n     }\
 \n     wout += trans[line].wout; \
 \n     out += replace_from_queue(q,trans[line].out); \
 \n   } \
 \n   wout += trans[line].final_wout; \
 \n   out += replace_from_queue(q,trans[line].final_out); \
 \n   return trans[line].final; \
 \n }\
 \n",[Ucode,Qcode]).

format_main(off,_Lang,_Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable).
format_main(on,Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable) :-
    format_main(Lang,Type,Label,Ini,FinalOutput,FinalTable,NextTable).

format_main(c,Type,_,_,_,_,_):-
    format_main_c(Type).

format_main(cpp,Type,_,_,_,_,_):-
    format_main_cpp(Type).

format_main(java,_,Label,_,_,_,_):-
    format("\
 \n     public static void main(String argv[]) {\
 \n     DFA.main(argv, new ~w());\
 \n     }",[Label]).

format_main_c(recognizer) :-
    format("\
 \n \
 \n \
 \n \
 \n #include <stdio.h>\
 \n #define MAXLINE 131072        /* this constrains line length */\
 \n int main() {\
 \n   char in[MAXLINE];\
 \n \
 \n   while(fgets(in,MAXLINE,stdin) != NULL) {\
 \n     in[strlen(in)-1] = '\\0';                       /* chop */\
 \n     if (accepts(in)) \
 \n       fprintf(stderr,\"yes\\n\");\
 \n     else\
 \n       fprintf(stderr,\"no\\n\");\
 \n   }                                         /* end-of-file */\
 \n   return 0;\
 \n } \
 \n \n",[]).

format_main_c(transducer(_)) :-
    format("\
 \n \
 \n \
 \n \
 \n #include <stdio.h>\
 \n #define MAXLINE 131072        /* this constrains line length */\
 \n int main() {\
 \n   char in[MAXLINE];\
 \n   char out[MAXLINE];\
 \n   int return_value;\
 \n \
 \n   while(fgets(in,MAXLINE,stdin) != NULL) {\
 \n     in[strlen(in)-1] = '\\0';                       /* chop */\
 \n     return_value = t_accepts(in,out,MAXLINE);\
 \n     switch(return_value) {\
 \n     case 0:\
 \n       fprintf(stderr,\"no\\n\");\
 \n       break;\
 \n     case 1:\
 \n       printf(\"%s\\n\",out);\
 \n       break;\
 \n     case 2:\
 \n       fprintf(stderr,\"error: output line too long\\n\");\
 \n       break;\
 \n     }\
 \n   }                                         /* end-of-file */\
 \n   return 0;\
 \n } \
 \n \n",[]).

format_main_c(w_recognizer) :-
    format("\
 \n \
 \n \
 \n \
 \n #include <stdio.h>\
 \n #define MAXLINE 131072        /* this constrains line length */\
 \n int main() {\
 \n   char in[MAXLINE];\
 \n   unsigned int out;\
 \n \
 \n   while(fgets(in,MAXLINE,stdin) != NULL) {\
 \n     in[strlen(in)-1] = '\\0';                       /* chop */\
 \n     if (w_accepts(in,&out)) \
 \n       printf(\"%d\\n\",out);\
 \n     else\
 \n       fprintf(stderr,\"no\\n\");\
 \n   }                                         /* end-of-file */\
 \n   return 0;\
 \n } \
 \n \n",[]).


format_main_c(w_transducer(_)) :-
    format("\
 \n \
 \n \
 \n \
 \n #include <stdio.h>\
 \n #define MAXLINE 131072        /* this constrains line length */\
 \n int main() {\
 \n   char in[MAXLINE];\
 \n   char out[MAXLINE];\
 \n   unsigned int wout;\
 \n   int return_value;\
 \n \
 \n   while(fgets(in,MAXLINE,stdin) != NULL) {\
 \n     in[strlen(in)-1] = '\\0';                       /* chop */\
 \n \
 \n     return_value = wt_accepts(in,out,&wout,MAXLINE);\
 \n     switch(return_value) {\
 \n     case 0:\
 \n       fprintf(stderr,\"no\\n\");\
 \n       break;\
 \n     case 1:\
 \n       printf(\"%d:%s\\n\",wout,out);\
 \n       break;\
 \n     case 2:\
 \n       fprintf(stderr,\"error: output line too long\\n\");\
 \n       break;\
 \n     }\
 \n   }                                         /* end-of-file */\
 \n   return 0;\
 \n } \
 \n \n",[]).

					 
format_main_cpp(recognizer) :-
    format("\
 \n int main() { \
 \n  \
 \n   string in; \
 \n   while(getline(cin,in)) { \
 \n     if (accepts(in.begin(),in.end()))  \
 \n       cerr << \"yes\" << endl;\
 \n     else \
 \n       cerr << \"no\" << endl; \
 \n   } \
 \n } \
 \n",[]).


format_main_cpp(transducer(_)) :-
    format("\
 \n int main() { \
 \n  \
 \n   string in, out; \
 \n   while(getline(cin,in)) { \
 \n     if (t_accepts(in.begin(),in.end(),out))  \
 \n       cout << out << endl; \
 \n     else \
 \n       cerr << \"no\" << endl; \
 \n   } \
 \n } \
 \n",[]).


format_main_cpp(w_recognizer) :-
    format("\
 \n int main() { \
 \n  \
 \n   unsigned int wout; \
 \n   string in; \
 \n   while(getline(cin,in)) { \
 \n     if (w_accepts(in.begin(),in.end(),wout))  \
 \n       cout << wout << endl; \
 \n     else \
 \n       cerr << \"no\" << endl; \
 \n   } \
 \n } \
 \n",[]).


format_main_cpp(w_transducer(_)) :-
    format("\
 \n int main() { \
 \n  \
 \n   unsigned int wout; \
 \n   string in, out; \
 \n   while(getline(cin,in)) { \
 \n     if (wt_accepts(in.begin(),in.end(),out,wout))  \
 \n       cout << wout << ':' << out << endl; \
 \n     else \
 \n       cerr << \"no\" << endl; \
 \n   } \
 \n } \
 \n",[]).


format_tail(c,_,_,_,_,_,_).
format_tail(cpp,_,_,_,_,_,_).
format_tail(java,_Type,_Label,_Ini,_FinalOutput,_FinalTable,_NextTable) :-
    format("~n}~n~n",[]).

nil_out_mark(recognizer,[]).
nil_out_mark(transducer(_),[34,34]).    % ""
nil_out_mark(w_recognizer,[48]).     % 0
nil_out_mark(w_transducer(_),[34,34]/[48]).     

format_unknown_symbol(Lang,Syms) :-
    format_unknown_symbol_start(Lang),
    format_unknown_symbol_list(Syms,Lang),
    format_unknown_symbol_end(Lang).

format_unknown_symbol_start(c) :-
    format("\
 \n \n \n int unknown_symbol(char c) {\
 \n   return ( 1~n",[]).

format_unknown_symbol_start(cpp) :-
    format("\
 \n \n \n bool unknown_symbol(char c) {\
 \n   return ( 1~n",[]).

format_unknown_symbol_start(java) :-
    format(' \n    private static String unknown_string="',[]). %"

format_unknown_symbol_end(cpp) :-
    format("       );  ~n ~n }",[]).
format_unknown_symbol_end(java) :-
     format('";~n',[]).              %"
format_unknown_symbol_end(c) :-
    format("       );  ~n ~n }",[]).   


format_unknown_symbol_list([],_).
format_unknown_symbol_list([H|T],Lang) :-
    (	atom(H)
    ->	atom_codes(H,[N])
    ;	number(H)
    ->	number_codes(H,[N])
    ;	fail
    ),
    format_unknown_symbol_(Lang,N,H),
    format_unknown_symbol_list(T,Lang).

format_unknown_symbol_(cpp,N,_) :-
    format("        && c != (char)~w~n",[N]).
format_unknown_symbol_(java,_,H) :-
    format("~w",[H]).
format_unknown_symbol_(c,N,_) :-
    format("        && c != (char)~w~n",[N]).

%% returns the *input* symbols of Fa
symbols_talked_about(Fa,Syms) :-
    fsa_data:fsa_transitions(Fa,Trans),
    fsa_data:fsa_type(Fa,Type),
    fsa_data:fsa_symbols_decl(Fa,Decl),
    transition_symbols_talked_about(Trans,Type,Decl,Syms0),
    sort(Syms0,Syms).    

transition_symbols_talked_about([],_,_,[]).
transition_symbols_talked_about([trans(_,S,_)|Tr],Type,Decl,Syms):-
    transition_symbol_talked_about(Type,Decl,S,Syms,Syms1),
    transition_symbols_talked_about(Tr,Type,Decl,Syms1).

transition_symbol_talked_about(recognizer,r(Mod),S,Syms0,Syms) :-
    pred_talked_about(Mod,S,Syms0,Syms).
transition_symbol_talked_about(w_recognizer,r(Mod,_),S/_,Syms0,Syms) :-
    (	S == []
    ->	Syms0=Syms
    ;   pred_talked_about(Mod,S,Syms0,Syms)
    ).
transition_symbol_talked_about(transducer(Sub),t(Ma,Mb),Sa/Sb,Syms0,Syms):-
    transition_symbol_talked_about(Sub,Ma,Mb,Sa,Sb,Syms0,Syms).
transition_symbol_talked_about(w_transducer(Sub),t(Ma,Mb,_),Sa/(Sb/_),Syms0,
			       Syms):-
    transition_symbol_talked_about(Sub,Ma,Mb,Sa,Sb,Syms0,Syms).

transition_symbol_talked_about(_,Ma,_Mb,Sa,_Sb,Syms0,Syms):-
    (	Sa==[]
    ->	Syms0=Syms
    ;	Sa='$@'(Sa1)
    ->	pred_talked_about(Ma,Sa1,Syms0,Syms)
    ;   pred_talked_about(Ma,Sa,Syms0,Syms)
    ).

pred_talked_about(M,H,S0,S) :-
    findall(Sym,fsa_pred_module:talks_about(M,H,Sym),S0,S).


format_queue(recognizer,_).
format_queue(w_recognizer,_).
format_queue(transducer(_),Lang) :-
    format_queue_(Lang).
format_queue(w_transducer(_),Lang) :-
    format_queue_(Lang).

format_queue_(java).
format_queue_(cpp) :-
    queue_symbol(Q),
    atom_codes(Q,[Code]),
    format("\
 \n string replace_from_queue(queue< char >& q,string in)\
 \n {\
 \n   int i=0;\
 \n   char c;\
 \n   c=in[i];\
 \n \
 \n   while (c != '\\0') {\
 \n     if (c == ~w) {\
 \n       in[i]=q.front();\
 \n       q.pop();\
 \n     }\
 \n     i++;\
 \n     c=in[i];\
 \n   }\
 \n   return in;\
 \n }\
 \n       ",[Code]).


format_queue_(c) :-
    format("\
 \n \
 \n /*\
 \n  * Queue ADT\
 \n  *\
 \n  * Taken from http://purists.org/queue/\
 \n  * Institute of Applied Iconoclasm\
 \n  *\
 \n  * GvN: added explicit casts for malloc() to prevent warnings\
 \n  */\
 \n \
 \n #define QDATUM char\
 \n       \
 \n /*\
 \n  *  Queue item structure\
 \n  */\
 \n struct qentry\
 \n {\
 \n         struct qentry *next;\
 \n         QDATUM d;\
 \n };",[]),
format("\
 \n /*\
 \n  *  Queue container structure\
 \n  *\
 \n  *  This structure encapsulates a linked list of qentry items.\
 \n  */\
 \n struct queue\
 \n {\
 \n         struct qentry *begin, **end;\
 \n };",[]),
format("\
 \n /*\
 \n  *  qinit: Initialize queue.\
 \n  *\
 \n  *  Parameters:\
 \n  *\
 \n  *    q         Pointer to a queue, or NULL if the user wishes to leave \
 \n  *              it to qinit to allocate the queue.\
 \n  *\
 \n  *  Return values:\
 \n  *\
 \n  *    non-NULL  Queue has been initialized.\
 \n  *\
 \n  *    NULL      Insufficient memory.\
 \n  */\
 \n struct queue *\
 \n qinit(struct queue *q)\
 \n {\
 \n         if (q || (q = (struct queue *) malloc(sizeof(struct queue))) != NULL) {\
 \n                 q->begin = NULL;\
 \n                 q->end = &q->begin;\
 \n         }\
 \n         return q;\
 \n }",[]),
format("\
 \n /*\
 \n  *  qinsert: append an item to the queue.\
 \n  *\
 \n  *  Parameters:\
 \n  *\
 \n  *    q         Pointer to a queue. It is assumed the queue has been\
 \n  *              initialized by a call to qinit.\
 \n  *\
 \n  *    d         Item to be appended.\
 \n  *\
 \n  *  Return values:      \
 \n  *\
 \n  *    1         The item has been appended.\
 \n  *\
 \n  *    0         The item could not be appended. Either the queue \
 \n  *              pointer provided was NULL, or the function was unable \
 \n  *              to allocate the amount of memory needed for a new \
 \n  *              queue item.     \
 \n  */\
 \n int\
 \n qinsert(struct queue *q, QDATUM d)\
 \n {\
 \n         if (!q || !(*q->end = (struct qentry *) malloc(sizeof(struct qentry)))) return 0;\
 \n         (*q->end)->d = d;\
 \n         (*q->end)->next = NULL;\
 \n         q->end = &((*q->end)->next);\
 \n         return 1;\
 \n }",[]),
format("\
 \n /*\
 \n  *  qremove: remove an item from the queue.\
 \n  *\
 \n  *  Parameters:\
 \n  *\
 \n  *    q         Pointer to a queue.\
 \n  *\
 \n  *    d         Pointer to the QDATUM variable that will hold the datum\
 \n  *              corresponding to the queue item.\
 \n  *\
 \n  *  Return values:\
 \n  *\
 \n  *    non-NULL  An item has been removed. The variable that d points \
 \n  *              to now contains the datum associated with the item \
 \n  *              in question.\
 \n  *              \
 \n  *    NULL      No item could be removed. Either the queue pointer \
 \n  *              provided was NULL, or the queue was empty. The memory \
 \n  *              location that d points to has not been modified. \
 \n  */\
 \n QDATUM *\
 \n qremove(struct queue *q, QDATUM *d)\
 \n {\
 \n         struct qentry *tmp;\
 \n                 \
 \n         if (!q || !q->begin) return NULL;\
 \n         tmp = q->begin;\
 \n         if (!(q->begin = q->begin->next)) q->end = &q->begin;\
 \n         *d = tmp->d;\
 \n         free(tmp);\
 \n         return d;       \
 \n }",[]),
format("\
 \n /*\
 \n  *  qpeek: access an item without removing it from the queue.\
 \n  *\
 \n  *  Parameters:\
 \n  *\
 \n  *    q         Pointer to a queue.\
 \n  *\
 \n  *    d         Pointer to the QDATUM variable that will hold the datum\
 \n  *              associated with the first item in the queue, i. e.,\
 \n  *              the item that would be removed had qremove been called\
 \n  *              instead of qpeek.\
 \n  *\
 \n  *  Return values:\
 \n  * \
 \n  *    See qremove.\
 \n  */\
 \n QDATUM *\
 \n qpeek(struct queue *q, QDATUM *d)\
 \n {\
 \n         if (!q || !q->begin) return NULL;\
 \n         *d = q->begin->d;\
 \n         return d;\
 \n }",[]),
    queue_symbol(Q),
    atom_codes(Q,[Code]),
format("\
 \n void replace_from_queue(struct queue *q,char *in)\
 \n {\
 \n   int i=0;\
 \n   char c;\
 \n   char d;\
 \n   c=in[i];\
 \n   while (c != '\\0') {\
 \n     if (c == ~w) {\
 \n       qremove(q,&d);\
 \n       in[i]=d;\
 \n     }\
 \n     i++;\
 \n     c=in[i];\
 \n   }\
 \n \
 \n }",[Code]).



new_symbols_decl(recognizer,Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,_,r(fsa_frozen)).
new_symbols_decl(transducer(_Sub),Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,_,t(fsa_frozen,fsa_frozen)).
new_symbols_decl(w_recognizer,Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,_,r(fsa_frozen,fsa_frozen)).
new_symbols_decl(w_transducer(_Sub),Fa0,Fa) :-
    fsa_data:fsa_copy_except(symbols,Fa0,Fa,_,t(fsa_frozen,fsa_frozen,fsa_frozen)).

%%% limit input symbols to singleton set predicates of one-letter atoms
%%% for open-ended alphabets, use symbol '\002' 
%%% for identities over open-ended alphabets, use symbol
%%% '\001\' to indicate enqueue/dequeue
n_expand_predicates_fa(recognizer,Fa0,Fa,DefaultSymbols):-
    n_expand_predicates_fa(Fa0,Fa1,DefaultSymbols),
    new_symbols_decl(recognizer,Fa1,Fa).
n_expand_predicates_fa(w_recognizer,Fa0,Fa,DefaultSymbols):-
    n_expand_predicates_fa(Fa0,Fa1,DefaultSymbols),
    new_symbols_decl(w_recognizer,Fa1,Fa).
n_expand_predicates_fa(transducer(Sub),Fa0,Fa,DefaultSymbols):-
    fsa_regex:to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    n_expand_predicates_fa(Fa1,Fa2,DefaultSymbols),
    new_symbols_decl(transducer(Sub),Fa2,Fa).
n_expand_predicates_fa(w_transducer(Sub),Fa0,Fa,DefaultSymbols):-
    fsa_regex:w_to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    n_expand_predicates_fa(Fa1,Fa2,DefaultSymbols),
    new_symbols_decl(w_transducer(Sub),Fa2,Fa).

n_expand_predicates_fa(Fa0,Fa,DefaultSymbols) :-
    fsa_data:fsa_symbols_decl(Fa0,Mods),
    fsa_data:fsa_copy_except(transitions,Fa0,Fa1,Trans0,Trans),
    fsa_data:fsa_copy_except(states,Fa1,Fa,Nr0,Nr),
    n_expand_transitions(Trans0,Trans1,Mods,DefaultSymbols,Nr0,Nr),
    sort(Trans1,Trans).

n_expand_transitions([],[],_,_Ds,Nr,Nr).
n_expand_transitions([H|T],Tr,Mods,Ds,Nr0,Nr):-
    n_expand_transition(Mods,H,Tr,Tr1,Ds,Nr0,Nr1),
    n_expand_transitions(T,Tr1,Mods,Ds,Nr1,Nr).

n_expand_transition(r(Mod),trans(P,Pred,Q),Tr0,Tr,Ds,Nr0,Nr):-
    findall(trans(P,Sym,Q),n_eval_pred(Mod,Pred,Sym,Ds),Tr1),
    expand_multi(Tr1,Tr0,Tr,Nr0,Nr,r(Mod)).

n_expand_transition(r(Mod,F),trans(P,Pred/W,Q),Tr0,Tr,Ds,Nr0,Nr):-
    findall(trans(P,Sym/W,Q),n_eval_pred(Mod,Pred,Sym,Ds),Tr1),
    expand_multi(Tr1,Tr0,Tr,Nr0,Nr,r(Mod,F)).

n_expand_transition(t(ModD,ModR),trans(P,PredD/PredR,Q),Tr0,Tr,Ds,Nr0,Nr):-
    findall(trans(P,SymD/SymR,Q),
	    n_eval_pred_pair(ModD,ModR,PredD,PredR,SymD,SymR,Ds),
	    Tr1),
    expand_multi(Tr1,Tr0,Tr,Nr0,Nr,t(ModD,ModR)).

n_expand_transition(t(ModD,ModR,F),trans(P,PredD/(PredR/W),Q),Tr0,Tr,Ds,Nr0,Nr):-
    findall(trans(P,SymD/(SymR/W),Q),
	    n_eval_pred_pair(ModD,ModR,PredD,PredR,SymD,SymR,Ds),
	    Tr1),
    expand_multi(Tr1,Tr0,Tr,Nr0,Nr,t(ModD,ModR,F)).

expand_multi([],Tr,Tr,Nr,Nr,_).
expand_multi([H|T],Tr0,Tr,Nr0,Nr,Mod) :-
    expand_multi_(Mod,H,Tr0,Tr1,Nr0,Nr1),
    expand_multi(T,Tr1,Tr,Nr1,Nr,Mod).

expand_multi_(r(_),trans(P,Sym,Q),Trs0,Trs,Nr0,Nr) :-
    (	Sym==[]
    ->	Nr0=Nr,
	Trs0=[trans(P,Sym,Q)|Trs]
    ;	atom_to_atoms(Sym,Syms,[]),
	exp_transitions_r(Syms,P,Q,Trs0,Trs,Nr0,Nr)
    ).

expand_multi_(r(_,_),trans(P,Sym/W,Q),Trs0,Trs,Nr0,Nr) :-
    (	Sym==[]
    ->	Nr0=Nr,
	Trs0=[trans(P,Sym/W,Q)|Trs]
    ;	atom_to_atoms(Sym,Syms,[]),
	exp_transitions_wr(Syms,W,P,Q,Trs0,Trs,Nr0,Nr)
    ).

expand_multi_(t(_,_),trans(P,Sym/Out,Q),Trs0,Trs,Nr0,Nr) :-
    (	Sym == []
    ->	Nr0=Nr,
	Trs0=[trans(P,Sym/Out,Q)|Trs]
    ;	atom_to_atoms(Sym,Syms,[]),
	exp_transitions_t(Syms,Out,P,Q,Trs0,Trs,Nr0,Nr)
    ).

expand_multi_(t(_,_,_),trans(P,Sym/W,Q),Trs0,Trs,Nr0,Nr) :-
    (	Sym == []
    ->	Nr0=Nr,
	Trs0=[trans(P,Sym/W,Q)|Trs]
    ;	atom_to_atoms(Sym,Syms,[]),
	exp_transitions_wt(Syms,W,P,Q,Trs0,Trs,Nr0,Nr)
    ).

exp_transitions_r([H|T],P,Q,[trans(P,H,Q1)|Trs0],Trs,Nr0,Nr):-
    exp_transitions_r_(T,Q1,Q,Trs0,Trs,Nr0,Nr).

exp_transitions_r_([],Q,Q,Trs,Trs,Nr,Nr).
exp_transitions_r_([H|T],Q0,Q,[trans(Q0,H,Q1)|Trs0],Trs,Q0,Nr) :-
    Nr1 is Q0+1,
    exp_transitions_r_(T,Q1,Q,Trs0,Trs,Nr1,Nr).

exp_transitions_wr([H|T],W,P,Q,[trans(P,H/W,Q1)|Trs0],Trs,Nr0,Nr):-
    exp_transitions_wr_(T,Q1,Q,Trs0,Trs,Nr0,Nr).

exp_transitions_wr_([],Q,Q,Trs,Trs,Nr,Nr).
exp_transitions_wr_([H|T],Q0,Q,[trans(Q0,H/Zero,Q1)|Trs0],Trs,Q0,Nr) :-
    fsa_semiring:zero(fsa_weights,Zero),
    Nr1 is Q0+1,
    exp_transitions_wr_(T,Q1,Q,Trs0,Trs,Nr1,Nr).

exp_transitions_t([H|T],Out,P,Q,[trans(P,H/Out,Q1)|Trs0],Trs,Nr0,Nr):-
    exp_transitions_t_(T,Q1,Q,Trs0,Trs,Nr0,Nr).

exp_transitions_t_([],Q,Q,Trs,Trs,Nr,Nr).
exp_transitions_t_([H|T],Q0,Q,[trans(Q0,H/Zero,Q1)|Trs0],Trs,Q0,Nr) :-
    fsa_semiring:zero(fsa_strings,Zero),
    Nr1 is Q0+1,
    exp_transitions_t_(T,Q1,Q,Trs0,Trs,Nr1,Nr).

exp_transitions_wt([H|T],W,P,Q,[trans(P,H/W,Q1)|Trs0],Trs,Nr0,Nr):-
    exp_transitions_wt_(T,Q1,Q,Trs0,Trs,Nr0,Nr).

exp_transitions_wt_([],Q,Q,Trs,Trs,Nr,Nr).
exp_transitions_wt_([H|T],Q0,Q,[trans(Q0,H/Zero,Q1)|Trs0],Trs,Q0,Nr) :-
    fsa_semiring:zero(fsa_weighted_strings,Zero),
    Nr1 is Q0+1,
    exp_transitions_wt_(T,Q1,Q,Trs0,Trs,Nr1,Nr).

n_eval_pred_pair(Mod,Mod,'$@'(Pred0),'$@'(Pred0),Sym,Sym,Ds):-
    !,
    n_eval_pred(Mod,Pred0,Sym0,Ds),
    (	unknown_symbol(Sym0)
    ->	queue_symbol(Sym)
    ;	Sym0=Sym
    ).

n_eval_pred_pair(ModD,ModR,PredD,PredR,SymD,SymR,Ds):-
    (	PredD==[]
    ->  SymD=[]
    ;   n_eval_pred(ModD,PredD,SymD,Ds)
    ),
    (	PredR==[]
    ->  SymR=[]
    ;   n_eval_pred(ModR,PredR,SymR)
    ).

n_eval_pred(Mod,Pred,Sym,Ds) :-
    lists:member(Sym0,Ds),
    fsa_pred_module:evaluate_predicate(Mod,Pred,Sym0),
    ensure_atom(Sym0,Sym).

ensure_atom(Sym0,Sym) :-
    (	atom(Sym0)
    ->	Sym0=Sym
    ;	number(Sym0)
    ->	number_codes(Sym0,Codes),
	atom_codes(Sym,Codes)
    ;	fsa_util:term_atom(Sym0,Sym)
    ).

n_eval_pred(Mod,Pred,Sym) :-
    fsa_pred_module:evaluate_predicate(Mod,Pred,Sym0),
    (	var(Sym0)
    ->	fsa_globals:fsa_global_get(fl_arbitrary_symbol,Sym0)
    ;	true
    ),
    ensure_atom(Sym0,Sym).

unknown_symbol(X) :-
    (	var(X)
    ->	atom_codes(X,[1])
    ;	atom(X),
	atom_codes(X,[1])
    ).

queue_symbol(X) :-
    (	var(X)
    ->	atom_codes(X,[2])
    ;	atom(X),
	atom_codes(X,[2])
    ).
