:- module( fsa_frozen, [] ).
%% don't export predicates, since there are competing regimes
%% such as fsa_preds, fsa_types, defining the same predicates.

#swi# :- style_check([-atom]).

#!swi# :- use_module(library(ordsets),[ ord_subtract/3 ]).
#swi# :- use_module(ordsets,[ ord_subtract/3 ]).

:- public true/1, regex_notation_to_predicate/2, evaluate_predicate/2,
    conjunction/3, negation/2, display_predicate/2, talks_about/2,
    prepare_complement_of_set/2, complement_of_set/3, determinize_preds/2,
    t_determinize_preds/2, identity/2, class_to_pred/2,
    negated_class_to_pred/2, cleanup/2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% predicates on labels %%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% fsa_frozen has _no_ support for ?, and therefore not for complement
%% either. It is useful for very fast determinization of large automata
%% which don't use the `any' symbol anyway. This module is also used
%% internally for transducers which need to be treated temporarily as
%% recognizers (over pairs of symbols).

true(_) :-
    user:throw(fsa_error("fsa_frozen:true/1 not supported~n",[])).

regex_notation_to_predicate(A,A).

evaluate_predicate(A,A).

conjunction(A,A,A).

negation(_,_) :-
    user:throw(fsa_error("fsa_frozen:negation/2 not supported~n",[])).

display_predicate(A,A).

talks_about(A,A).

prepare_complement_of_set(Fa,Set):-
    fsa_data:fsa_symbols_set(Fa,Set).

complement_of_set(SetOfPredicates,Set,Complements):-
    ord_subtract(Set,SetOfPredicates,Complements).

determinize_preds(L,L).

t_determinize_preds(L,L).

identity(Sym,Sym).

class_to_pred(L,L).

negated_class_to_pred(_,_):-
    user:throw(fsa_error(
            "fsa_frozen:negated_class_to_pred/2 not supported~n",[])).

cleanup(List,List).

