:- module(fsa_help, [fsa_help_listing/0,
		 fsa_help/1,
		 fsa_help/2,
#sic# 		 fsa_help_add_to_menu/2,
		 fsa_help_module/0, fsa_help_module/1,
		 fsa_help_class/1,  fsa_help_class/2,
		 fsa_help_key/1,    fsa_help_key/2,    fsa_help_key/3
		]).

#swi# :- style_check([-atom]).

#yap# :- use_module(library(lists), [ member/2 ]).
#yap# :- use_module(library(charsio), [ format_to_chars/3 ] ).
#!yap# :- use_module(charsio, [ format_to_chars/3 ] ).
#sic# :- use_module(library(tcltk), [ tcl_eval/3 ]).

:- meta_predicate hook(:).
:- meta_predicate is_current_predicate(:).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.
:- public help_info/4.

user:help_info(module,help,"The Help System","\
 \n The help module provides support to create both on-line and off-line\
 \n documentation on Prolog programs. Documentation must be defined by the\
 \n hook predicate help_info/4. Documentation on a per module basis is\
 \n provided if a help_info(module,Module,TitleString,DescriptionString) \
 \n definition is given for Module. In that case the system also checks\
 \n for Module:help_info/4 definitions.  \
 \n \
 \n The module supports production of the help information on standard\
 \n output, (which can be converted into html format), and there also \
 \n is an interface to a graphical user interface based on library(tcltk).").

help_info(class,pred,"List of Predicates","\
 \n This section lists the predicates defined by the help module.").

help_info(class,hook,"List of Hook Predicates","\
 \n This section lists the hook predicates which an application can define \
 \n for the help module.").

help_info(pred,fsa_help_listing,"fsa_help_listing","Lists all help information.").

help_info(pred,fsa_help,"fsa_help/fsa_help(Module)/fsa_help(Module,Class)","\
 \n Use fsa_help/0 to see for which modules help is available. Use fsa_help/1 for\
 \n an overview which classes are available for a given module. Use \
 \n fsa_help(Module,Class) to see for which keys help is available. "). 

help_info(pred,fsa_help_module,"fsa_help_module[(M)]",
"Use fsa_help_module(M) for a full listing of the help information\
 \n available on module M. Without M uses module user.").

help_info(pred,fsa_help_class,"fsa_help_class(C[,M])",
"Use fsa_help_class(C,M) for a full listing of the help information\
 \n available for class C in module M. Without M module user is\
 \n assumed."). 

help_info(pred,fsa_help_key,"fsa_help_key(K[,C[,M]])","\
 \n Use fsa_help_key(K,C,M) for a full listing of the help information\
 \n available for key K in class C in module M. If C (and M) are not\
 \n given, then use variable for C (and M)."). 

help_info(hook,help_info,"help_info(Class,Key,Usage,Expl)","\
 \n Provides help information for Class and Key (both must be\
 \n atoms). Usage and Expl are Prolog strings. Typically the Usage string\
 \n is a short summary, and Expl is a longer explanation. Class is\
 \n typically pred, hook, flag, command, option, etc. Note that each\
 \n module can have its own help_info predicates. You can also define\
 \n user:help_info/4 declarations on the special class module. In that\
 \n case, if a full documentation on a module is requested the Usage\
 \n string is used as the title and the Expl string as an introduction to\
 \n the module. There can also be Module:help_info/4 declarations on the\
 \n special class `class'. If a full listing on a class in Module is\
 \n requested, then Usage and Expl are used as the title and introduction\
 \n to that section. "). 

non_user_module(M) :-
    a_module(M),
    M \== user.

#sic# help_info(pred,fsa_help_add_to_menu,"fsa_help_add_to_menu(Menu,Interp)","\
#sic#  \n Interface of the help system and a graphical user interface based on\
#sic#  \n library(tcltk). Menu must be a menu already existing for Tcl/Tk\
#sic#  \n interpreter Interp. The various help messages are added as cascaded\
#sic#  \n menu entries in Menu. Cf. also the fsa_help/1 predicate and the\
#sic#  \n help_info/4 hook predicate."). 
#sic#  
#sic# fsa_help_add_to_menu(Menu,Interp) :-
#sic#     tcl_eval(Interp,
#sic# 		   format(' \
#sic#  \n proc display_help {module cl k} { \
#sic#  \n   global prolog_variables\
#sic#  \n   prolog fsa_help:gen_widget(W)\
#sic#  \n   set w $prolog_variables(W)\
#sic#  \n   prolog "(fsa_help:help_info($module,$cl,$k,U0,E0),atom_codes(U,U0),atom_codes(E,E0))"\
#sic#  \n   set text $prolog_variables(E)\
#sic#  \n   set header $prolog_variables(U)\
#sic#  \n   set title "Help. Module: $module   Class: $cl"\
#sic#  \n   catch {destroy $w}\
#sic#  \n   toplevel $w\
#sic#  \n   wm title $w $title\
#sic#  \n   frame $w.bot -relief raised -bd 1\
#sic#  \n   pack $w.bot -side bottom -fill both\
#sic#  \n   text $w.text -wrap word -relief flat -bd 10 -yscrollcommand "$w.scroll set" -setgrid true -width 80 -height 20\
#sic#  \n   scrollbar $w.scroll  -command "$w.text yview"\
#sic#  \n   pack $w.scroll -side right -fill y\
#sic#  \n   pack $w.text -expand yes -fill both\
#sic#  \n   $w.text tag configure color2 -foreground red\
#sic#  \n   $w.text insert 0.0 $header color2\
#sic#  \n   $w.text insert end "\n\n"\
#sic#  \n   $w.text insert end $text color1\
#sic#  \n   $w.text configure -state disabled\
#sic#  \n   button $w.bot.button -command "destroy $w" -text {OK}\
#sic#  \n   pack $w.bot.button\
#sic#  \n   }',[]),_),
#sic#     findall(Class,a_class(user,Class),Classes),
#sic#     help_add_to_menu_ls(Classes,Menu,Interp,user),
#sic#     findall(Module,non_user_module(Module),Modules),
#sic#     help_add_to_menu_modules(Modules,Menu,Interp).
#sic# 
#sic# help_add_to_menu_modules([],_,_).
#sic# help_add_to_menu_modules([M|Ms],Menu,Interp):-
#sic#     help_add_to_menu_module(M,Menu,Interp),
#sic#     help_add_to_menu_modules(Ms,Menu,Interp).
#sic# 
#sic# help_add_to_menu_module(Module,Menu,Interp) :-
#sic#     tcl_eval(Interp,
#sic#       format("~w add cascade -label {~w} -menu ~w.module~w",
#sic# 	    [Menu,Module,Menu,Module]),_),
#sic#     tcl_eval(Interp,
#sic#       format('catch "destroy ~w.module~w"',[Menu,Module]),_), 
#sic#     tcl_eval(Interp,
#sic#       format("menu ~w.module~w",[Menu,Module]),NewMenu0),
#sic#     atom_codes(NewMenu,NewMenu0),
#sic#     tcl_eval(Interp,
#sic#       format("~w add command -label Introduction -command {display_help user module {~q}}",[NewMenu,Module]),_),
#sic#     findall(Class,a_class(Module,Class),Classes),
#sic#     help_add_to_menu_ls(Classes,NewMenu,Interp,Module).
#sic# 
#sic# help_add_to_menu_ls([],_menu,_interp,_module).
#sic# help_add_to_menu_ls([Class|Classes],Menu,Interp,Module) :-
#sic#     help_add_to_menu_l(Class,Menu,Interp,Module),
#sic#     help_add_to_menu_ls(Classes,Menu,Interp,Module).
#sic# 
#sic# help_add_to_menu_l(Class,Menu,Interp,Module) :-
#sic#     help_info(Module,class,Class,ClassText,_),!,
#sic#     tcl_eval(Interp,
#sic#       format("~w add cascade -label {~s} -menu ~w.~w",
#sic# 	     [Menu,ClassText,Menu,Class]),_),
#sic#     tcl_eval(Interp,
#sic#       format('catch "destroy ~w.~w"',[Menu,Class]),_), 
#sic#     tcl_eval(Interp,
#sic#       format("menu ~w.~w",[Menu,Class]),NewMenu0),
#sic#     atom_codes(NewMenu,NewMenu0),
#sic#     tcl_eval(Interp,
#sic#       format("~w add command -label Introduction -command {display_help user class {~q}}",[NewMenu,Class]),_),
#sic#     findall(Key,help_info(Module,Class,Key,_,_),Keys0),
#sic#     %sort(Keys0,Keys),
#sic#     Keys0=Keys,
#sic#     help_add_to_menu_l_ls(Keys,Class,NewMenu,Interp,Module).
#sic# 
#sic# help_add_to_menu_l_ls([],_,_,_,_).
#sic# help_add_to_menu_l_ls([Key|Keys],Class,NewMenu,Interp,Module):-
#sic#     %% update is required, since otherwise winfo height doesn't work
#sic#     tcl_eval(Interp,update,_),
#sic#     tcl_eval(Interp,format("winfo reqheight ~w",[NewMenu]),ReqHeight0),
#sic#     tcl_eval(Interp,format("winfo screenheight ~w",[NewMenu]),Height0),
#sic#     number_codes(ReqHeight,ReqHeight0),
#sic#     number_codes(Height,Height0),
#sic#     (	ReqHeight > Height-50
#sic#     ->  tcl_eval(Interp,format("\
#sic# \n           ~w add cascade -label More -menu ~w.more\
#sic# \n           menu ~w.more\
#sic# \n           ",[NewMenu,NewMenu,NewMenu]),Menu0),
#sic# 	atom_codes(Menu,Menu0),
#sic# 	help_add_to_menu_l_ls([Key|Keys],Class,Menu,Interp,Module)
#sic#     ;  tcl_eval(Interp,
#sic# 	format("~w add command -label {~w} -command {display_help {~q} {~q} {~q} }",
#sic# 	       [NewMenu,Key,Module,Class,Key]),_),
#sic#        help_add_to_menu_l_ls(Keys,Class,NewMenu,Interp,Module)
#sic#     ).
#sic# 
#sic# :- public gen_widget/1.
#sic# gen_widget(Symbol):-
#sic#     Prefix='.dt',
#sic#     (	bb_get(Prefix,Integer)
#sic#     ->  true
#sic#     ;   Integer=0
#sic#     ),
#sic#     Next is Integer+1,
#sic#     bb_put(Prefix,Next),
#sic#     format_to_chars('~w~w',[Prefix,Integer],SymbolChars),
#sic#     atom_codes(Symbol,SymbolChars).
#sic# 
#sic# 



:- multifile user:user_help/0.
user:user_help :-
    findall(M,a_module(M),Modules),
    format(user_error,"Help is available on the modules ~w~n",[Modules]),
    format(user_error,"Use fsa_help(Module) fsa_help(Module,Class) to get an overview~n",[]),
    format(user_error,"on what help is available. Use fsa_help_module(Module)~n",[]),
    format(user_error,"fsa_help_class(Class,Module) and fsa_help_key(Key,Class,Module)~n",[]),
    format(user_error,"for an actual listing of the corresponding help information~n",[]),
    format(user_error,"Finally, fsa_help_listing/0 is used to get all help information~n",[]).

fsa_help(Module) :-
    a_module(Module),
    findall(C,a_class(Module,C),Classes),
    format("Help in module ~w is available on classes:~n~w~n",
	   [Module,Classes]).

fsa_help(Module,Class) :-
    a_module(Module),
    a_class(Module,Class),
    findall(K,a_key(Module,Class,K),Keys),
    format("Help in module ~w, class ~w is available for keys:~n~w~n",
	   [Module,Class,Keys]).

fsa_help_listing :-
    findall(M,non_user_module(M),Modules),
    help_user_module(Next),
    help_modules(Modules,Next).

help_modules([],_).
help_modules([M|Ms],N0):-
    fsa_help_module(M,N0),
    N1 is N0+1,
    help_modules(Ms,N1).

fsa_help_module :-
    fsa_help_module(user,1).
fsa_help_module(Module) :-
    a_module(Module),
    fsa_help_module(Module,1).
fsa_help_module(Module,Section) :-
    user:help_info(module,Module,Title,Intro),
    format("~w. ~w: ~s~n~n~s~n~n",[Section,Module,Title,Intro]),
    findall(Class,a_class(Module,Class),Classes),
    help_classes(Classes,Module,Section,1).

help_classes([],_,_,_N).
help_classes([Class|Classes],Module,Section,N0) :-
    fsa_help_class(Class,Module,Section,N0),
    N1 is N0+1,
    help_classes(Classes,Module,Section,N1).

fsa_help_class(Class) :-
    fsa_help_class(Class,user,1,1).
fsa_help_class(Class,Module) :-
    a_module(Module),
    fsa_help_class(Class,Module,1,1).
fsa_help_class(Class,Module,Section,N0) :-
    help_info(Module,class,Class,Title,Intro),
    format("~n~n~w.~w. ~s~n~n~s~n~n",[Section,N0,Title,Intro]),
    findall(Key,help_info(Module,Class,Key,_,_),Keys0),
%%    sort(Keys0,Keys),
    Keys0=Keys,
    help_keys(Keys,Class,Module,Section,N0,1).

help_keys([],_Class,_Module,_S,_Su,_N).
help_keys([Key|Keys],Class,Module,Na,Nb,N0) :-
    help_key(Key,Class,Module,Na,Nb,N0),
    N1 is N0+1,
    help_keys(Keys,Class,Module,Na,Nb,N1).

fsa_help_key(Key):-
    fsa_help_key(Key,_,_).
fsa_help_key(Key,Class) :-
    fsa_help_key(Key,Class,_).
fsa_help_key(Key,Class,Module) :-
    a_module(Module),
%    help_key(Key,Class,Module,1,1,1).
    help_info(Module,Class,Key,Title,Expl),
    format("~nModule: ~w~n",[Module]),
    format("Class:  ~w~n",[Class]),
    format("~s~n~n~s~n~n",[Title,Expl]).

help_key(Key,Class,Module,Na,Nb,N0) :-
    help_info(Module,Class,Key,Title,Expl),
    format("~n~n~w.~w.~w. ~s~n~n~s~n~n",[Na,Nb,N0,Title,Expl]).

a_module(Module) :-
    findall(Mod,(user:help_info(module,Mod,_,_),Mod\==user),Modules),
    %%sort(Modules0,Modules),
    member(Module,[user|Modules]).

/*
a_class(Module,Class) :-
    findall(Class,(help_info(Module,Class,_,_,_),
		   Class\==module,
		   Class\==class),
		   Classes0),
    sort(Classes0,Classes),
    member(Class,Classes).

a_class(Module,Class) :-
    findall(Class,help_info(Module,class,Class,_,_),Classes0),
    sort(Classes0,Classes),
    member(Class,Classes).
*/
a_class(Module,Class) :-
    help_info(Module,class,Class,_,_).


a_key(Module,Class,Key) :-
    findall(K,help_info(Module,Class,K,_,_),Keys0),
    %%sort(Keys0,Keys),
    Keys0=Keys,
    member(Key,Keys).

%% only called from help_listing.
help_user_module(Next) :-
    user:help_info(module,user,Title,Intro),
    format("~w. ~s~n~n~s~n~n",[1,Title,Intro]),
    findall(Class,a_class(user,Class),Classes),
    help_user_classes(Classes,2,Next).


help_user_classes([],N,N).
help_user_classes([Class|Classes],N0,N) :-
    help_user_class(Class,N0),
    N1 is N0+1,
    help_user_classes(Classes,N1,N).

help_user_class(Class,N0) :-
    user:help_info(class,Class,Title,Intro),
    format("~n~n~w. ~s~n~n~s~n~n",[N0,Title,Intro]),
    findall(Key,user:help_info(Class,Key,_,_),Keys0),
    %%sort(Keys0,Keys),
    Keys0=Keys,
    help_user_keys(Keys,Class,N0,1).

help_user_keys([],_Class,_Su,_N).
help_user_keys([Key|Keys],Class,Nb,N0) :-
    help_user_key(Key,Class,Nb,N0),
    N1 is N0+1,
    help_user_keys(Keys,Class,Nb,N1).

help_user_key(Key,Class,Nb,N0) :-
    user:help_info(Class,Key,Title,Expl),
    format("~n~n~w.~w. ~s~n~n~s~n~n",[Nb,N0,Title,Expl]).


help_info(Module,Class,Key,Title,Expl) :-
    hook(Module:help_info(Class,Key,Title,Expl)).

%% from hdrug_util
hook(Module:Call) :-
    is_current_predicate(Module:Call),
    Module:Call.

is_current_predicate(Module:Call) :-
    (	current_predicate(_,Module:Call)
    ->  true
    ;	predicate_property(Module:Call,imported_from(NewModule)),
	is_current_predicate(NewModule:Call)
    ).
