%%% -*- Module: fsa_java; -*-
:- module(fsa_java,
	  [ fsa_java/2,
	    fsa_java_fa/2
	  ]).

#swi# :- style_check([-atom]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.


user:help_info(class,fsa_java,"JAVA Code Generation","\
 \n FSA supports the production of JAVA code on the basis of\
 \n a finite automaton.  Before the automaton is translated into C code, \
 \n it is determinized. In the case of transducers, Mohri's determinization\
 \n algorithm is applied. Note however that certain transductions cannot be\
 \n determinized: in that case the algorithm will not terminate. \
 \n \
 \n The JAVA program will define a class (named in accordance with the given \
 \n output file name) which inherits from Applet. The class defines the method:\
 \n \
 \n         static void main(String argv[])\
 \n \
 \n The instance itself is an applet in which you can write strings which\
 \n are checked against the automaton.  The JAVA program starts a\
 \n graphical user interface in which you can input strings (if the option\
 \n -w is the single option), or reads lines from standard input and\
 \n writes the result of applying the automaton to standard output.\
 \n \
 \n         public void gui();\
 \n \
 \n starts a graphical user interface in which you can input strings.\
 \n \
 \n         public DFA automaton();\
 \n \
 \n returns the automaton part of the applet. This DFA class in turn defines the \
 \n following methods:\
 \n \
 \n         public boolean Recognizer();\
 \n     public boolean Transducer();\
 \n     public boolean WeightedRecognizer();\
 \n     public boolean WeightedTransducer();\
 \n     \
 \n As well as:\
 \n \
 \n         public void filter ()\
 \n \
 \n reads lines from standard input and displays the result of running each line\
 \n through the automaton to standard output.\
 \n \
 \n         public boolean accepts ( String in )\
 \n     public String transduces ( String in )\
 \n     public Integer weighs ( String in )\
 \n     public StringWeightPair string_weight( String in )\
 \n \
 \n The `main' method is provided only if the global variable *java_with_main* is\
 \n set to on.\
 \n \
 \n In order to be able to run the JAVA code generated by FSA you need a java\
 \n compiler, as well as the class files which are distributed with FSA. You\
 \n have to ensure that the JAVA compiler knows where to find these files. For\
 \n instance, if you have installed FSA in /usr/local/lib then the class files\
 \n are in /usr/local/lib/fsa/Java. For instance, after running the FSA command:\
 \n \
 \n         % fsa write=java -r '[a,b,c,? *]' z.java\
 \n \
 \n you compile the JAVA file with e.g.:\
 \n \
 \n         % javac  -classpath /usr/local/lib/fsa/Java:\\         \
 \n                         /usr/lib/java/lib/classes.zip:.  z.java\
 \n \
 \n Instead of the -classpath option to javac it is preferable to include the\
 \n relevant class directories in the CLASSPATH environment variable. You can\
 \n now run the program using one of:\
 \n \
 \n         % java z -w\
 \n         % java z\
 \n \
 \n \
 \n \
 \n The representation of a finite-automaton in JAVA is similar to the\
 \n technique explained on page 43 (table 4.2) of Jan Daciuk's\
 \n dissertation `Incremental Construction of Finite-State Automata and\
 \n Transducers and their use in the Natural Language Processing'.\
 \n Politechnika Gdanska, 1998, except that instead of the number\
 \n of transitions we have a boolean flag indicating for each\
 \n line whether that line is the last transition for the current state.\
 \n \
 \n The special input symbol ^A is used in the representation of the automaton\
 \n in C to indicate a symbol not otherwise mentioned in the automaton: it will\
 \n match any such symbol. Similarly, in transducers the symbol ^B is used\
 \n to indicate an unknown symbol with an associated identity. The corresponding\
 \n output symbol is also ^B. When a string is transduced this ^B is replaced\
 \n by the actual input symbol (by means of a queue).  An unknown\
 \n output symbol without an associated identity is represented using the\
 \n symbol given by the global flag fl_arbitrary_symbol.  In the case of\
 \n final states with multiple outputs a special meta-notation is used using\
 \n a special symbol given by the global variable fl_multiple_symbol_start\
 \n which starts a sequence of possible outputs where each output is seperated\
 \n using a symbol given by the global variable fl_multiple_symbol_sep.").


user:help_info(xpred,fsa_java,
"fsa_java(+FileIn,+FileOut)","\
 \n A JAVA program is written to FileOut for the recognizer read\
 \n from FileIn. The JAVA program will read lines from standard input\
 \n and apply the automaton to each line. ").

fsa_java(Source,Target):-
    fsa_fl:fsa_fl_file(java,Source,Target).

fsa_java_fa(Fa,Target):-
    fsa_fl:fsa_fl_fa(java,Fa,Target).
