:- module(fsa_m_hash, [%% updatable hashes using mutable terms
		       fsa_m_hash_new/1,
                       fsa_m_hash_new/2,
		       fsa_m_hash_get/3,
		       fsa_m_hash_get/4,
		       fsa_m_hash_put/3,
                       fsa_m_hash_put/5 
		    ]).

#swi# :- style_check([-atom]).

:- discontiguous help_info/4.
:- public help_info/4.

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

user:help_info(module,fsa_m_hash,"Mutable Hashes",
"This module provides a mutable hash datastructure on top of\
 \n the fsa_hash datastructure.\
 \n \
 \n Here's an overview of the predicates provided:  \
 \n \
 \n * fsa_m_hash_new/[1,2]        create a new mutable hash\
 \n * fsa_m_hash_get/[3,4]        lookup a value from a mutable hash\
 \n * fsa_m_hash_put/[3,5]        update a value in a mutable hash").

help_info(class,pred,"List of Predicates",
"This section lists the predicates defined by this module.").

help_info(pred,fsa_m_hash_new,"fsa_m_hash_new(-MutableFsaHash[,Size])",
"Initializes a new MutableFsaHash with size Size; or default size if\
 \n there is no second argument. The default is determined by the global\
 \n variable *hash_size*."). 

fsa_m_hash_new(Thing):-
    fsa_hash:fsa_hash_new(Thing).

fsa_m_hash_new(Thing,Size):-
    fsa_hash:fsa_hash_new(Thing,Size).

help_info(pred,fsa_m_hash_get,
"fsa_m_hash_get(+Key,?Val,+MutableFsaHash) \
 \n fsa_m_hash_get(+Key,?Val,?Default,+MutableFsaHash)",
"Val is unified with the value associated with Key in\
 \n MutableFsaHash. If no such key exists in MutableFsaHash already, then\
 \n the predicate *succeeds* without binding Val (in the first form) or\
 \n unifies Val and Default (second form)."). 

fsa_m_hash_get(Key,Val,Array) :-
    fsa_hash:fsa_hash_access(Key,Val0,Array),
    (	var(Val0)
    ->	true
    ;   get_mutable(Val,Val0)
    ).

fsa_m_hash_get(Key,Val,Default,Array):-
    fsa_hash:fsa_hash_access(Key,Val0,Array),
    (	var(Val0)
    ->  Val=Default
    ;   get_mutable(Val,Val0)
    ).

help_info(pred,fsa_m_hash_put,
"fsa_m_hash_put(+Key,?Val,+MutableFsaHash)\
 \n fsa_m_hash_put(+Key,?OldVal,?Default,?Val,+MutableFsaHash)",
"The value associated with Key in MutableFsaHash is updated to Val\
 \n (using the SICStus built-in update_mutable). OldVal is unified with\
 \n the old value (it it existed) or with Default (if it didn't exist)."). 

fsa_m_hash_put(Key,Val,Array) :-
    fsa_hash:fsa_hash_access(Key,Val0,Array),
    (	var(Val0)
    ->	create_mutable(Val,Val0)
    ;	update_mutable(Val,Val0)
    ).

fsa_m_hash_put(Key,ValOld,ValDefault,Val,Array) :-
    fsa_hash:fsa_hash_access(Key,Val0,Array),
    (	var(Val0)
    ->	ValOld=ValDefault,
	create_mutable(Val,Val0)
    ;   get_mutable(ValOld,Val0),
	update_mutable(Val,Val0)
    ).

