#swi# :- style_check([-atom]).

#yap# :- style_check([single_var,discontiguous]).

#yap# :- op(1150,fx,discontiguous).

#yap# :- yap_flag(character_escapes,on).

#yap# :- use_module(library(charsio),[ format_to_chars/3,read_from_chars/2 ]).
#!yap# :- use_module(charsio,[ format_to_chars/3,read_from_chars/2 ]).
#yap# :- use_module(library(lists), [ append/3 ]).

:- multifile help_info/4.
:- discontiguous help_info/4.

help_info(module,user,
"FSA6: Finite State Automata Utilities Version 6",Text):-
    fsa_globals:fsa_global_get(fsa_version,Atom),
    charsio:format_to_chars("(manual generated with FSA Utilities version ~a)~n~n",[Atom],VersionInfo),
    append(VersionInfo,
"FSA6 is a collection of utilities to\
 \n \
 \n          * *construct* finite automata (from regular expressions)\
 \n          * *manipulate* finite automata\
 \n          * *visualise* finite automata\
 \n          * *apply* finite automata\
 \n \
 \n FSA6 supports a number of different types of automata:\
 \n          * recognizers\
 \n          * weighted recognizers (aka string-to-weight transducers)\
 \n          * transducers (aka string-to-string transducers)\
 \n          * weighted transducers (aka string-to-string-weight transducers)\
 \n \
 \n 1.1 Functionality\
 \n \
 \n 1.1.1 Constructing Finite Automata with Regular Expressions\
 \n \
 \n Many basic regular expression operators are provided, both for\
 \n acceptors and transducers. Moreover, it is easy to define new regular\
 \n expression operators. The built-in regular expression operators\
 \n include:\
 \n \
 \n * Concatenation; Kleene star; Kleene plus; Option; Union\
 \n * Complement; Difference; Intersection\
 \n * Reversal; Containment; Ignore\
 \n * Composition; Cross-product; Domain; Range; Identity; Inversion;\
 \n * Interval\
 \n * `Any' meta-symbol. \
 \n * Arbitrary predicates instead of symbols\
 \n * operators to construct weighted automata\
 \n \
 \n 1.1.2 Manipulating Finite Automata\
 \n \
 \n Tools are provided to manipulate finite automata. Such manipulations\
 \n include determinization and minimization (both the classical\
 \n algorithms for recognizers and the recent algorithms for transducers\
 \n are provided). \
 \n \
 \n * Determinization.  Currently there are three different implementations of\
 \n this algorithm, depending on how epsilon transitions (jumps) are\
 \n treated. There is also an implementation of Mohri's determinization\
 \n algorithm, both for ordinary (string-to-string) transducers and\
 \n string-to-weight transducers. The implementation is described\
 \n in a paper in Computational Linguistics, available from\
 \n http://www.let.rug.nl/~vannoord/papers/\
 \n \
 \n * Minimization.  Three different minimization algorithms are\
 \n supported. There is also an implementation of Mohri's minimization\
 \n algorithm, both for ordinary (string-to-string) transducers and\
 \n string-to-weight transducers. \
 \n \
 \n * Random generation of finite automata, based on the algorithm in\
 \n Leslie (1995).\
 \n \
 \n * Epsilon-removal.\
 \n \
 \n * Completion and Incompletion: extending a given automaton in order to \
 \n make the transition table total (typically by adding a sink state and\
 \n adding transitions to this sink state); and removing transitions leading to\
 \n sink states. \
 \n \
 \n * Regular manipulations. The various regular expression operators can\
 \n be applied to automata directly as well. \
 \n \
 \n \
 \n 1.1.3 Applying Finite Automata\
 \n \
 \n * Acceptance. Tools to check a given string for acceptance by a\
 \n recognizer. \
 \n \
 \n * Transduction. Tools to apply a transducer to a given input string. \
 \n \
 \n * Production. Tools to produce strings of a given recognizer, and\
 \n pairs of strings for a given transducer.\
 \n \
 \n * Code Generation. Tools to compile finite automata into efficient\
 \n Prolog, C, C++, and JAVA programs which can be used to check whether a\
 \n given string is in the language defined by the automaton, or to generate\
 \n the transduction of a given string w.r.t. a given transducer.\
 \n \
 \n \
 \n 1.1.4 Visualizing Finite Automata\
 \n \
 \n Much attention has been paid to be able to visualize finite state\
 \n recognizers and finite state transducers. Support includes built- in\
 \n visualization and interfaces to third party software:\
 \n \
 \n * DOT. The program is able to produce a representation of a\
 \n finite automaton compatible with the DOT graph visualisation program.\
 \n DOT (part of AT&T's graphviz) is a tool that automatically figures\
 \n out how a graph is best displayed (crossing-edges reduction, etc).\
 \n It can produce e.g. Postscript output. An example is\
 \n http://www.let.rug.nl/~vannoord/Fsa/Manual/dot.png.\
 \n \
 \n * VCG. The program is able to produce a representation of a\
 \n finite state automaton compatible with the VCG graph\
 \n visualisation program. VCG is a tool that automatically figures\
 \n outhow a graph is best displayed (crossing-edges reduction, etc). \
 \n An example is http://www.let.rug.nl/~vannoord/Fsa/Manual/vcg.png.\
 \n The VCG program is now also known as aiSee.\
 \n \
 \n * daVinci.  The program is able to produce a representation of a\
 \n finite state automaton compatible with the daVinci 1.4 graph\
 \n visualisation program. This program automatically computes the most\
 \n optimal way to view the finite-state automaton by minimizing the\
 \n number of crossing edges. Postscript output can easily be generated\
 \n from the result. An example is http://www.let.rug.nl/~vannoord/Fsa/Manual/daVinci.png.\
 \n \
 \n * TK Widget.  The package contains an interface to a TK Widget to browse\
 \n finite state automata, providing a graphical user-interface for the\
 \n toolbox. The TK Widget is explained in much more detail below. An\
 \n example is http://www.let.rug.nl/~vannoord/Fsa/Manual/dump.png. Note that the\
 \n GUI is not an integral part of the toolbox; it makes perfect sense to\
 \n use the program in batch mode. The graphical user interface is only\
 \n available under SICStus.\
 \n \
 \n * LaTeX (if you want to be able to use the result you need the \
 \n pstricks package).  An example is\
 \n http://www.let.rug.nl/~vannoord/Fsa/Manual/pstricks.png. \
 \n \
 \n * Postscript (thanks to Peter Kleiweg).\
 \n An example is http://www.let.rug.nl/~vannoord/Fsa/Manual/pk.png. \
 \n \
 \n \
 \n 1.2 How to use the toolbox\
 \n \
 \n There are a number of ways that the toolbox is meant to be used:\
 \n \
 \n * interactively using a command interpreter and/or a graphical user\
 \n interface. For example, in order to use fsa interactively with the\
 \n graphical user interface, use the command:\
 \n \
 \n         % fsa -tk\
 \n \
 \n * as a UNIX-like filter. In such cases you use the fsa command with a\
 \n number of options. For instance:\
 \n \
 \n         % fsa write=postscript -r '[a,b+,c*,d]' | ghostview -\
 \n \
 \n * as a library in your own Prolog program. You can incorporate the FSA\
 \n program in your own program, just as you can use other Prolog\
 \n libraries. In order for this to work, you simply need to load the\
 \n file fsa_library.pl in the installation directory. For example (if you\
 \n use SWI Prolog, you need 'consult' instead of 'use_module'):\
 \n \
 \n         % sicstus\
 \n         SICStus ...\
 \n         Licensed to ...\
 \n         | ?- use_module(fsa_library).\
 \n         ...\
 \n         ...\
 \n         ...\
 \n         yes\
 \n         | ?- fsa_regex_atom_compile('[a*,b^,{d,e}]',L).\
 \n         L = fa(r(fsa_preds),3,[0],[1],[trans(0,a,0),trans(0,b,2),\
 \n             trans(0,d,1),trans(0,e,1),trans(2,d,1),trans(2,e,1)],[]) ?   \
 \n         yes\
 \n         | ?- fsa_regex_transduces('{a:b,? -a}*',\"ababac\",L), atom_codes(Atom,L).\
 \n         L = [98,98,98,98,98,99],\
 \n         Atom = bbbbbc ?\
 \n         yes\
 \n         | ?-\
 \n \
 \n All predicates that are imported have names starting with *fsa*. All\
 \n module names start with *fsa* as well. \
 \n         \
 \n 1.3 Examples\
 \n \
 \n The package comes with a number of larger examples These examples\
 \n include both automata and extended regular expression definitions.\
 \n \
 \n * Examples/Automata\
 \n \
 \n Regular expression operators which allows to input an automaton\
 \n by listing its transitions, start states, and final states.\
 \n Contributed by Dale Gerdemann.\
 \n \
 \n * Examples/Booleans\
 \n \
 \n A collection of regular expression operators including boolean\
 \n operators and various predicates of automata. Contributed by Dale\
 \n Gerdemann.\
 \n \
 \n * Examples/Bouma\
 \n \
 \n A finite-state automaton of all possible Dutch monosyllabic\
 \n words. Contributed by Gosse Bouma.\
 \n \
 \n * Examples/DutchWords\
 \n \
 \n Dutch words, taken somewhere from ftp site (see ftp_info.txt). This\
 \n list of words can then be used to experiment with the option to create\
 \n perfect hashes (-dict2ph).\
 \n \
 \n * Examples/GerdemannVannoord99\
 \n \
 \n The replace operator as defined in our EACL 99 paper. Also some\
 \n further examples with longest match and finite-state parsing. Contributed\
 \n by Gerdemann and van Noord.\
 \n \
 \n * Examples/Graph2Phon\
 \n \
 \n Grapheme to Phoneme conversion for Dutch, as described in Bouma's ACL\
 \n 2000 paper. Uses the Celex format for phonemes. Contributed by Gosse Bouma.\
 \n \
 \n * Examples/Grimley-Evans\
 \n \
 \n Implementation of the Hopcroft minimization algorithm by Edmund\
 \n Grimley-Evans, as described in his WIA 97 paper. This code made me\
 \n re-implement the FSA implementation of Hopcroft's\
 \n algorithm. Contributed by Grimley-Evans.\
 \n \
 \n * Examples/HMM\
 \n \
 \n HMM's can be seen as a special type of weighted finite automata. This\
 \n example implements the Baum-Welch training algorithm. Fairly\
 \n simple-minded implementation.\
 \n \
 \n * Examples/Hyphenation\
 \n \
 \n Code to convert LaTeX hyphenation patterns into a single (large)\
 \n finite-state transducer that can be used directly to hyphenate a given\
 \n word. By Gosse Bouma and van Noord, after a suggestion by Lauri\
 \n Karttunen.\
 \n \
 \n * Examples/KaplanKay94\
 \n \
 \n These examples are taken from Kaplan and Kay, Regular Models of\
 \n Phonological Rule Systems, Computational Linguistics, 20(3), 1994.\
 \n Simple examples of transducers, and composition.\
 \n \
 \n * Examples/Karttunen91\
 \n \
 \n These examples are taken from Karttunen, Finite-state Constraints,\
 \n Proceedings International Conference on Current Issues in\
 \n Computational Linguistics, Universiti Sains Malaysia, Penang,\
 \n 1991. Simple examples of transducers, and composition.\
 \n \
 \n * Examples/Karttunen95\
 \n \
 \n Lauri Karttunen, The Replace Operator, ACL 1995, MIT Boston. Fairly\
 \n complex examples of regular expression operator definitions.\
 \n \
 \n * Examples/Karttunen96\
 \n \
 \n Lauri Karttunen, Directed Replacement, ACL 1996. Includes soundex\
 \n example from MLTT home page.\
 \n \
 \n * Examples/Karttunen97\
 \n \
 \n Lauri Karttunen, The Replace Operator, 1997. In volume edited by Roche\
 \n and Schabes.\
 \n \
 \n * Examples/Mohri97\
 \n \
 \n Simple examples of weighted automata.\
 \n \
 \n * Examples/MohriSproat96\
 \n \
 \n Mehryar Mohri and Richard Sproat, An Efficient Compiler for Weighted\
 \n Rewrite Rules. 34th Annual Meeting of the ACL, Santa Cruz 1996, pages\
 \n 230-238. This only treats the non-weighted case. Nice example of the\
 \n power of the regular expression language: their algorithm only takes a\
 \n few paragraphs in FSA6.\
 \n \
 \n * Examples/MSOL\
 \n \
 \n Nathan Vailette, Logical Specifications of Transducers for NLP. In:\
 \n FSMNLP 2001, ESSLLI Helsinki. electronically available from\
 \n http://www.let.rug.nl/~vannoord/alp/esslli_fsmnlp Contributed by\
 \n Nathan Vailette.\
 \n \
 \n * Examples/Nederhof\
 \n \
 \n These are examples used by Mark-Jan Nederhof while\
 \n investigating finite-state approximations of context-free\
 \n grammars. The larger examples were used in my Computational\
 \n Linguistics paper, The Treatment of Epsilon-moves in Subset Construction, \
 \n available from http://www.let.rug.nl/~vannoord/papers/\
 \n Contributed by Nederhof.\
 \n \
 \n * Examples/Nerbonne\
 \n \
 \n examples from http://www.let.rug.nl/~nerbonne/teach.html material for\
 \n a course on computational morphology. Simple examples of transducers.\
 \n \
 \n * Examples/OptimalityTheory\
 \n \
 \n Implementation of Lauri Karttunen, The Proper Treatment of Optimality\
 \n in Computational Phonology. FSMNLP 1998, Ankara.  Includes definition\
 \n of lenient composition operator and syllibification algorithm. Also\
 \n includes Gerdemann/van Noord (even more proper?) alternative\
 \n implementation. Contributed by Gerdemann and van Noord.\
 \n \
 \n * Examples/PredModules\
 \n \
 \n Examples of predicate modules; for example using bitvectors to\
 \n represent sets of symbols, or using types. The bitvector stuff is only\
 \n available under SICStus.\
 \n \
 \n * Examples/PereiraRiley96\
 \n \
 \n Fernando C. N. Pereira and Michael D. Riley, Speech Recognition by\
 \n Composition of Weighted Finite Automata, 1996 (on cmp-lg). Also\
 \n appears as chapter 15 of the volume edited by Roche and\
 \n Schabes. Simple examples of weighted composition. Definition of their\
 \n version of the composition operator (filtering our spurious paths).\
 \n \
 \n * Examples/Queens\
 \n \
 \n Solving the N-queens problem with regular expressions, by Dale\
 \n Gerdemann. Another solution by G. van Noord. Interesting examples of\
 \n definitions of regular expression operators.\
 \n \
 \n * Examples/Random\
 \n \
 \n Random automata. Used for the experiments documented in my FSMNLP98\
 \n paper, on the treatment of epsilon-moves in subset construction.\
 \n \
 \n * Examples/Recognizers\
 \n \
 \n Small examples.\
 \n \
 \n * Examples/RocheSchabes95\
 \n \
 \n Emmanuel Roche and Yves Schabes, Deterministic Part-of-speech Tagging\
 \n with Finite-state Transducers, Computational Linguistics, 21(2),\
 \n 1995. Small examples of transducers. Also includes a definition of the\
 \n local extension operator.\
 \n \
 \n * Examples/RocheSchabes97\
 \n \
 \n Roche and Schabes, Introduction. In: Roche and Schabes (eds), Finite\
 \n State Language Processing. MIT Press 1997. Includes implementations of\
 \n is_functional, unambiguous, is_subsequential, build_bimachine,\
 \n bitransform. Also has simple utils to apply and visualize bimachines.\
 \n \
 \n * Examples/Spell\
 \n \
 \n Implements a simple spell-checker as the combination of a dictionary\
 \n and strings within Levenshtein distance d of the words in the\
 \n dictionary (for some fixed d). Interesting application of the priority\
 \n union operator of Karttunen (1998).\
 \n \
 \n * Examples/Transducers\
 \n \
 \n small stuff, including my attempt to translate Dutch temporal\
 \n expression into a numerical format (that one is quite large, in fact).\
 \n \
 \n * Examples/twolevel\
 \n \
 \n Definitions to implement twolevel rules in the style of Kimmo. Contributed\
 \n by Rob Malouf, Gosse Bouma, Gertjan van Noord.\
 \n \
 \n * Examples/Weights\
 \n \
 \n Small stuff, weighted automata.\
 \n \
 \n * Examples/Wordgraphs\
 \n \
 \n Some small acyclic weighted automata.\
 \n \
 \n 1.5 References\
 \n \
 \n * Alfred V. Aho and John E. Hopcroft and Jeffrey D. Ullman, The Design\
 \n and Analysis of Computer Algorithms. Addison Wesley, 1974.\
 \n \
 \n * Meera Blattner and Tom Head, Single-Valued a-Transducers. In: Journal\
 \n of Computer and System Sciences, 15. Pp. 310--327. 1977.\
 \n \
 \n * J.A. Brzozowski. Canonical Regular Expressions and Minimal State\
 \n Graphs for Definite Events. In: Mathematical Theory of Automata,\
 \n Polytechnic Press Brooklyn, 1962. \
 \n \
 \n * Gosse Bouma, A Modern CL course using Dutch.\
 \n In: EACL 99 Postconference Workshop ``Computer and Internet\
 \n supported Education in Language and Speech Technology'', June 12 1999,\
 \n Bergen Norway.\
 \n \
 \n * Jan Daciuk, Incremental Construction of Finite-State Automata and\
 \n Transducers and their use in the Natural Language\
 \n Processing. Thesis, Politechnika Gdanska, 1998.\
 \n \
 \n * Jeffrey Friedl. Mastering Regular Expressions. O'Reilly 1997.\
 \n \
 \n * Dale Gerdemann and Gertjan van Noord. Transducers from Rewrite Rules\
 \n with Backreferences. EACL 1999 Bergen.\
 \n \
 \n * John E. Hopcroft. An n log n algorithm for minimizing the states in a\
 \n finite automaton. In: Z. Kohavi (editor), The Theory of Machines and\
 \n Computations. Academic Press 1971.\
 \n \
 \n * John E. Hopcroft and Jeffrey D. Ullman. Introduction to Automata\
 \n Theory, Languages and Computation. Addison Wesley 1979.\
 \n \
 \n * C. Douglas Johnson. Formal Aspects of Phonological\
 \n Descriptions. Mouton The Hague, 1972.\
 \n \
 \n * Ronald M. Kaplan and Martin Kay. Regular Models of Phonological Rule\
 \n Systems. Computational Linguistics 20 (3) 1994.\
 \n \
 \n * Lauri Karttunen. Finite-state Constraints. In: Proceedings\
 \n International Conference on Current Issues in Computational\
 \n Linguistics. Unviersiti Sains Malaysia, Penang. 1991.\
 \n \
 \n * Lauri Karttunen. The Replace Operator. 33rd ACL, Boston, 1995.\
 \n \
 \n * Lauri Karttunen and Jean-Pierre Chanod and Gregory Grefenstette and\
 \n Anne Schiller, Regular Expressions for Language Engineering. Natural\
 \n Language Engineering 2 (4) 1996.\
 \n \
 \n * Lauri Karttunen, Directed Replacement. ACL 1996 Santa Cruz.\
 \n \
 \n * Lauri Karttunen, The Proper Treatment of Optimality Theory in\
 \n Computational Phonology. In: FSMNLP 1998, Ankara.\
 \n \
 \n * George Anton Kiraz and Edmund Grimley-Evans, Multi-Tape Automata\
 \n for Speech and Language Systems: A Prolog Implementation.  \
 \n In D. Wood and S. Yu (eds.), Automata Implementation,\
 \n Lecture Notes in Computer Science 1436, Springer, 1998.\
 \n \
 \n * Ted Leslie, Efficient Approaches to Subset Construction. University\
 \n of Waterloo 1995. \
 \n \
 \n * Mehryar Mohri, Compact Representations by Finite-state Transducers.\
 \n ACL New Mexico 1994.\
 \n \
 \n * Mehryar Mohri, Finite-State Transducers in Language and Speech\
 \n Processing, Computational Linguistics 23 (2) 1997.\
 \n \
 \n * Mehryar Mohri and Fernando C.N. Pereira and Michael Riley. A rational\
 \n design for a weighted finite-state transducer library. In: Automata\
 \n Implementation. WIA '97. Lecture Notes in Computer Science 1436. \
 \n Spring Verlag 1998.\
 \n \
 \n * Mehryar Mohri and Richard Sproat. An Efficient Compiler for Weighted\
 \n Rewrite Rules. ACL 1996 Santa Cruz.\
 \n \
 \n * Gertjan van Noord. FSA Utilities: A Toolbox to Manipulate Finite-state \
 \n Automata. In: Raymond, Woord, Yu (editors), Automata\
 \n Implementation. Lecture Notes in Computer Science 1260. Spring Verlag 1997. \
 \n \
 \n * Gertjan van Noord. The Treatment of Epsilon Moves in Subset\
 \n Construction. In: Computational Linguistics 26 (1) 2000.\
 \n \
 \n * Gertjan van Noord and Dale Gerdemann. An Extendible Regular Expression \
 \n Compiler for Finite-state Approaches in Natural Language\
 \n Processing. WIA 1999.\
 \n \
 \n * Gertjan van Norod and Dale Gerdemann. Finite State Transducers with\
 \n Predicates and Identity. Grammars 4 (3) 2001.\
 \n \
 \n * Emmanuel Roche and Yves Schabes. Deterministic Part-of-speech Tagging\
 \n with Finite-state Transducers. Computational Linguistics 21 (2) 1995.\
 \n \
 \n * Emmanuel Roche and Yves Schabes. Finite-State Language Processing. MIT \
 \n Press 1997.\
 \n \
 \n * Bruce Watson, Taxonomies and Toolkits of Regular Language\
 \n Algorithms. Ph.D. thesis TU Eindhoven. 1996.\
 \n \
 \n * Bruce Watson, Implementing and Using Finite Automata Toolkits. In:\
 \n Proceedings of the ECAI'96 Workshop Extended Finite State Models of\
 \n Language. 1996.\
 \n \
 \n 1.6 Papers using FSA\
 \n \
 \n * Gosse Bouma, A Modern CL course using Dutch.\
 \n In: EACL 99 Postconference Workshop ``Computer and Internet\
 \n supported Education in Language and Speech Technology'', June 12 1999,\
 \n Bergen Norway.\
 \n \
 \n * Gosse Bouma, A Finite State and Data-Oriented Method for Grapheme to\
 \n Phoneme Conversion. In: 1st Meeting of the North American Chapter of \
 \n the Association for Computational Linguistics, 303-310 Seattle 2000.\
 \n \
 \n * Gosse Bouma, A modern Computational Linguistic Course using Dutch. In: \
 \n Frank van Eynde and Ineke Schuurman, editors, CLIN 1998, Papers from the \
 \n ninth CLIN Meeting, Amsterdam, 1999. Rodopi Press. \
 \n \
 \n * Gosse Bouma, Finite State Methods for Hyphenation. In:\
 \n FSMNLP 2001, ESSLLI Helsinki.\
 \n \
 \n * Dale Gerdemann and Gertjan van Noord. Transducers from Rewrite Rules\
 \n with Backreferences. EACL 1999 Bergen.\
 \n \
 \n * Dale Gerdemann and Gertjan van Noord. Approximation and Exactness in\
 \n Finite State Optimality Theory. In: FINITE-STATE PHONOLOGY: SIGPHON 2000, \
 \n Fifth Meeting of the ACL Special Interest Group in Computational\
 \n Phonology, COLING 2000.\
 \n \
 \n * George Anton Kiraz, Multi-Tiered Nonlinear Morphology Using Multi-Tape\
 \n Finite Automata: A case study on Syriac and Arabic. In: Computational \
 \n Linguistics, 26 (1) 2000.\
 \n \
 \n * Edwin Kuipers, Ill-formed Input and Finite-state Devices. MA Thesis,\
 \n Rijksuniversiteit Groningen, 1997.\
 \n \
 \n * Gertjan van Noord, FSA Utilities: Manipulation of Finite-state Automata \
 \n implemented in Prolog, in: Proceedings of the First International Workshop \
 \n on Implementing Automata. London Ontario Canada, 1996. \
 \n \
 \n * Gertjan van Noord. FSA Utilities: A Toolbox to Manipulate Finite-state \
 \n Automata. In: Raymond, Woord, Yu (editors), Automata\
 \n Implementation. Lecture Notes in Computer Science 1260. Spring Verlag 1997. \
 \n \
 \n * Gertjan van Noord. The Treatment of Epsilon Moves in Subset\
 \n Construction. In: Computational Linguistics 26 (1) 2000.\
 \n \
 \n * Gertjan van Noord and Dale Gerdemann. An Extendible Regular Expression \
 \n Compiler for Finite-state Approaches in Natural Language\
 \n Processing. WIA 1999.\
 \n \
 \n * Gertjan van Norod and Dale Gerdemann. Finite State Transducers with\
 \n Predicates and Identity. Grammars X (X) 2001.\
 \n \
 \n * Nathan Vailette, Logical Specifications of Transducers for NLP. In:\
 \n FSMNLP 2001, ESSLLI Helsinki.\
 \n \
 \n * Markus Walther, Finite-State Reduplication in One-Level Prosodic Morphology.\
 \n In: 1st Meeting of the North American Chapter of the Association for \
 \n Computational Linguistics, 296-302 Seattle 2000.\
 \n \
 \n * Markus Walther, One-Level Prosodic Morphology. Marburger Arbeiten zur \
 \n Linguistic 1, University of Marburg. 64 pp.\
 \n \
 \n * Markus Walther, Temiar Reduplication in One-Level Prosodic\
 \n Morphology. In: FINITE-STATE PHONOLOGY: SIGPHON 2000, \
 \n Fifth Meeting of the ACL Special Interest Group in Computational\
 \n Phonology, COLING 2000.\
 \n \
 \n 1.7 Links\
 \n \
 \n * Up-to-date information on the program can be obtained via \
 \n http://www.let.rug.nl/~vannoord/Fsa/. The latest version of the\
 \n program should be available there too.\
 \n \
 \n * For information on the daVinci program, we refer to its homepage\
 \n http://www.informatik.uni-bremen.de/~inform/forschung/daVinci/.\
 \n \
 \n * For information on dot/GraphViz, we refer to\
 \n http://www.research.att.com:80/sw/tools/graphviz/. \
 \n \
 \n * For information on the VCG program: \
 \n http://www.cs.uni-sb.de/RW/users/sander/html/gsvcg1.html\
 \n \
 \n * A commercial version of VCG (called aiSee) is available from\
 \n http://www.aisee.com/gallery/\
 \n \
 \n * There is lots of interesting material at MLTT Xerox Grenoble: \
 \n http://www.xrce.xerox.com/competencies/content-analysis/fst/.\
 \n Be sure to read the documentation, including a number of nice examples.\
 \n \
 \n * AT&T's FSM toolset for weighted finite-state automata is available\
 \n from http://www.research.att.com/sw/tools/fsm. \
 \n \
 \n * The corresponding lextools package by Richard Sproat is now available\
 \n too, the url is http://www.research.att.com/sw/tools/lextools/\
 \n \
 \n * Grail.  Grail is a package of software for symbolic computation with \
 \n finite machines and regular expressions.  It is freely available \
 \n to students, educators, and anyone who simply wants to use the \
 \n software for their own amusement or education. The Grail homepage\
 \n is at http://www.csd.uwo.ca/staff/drraymon/.grail/grail.html\
 \n (somewhat outdated). A version for Linux is available from\
 \n http://http://www.cs.sun.ac.za/~lynette/merlin.html. \
 \n \
 \n * For a web interface to FSA, refer to\
 \n http://www.let.rug.nl/~vannoord/fsademo/.  \
 \n \
 \n * For a web interface to FSA3, c.f.: \
 \n http://i12www.ira.uka.de/Visualisierung.endlicher.Automaten/.\
 \n \
 \n * A tutorial for FSA by Gosse Bouma, in Dutch:\
 \n http://www.let.rug.nl/~gosse/tt/fsa.html\
 \n \
 \n * An even simpler tutorial for FSA (in Dutch as well) used for\
 \n highschool kids (!) is available as\
 \n http://www.let.rug.nl/~vannoord/fsademo/fsademo/klas.html\
 \n \
 \n * Electronic versions of some of the papers mentioned above are\
 \n available through the cmp-lg archive at\
 \n http://xxx.lanl.gov:80/cmp-lg/. \
 \n \
 \n * A list of related projects at University of Western Ontario by Darrell\
 \n Raymond is http://www.csd.uwo.ca/staff/drraymon/.grail/links.html. You can\
 \n also obtain a copy of Ted Leslie's thesis from that site, which\
 \n includes the algorithm to generate random automata, and which\
 \n discusses density of automata related to determinization.\
 \n \
 \n * Finite state Utilities by Jan Daciuk at\
 \n http://www.pg.gda.pl/~jandac/fsa.html. Useful tools for dictionary\
 \n construction and spell checking. Also read his dissertation.\
 \n \
 \n * SICStus Prolog home page: http://www.sics.se/isl/sicstus.html.\
 \n \
 \n * Collection of links on Prolog and Regular Expressions:\
 \n http://www.let.rug.nl/~vannoord/prolog-rx/PrologAndRegex.html \
 \n \
 \n * Wiese's Little Automata Builder at\
 \n http://www-ti.informatik.uni-tuebingen.de/~wiese/Automaton/\
 \n \
 \n * Another Java applet for finite automata at\
 \n http://www.cs.duke.edu/~rodger/tools/jflap/index.html\
 \n \
 \n * Interesting papers on Gene Myers' Home Page\
 \n http://www.cs.arizona.edu/people/gene\
 \n \
 \n * Application for interactive drawing of finite automata\
 \n http://triskam.virtualave.net/finomaton/finomaton.html\
 \n \
 \n 1.8 Copyright\
 \n \
 \n Copyright c 1995 - 2001 by Gertjan van Noord. This program is\
 \n distributed under Gnu General Public License (cf. the file COPYING in\
 \n distribution).\
 \n \
 \n This program is free software; you can redistribute it and/or modify\
 \n it under the terms of the GNU General Public License as published by\
 \n the Free Software Foundation; version 2 of the License.\
 \n \
 \n This program is distributed in the hope that it will be useful, but\
 \n WITHOUT ANY WARRANTY; without even the implied warranty of\
 \n MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU\
 \n General Public License for more details.\
 \n \
 \n You should have received a copy of the GNU General Public License\
 \n along with this program; if not, write to the Free Software\
 \n Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\
 \n \
 \n \
 \n 1.9 Acknowledgements\
 \n \
 \n Helpful suggestions, feedback, etc., from a variety of people, too\
 \n many to list here. Gosse Bouma and Dale Gerdemann were exceptionally\
 \n influential. \
 \n \
 \n \
 \n 1.10 Author\
 \n \
 \n Gertjan van Noord, \
 \n mailto:vannoord@let.rug.nl",Text). 

:- ensure_loaded(fsa_library).    % core functionality
:- use_module(fsa_options).       % command-line arguments
:- ensure_loaded(fsa_cmdint).     % command interpreter
:- use_module(fsa_help).

#sic# %% :- use_module(library(timeout), [ time_out/3 ]).
#sic# :- use_module(fsa_tk).          

#yap# :- yap_flag(unknown,warning).
#yap# :- leash(full).
%% #yap# :- write_depth(8,8).  I only want this when debugging...

#swi# :- set_prolog_flag(debug_on_error,false).

:- public fsa_main/0.

fsa_main:-
#!sic#     catch(fsa_main0,P,user:(print_message(error,P),halt(-1))).
#sic#      on_exception(P,fsa_main0,user:(print_message(error,P),halt(-1))).

fsa_main0 :-
    fsa_version,
#sic#     prolog_flag(argv,Args),
#yap#     unix(argv(Args)),
#swi#     current_prolog_flag(argv,Args0),
#swi#     (	append(_, [--|Args],Args0)
#swi#     ->	true
#swi#     ;   Args=[]
#swi#     ),
    (	fsa_options:parse(Cmd,Args,[])
    ->  fsa_run_command(Cmd,Args)
    ;	throw(fsa_error("cannot parse command line: ~w~n",[Args]))
    ),
    fsa_globals:fsa_global_get(interactive,IntOnOff),
    fsa_continue1(IntOnOff).

#swi# fsa_main_saved :-
#swi#     catch(fsa_main_saved0,P,user:print_message(error,P)).
#swi# 
#swi# fsa_main_saved0 :-
#swi#     fsa_version,
#swi#     bb_put(fsa:interactive,off),
#swi#     current_prolog_flag(argv,[_|Args]),
#swi#     (   fsa_options:parse(Cmd,Args,[])
#swi#     ->  fsa_run_command(Cmd,Args)
#swi#     ;   format(user_error,"error: cannot parse command line~n",[]),
#swi#         format(user_error,"fsa ~w~n",[Args])
#swi#     ),
#swi#     fsa_globals:fsa_global_get(interactive,IntOnOff),
#swi#     fsa_continue1(IntOnOff).

fsa_continue1(off) :-
#yap#    halt.
#swi#    halt.
#sic#    prolog_flag(system_type,DR),
#sic#    (   DR==runtime
#sic#    ->  fsa_globals:fsa_global_set(interactive,halt)
#sic# 	 ;   halt
#sic# 	 ).

fsa_continue1(on) :-
    true.

fsa_continue1(cmdint) :-
#yap# fsa_cmdint.
#swi# fsa_cmdint.
#sic#    prolog_flag(system_type,DR),
#sic#    (   DR==runtime
#sic# 	 ->  true
#sic# 	 ;   fsa_cmdint
#sic# 	 ).

fsa_run_command(true,_).
fsa_run_command(cmd(Cmd),Args) :-
#sic#       fsa_util:time(1,Cmd,Time,Space),
#yap#       fsa_util:time(1,Cmd,Time,Space),
#swi#       fsa_util:time(1,Cmd,Time0,Space),
#sic#       statistics(garbage_collection,[_,_,GcTime]),
#yap#       statistics(garbage_collection,[_,_,GcTime]),
#swi#       statistics(gctime,GcTime0), GcTime is integer(1000*GcTime0),
#swi#       (   GcTime < Time0 -> Time is Time0-GcTime ; Time0=Time ),
	fsa_util:debug_message(1,
                "total ~w: ~w msec; ~w kbytes (gc: ~w msec)~n",
		[Args,Time,Space,GcTime]).


#sic# %%% this is experimental. 
#sic# 
#sic# fsa_runtime_write_answer([]) :-
#sic#     format(user_error,"~nyes~n",[]).
#sic# fsa_runtime_write_answer([H|T]) :-
#sic#     format(user_error,"~n~w",[H]),
#sic#     fsa_runtime_write_answer0(T).
#sic# 
#sic# fsa_runtime_write_answer0([]) :-
#sic#     format(user_error," ? ",[]),
#sic#     fsa_runtime_more.
#sic# fsa_runtime_write_answer0([H|T]) :-
#sic#     format(user_error,",~n~w",[H]),
#sic#     fsa_runtime_write_answer0(T).
#sic#
#sic# :- public runtime_entry/1.
#sic# runtime_entry(start) :-
#sic#     fsa_globals:fsa_global_get(interactive,Value),
#sic#     fsa_runtime_start(Value).
#sic# 
#sic# fsa_runtime_start(on) :-
#sic#     fsa_runtime_cmd_interpreter.
#sic# 
#sic# fsa_runtime_start(halt).
#sic# 
#sic# fsa_runtime_start(cmdint) :-
#sic#     fsa_cmdint,
#sic#     fsa_runtime_cmd_interpreter.
#sic# 
#sic# 
#sic# fsa_runtime_cmd_interpreter :-
#sic#     format(user_error,
#sic#        "Welcome to FSA; (pseudo) Prolog prompt for runtime system~n",[]),
#sic#     format(user_error,
#sic#        "Type fsa_cmdint. to start the FSA command interpreter.~n",[]),
#sic#     format(user_error,"~n",[]),
#sic#     on_exception( Error, fsa_runtime_cmd_interpreter0,
#sic# 	   (   print_message(error,Error),
#sic# 	       fsa_runtime_cmd_interpreter
#sic# 	   )
#sic# 	 ).
#sic# 
#sic# fsa_runtime_cmd_interpreter0 :-
#sic#     %%	prompt(_,'o| ?- '),       % is written to user_error; but we want user
#sic#     format("o| ?- ",[]),
#sic#     ttyflush,
#sic#     read_term(Query,[variable_names(X)]),
#sic#     (	Query == halt
#sic#     ->	true
#sic#     ;   Query == end_of_file
#sic#     ->	true
#sic#     ;	(  call(Query),
#sic# 	    fsa_runtime_write_answer(X)
#sic# 	;   format(user_error,"~nno~n",[])
#sic# 	),
#sic#   !,   % according to spdet
#sic# 	fsa_runtime_cmd_interpreter0
#sic#     ).
#sic# 
#sic# fsa_runtime_more :-
#sic#     read_line(C),
#sic#     (	C = [59|_]
#sic#     ->	format(user_error,"~n",[]),fail
#sic#     ;	format(user_error,"~nyes~n",[])
#sic#     ).
#sic# 

help_info(class,xpred,"Exported Predicates",
"In addition to the graphical user interface and the FSA command interpreter,\
 \n it is also possible to use the toolkit as a library to your program. You can incorporate the FSA\
 \n program in your own program, just as you can use other Prolog\
 \n libraries. In order for this to work, you simply need to load the\
 \n file fsa_library.pl in the installation directory. For example:\
 \n \
 \n         % sicstus\
 \n         SICStus ...\
 \n         Licensed to ...\
 \n         | ?- use_module(fsa_library).\
 \n         ...\
 \n         ...\
 \n         ...\
 \n         yes\
 \n         | ?- fsa_regex_atom_compile('[a*,b^,{d,e}]',L).\
 \n         L = fa(r(fsa_preds),3,[0],[1],[trans(0,a,0),trans(0,b,2),\
 \n             trans(0,d,1),trans(0,e,1),trans(2,d,1),trans(2,e,1)],[]) ?   \
 \n         yes\
 \n         | ?- fsa_regex_transduces('{a:b,? -a}*',\"ababac\",L), atom_codes(Atom,L).\
 \n         L = [98,98,98,98,98,99],\
 \n         Atom = bbbbbc ?\
 \n         yes\
 \n         | ?-\
 \n \
 \n Most predicates that are imported have names starting with *fsa*. All\
 \n module names start with *fsa* as well.  Below, we list the predicates\
 \n exported by the FSA library.  The modules which efficiently implement\
 \n datastructures such as arrays, hashes, maps and sets are documented\
 \n separately; these modules are supposed to be useful independently from \
 \n their use in the FSA toolkit.").



#swi# :- fsa_main.
#yap# :- initialization(fsa_main).
