:- module( fsa_preds, []).

#swi# :- style_check([-atom]).

#sic# :- use_module(library(ordsets),
#yap# :- use_module(library(ordsets),
#swi# :- use_module(ordsets,
	                        [ ord_subtract/3,
 				  ord_intersection/3,
				  ord_union/3
				]).
#yap# :- use_module(library(lists),   [ member/2 ]).

:- public true/1, regex_notation_to_predicate/2, evaluate_predicate/2,
    conjunction/3, negation/2, display_predicate/2, talks_about/2,
    prepare_complement_of_set/2, complement_of_set/3, determinize_preds/2,
    t_determinize_preds/2, identity/2, class_to_pred/2,
    negated_class_to_pred/2, cleanup/2.

true(not_in([])).

regex_notation_to_predicate(Atom,Atom).

evaluate_predicate(Pred0,Symbol):-
    add_func(Pred0,Pred),
    evaluate_predicate0(Pred,Symbol).

evaluate_predicate0(in(List),Symbol) :-
    member(Symbol,List).

evaluate_predicate0(not_in(List),Symbol) :-
    non_member(Symbol,List).

add_func(in(El),T) :-
    !,
    in(El)=T.
add_func(not_in(El),T) :-
    !,
    not_in(El)=T.
add_func(El,in([El])).

del_func(in([H|T]),El):-
    del_func_in(T,H,El).
del_func(not_in(X),not_in(X)).

del_func_in([],H,H).
del_func_in([H|T],F,in([F,H|T])).

talks_about(Pred0,Sym) :-
    add_func(Pred0,Pred),
    talks_about0(Pred,Sym).

talks_about0(not_in(X),Sym) :-
    talks_about0(in(X),Sym).

talks_about0(in(X),Sym) :-
    member(Sym,X).

disjunction(A0,B0,C):-
    add_func(A0,A),
    add_func(B0,B),
    disjunction0(A,B,C0),
    del_func(C0,C).

disjunction0(not_in(A),B,C):-
    disjunction_not_in(B,A,C).

disjunction0(in(A),B,C):-
    disjunction_in(B,A,C).

disjunction_not_in(not_in(B),A,not_in(C)) :-
     ord_intersection(A,B,C).
disjunction_not_in(in(B),A,not_in(C)):-
    ord_subtract(A,B,C).

disjunction_in(in(B),A,in(C)):-
    ord_union(A,B,C).
disjunction_in(not_in(A),B,not_in(C)):-
    ord_subtract(A,B,C).

conjunction(A0,B0,C):-
    add_func(A0,A),
    add_func(B0,B),
    conjunction0(A,B,C0),
    del_func(C0,C).

conjunction0(not_in(A),B,C):-
    conjunction_not_in(B,A,C).

conjunction0(in(A),B,C):-
    conjunction_in(B,A,C).

conjunction_not_in(not_in(B),A,not_in(C)):-
    ord_union(A,B,C).
conjunction_not_in(in(B),A,in([Ch|Ct])):-
    ord_subtract(B,A,[Ch|Ct]).

conjunction_in(in(B),A,in([Ch|Ct])):-
    ord_intersection(A,B,[Ch|Ct]).
conjunction_in(not_in(A),B,in([Ch|Ct])):-
    ord_subtract(B,A,[Ch|Ct]).

negation(P0,P) :-
    add_func(P0,P1),
    negation0(P1,P2),
    del_func(P2,P).

negation0(in(P),not_in(P)).
negation0(not_in([H|T]),in([H|T])).

display_predicate(El,Conj):-
    add_func(El,In),
    display_predicate0(In,Conj).

display_predicate0(in(List),Conj):-
    fsa_visualization:abbreviate_label0(List,[H|T]),
    display_predicate_in(T,H,Conj).

display_predicate0(not_in(List),Not):-
    fsa_visualization:abbreviate_label0(List,Conj),
    display_predicate_not_in(Conj,Not).

display_predicate_in([],H,H).
display_predicate_in([H|T],A,{Conj}):-
    list_to_conj([H|T],A,Conj).

display_predicate_not_in([],?).
display_predicate_not_in([H|T], '`'(Conj)):-
    display_predicate_in(T,H,Conj).

list_to_conj([],A,A).
list_to_conj([H|T],A,(A,B)):-
    list_to_conj(T,H,B).

prepare_complement_of_set(_Fa,_).

complement_of_set(Preds,_,Comps):-
    true(True),
    (	conjoin_negated(Preds,True,Comp)
    ->	Comps=[Comp]
    ;   Comps=[]
    ).

conjoin_negated([],Pred,Pred).
conjoin_negated([P0|Preds],P1,Pred) :-
    negation(P0,NotP0),
    conjunction(P1,NotP0,P2),
    conjoin_negated(Preds,P2,Pred).

determinize_preds(U0,U):-
    fsa_determinizer:determinize_preds(U0,U,fsa_preds).

t_determinize_preds(U0,U):-
    fsa_t_determinizer:t_determinize_preds(U0,U,fsa_preds).

identity(in(L),P) :-
    !,
    P = '$@'(in(L)).
identity(not_in(L),P) :-
    !,
    P = '$@'(not_in(L)).
identity(P,P).

class_to_pred(List0,[Pred]):-
    sort(List0,List),
    del_func(in(List),Pred).

negated_class_to_pred(List0,[not_in(List)]):-
    sort(List0,List).

cleanup([],[]).
cleanup([H|T],[Pred]) :-
    cleanup(T,H,Pred).

cleanup([],P,P).
cleanup([H|T],Pred0,Pred) :-
    disjunction(H,Pred0,Pred1),
    cleanup(T,Pred1,Pred).

%% List is ground.
non_member(El,List) :-
    non_member_(List,El).

non_member_([],_).
non_member_([H|T],El) :-
#!swi#     dif(H,El),
#swi#      (   var(El)
#swi#      ->  throw(fsa_error("dif/2 not supported~n",[]))
#swi#      ;	  H \== El
#swi#      ),
    non_member_(T,El).
