:- module(fsa_u_hash, [%% updatable hashes using copying
		       fsa_u_hash_new/1,
                       fsa_u_hash_new/2,
		       fsa_u_hash_get/3,
		       fsa_u_hash_put/4,
                       fsa_u_hash_put/6
		    ]).

#swi# :- style_check([-atom]).
#!swi# :- use_module(library(terms), [ term_hash/4 ]).
#swi# :- use_module(terms, [ term_hash/4 ]).
#yap# :- use_module(library(lists), [ memberchk/2 ]).

:- discontiguous help_info/4.
:- public help_info/4.

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

user:help_info(module,fsa_u_hash,"Updatable Hashes",
"This module provides an updatable hash data-structure on top of\
 \n the updatable array datastructure.\
 \n \
 \n Here's an overview of the predicates provided:  \
 \n \
 \n * fsa_u_hash_new/[1,2]        create a new updatable hash\
 \n * fsa_u_hash_get/3            lookup a value from an updatable hash\
 \n * fsa_u_hash_put/[4,6]        update a value in an updatable hash\
 \n \
 \n The hash function is taken from library(terms). The default size of the\
 \n hashes is determined by the global variable *hash_size*.").

help_info(class,pred,"List of Predicates",
"This section lists the predicates defined by this module.").

help_info(pred,fsa_u_hash_new,"fsa_u_hash_new(-UpdatableFsaHash[,Size])",
"Initializes a new UpdatableFsaHash with size Size; or default size if\
 \n there is no second argument. The default size is determined by the\
 \n global variable *hash_size*."). 

fsa_u_hash_new(hash(_Thing,Size)):-
    fsa_globals:fsa_global_get(hash_size,Size).

fsa_u_hash_new(hash(_Thing,Size),Size).

help_info(pred,fsa_u_hash_get,"fsa_u_hash_get(+Key,?Val,+UpdatableFsaHash)",
"Val is unified with the value associated with Key in\
 \n UpdatableFsaHash. If no such key exists in UpdatableFsaHash already, then\
 \n the predicate fails."). 

fsa_u_hash_get(Term,Val,hash(Thing,Size)):-
    term_hash(Term,-1,Size,Index), % library(terms)
    fsa_u_array:fsa_u_array_get(Index,VarList,Thing),
    nonvar(VarList),
    memberchk(Term-Val,VarList).

help_info(pred,fsa_u_hash_put,
"fsa_u_hash_put(+Key,?Val,+UpdatableFsaHash0,?UpdatableFsaHash)\
 \n fsa_u_hash_put(+Key,?OldVal,?Default,?Val,+UpdatableFsaHash0,?UpdatableFsaHash)",
"The value associated with Key in UpdatableFsaHash0 is updated to Val,\
 \n resulting in the new hash UpdatableFsaHash. OldVal is unified with the \
 \n old value (it it existed) or with Default (if it didn't exist)."). 
fsa_u_hash_put(Term,Val,hash(Thing,Size),hash(NewThing,Size)):-
    term_hash(Term,-1,Size,Index), % library(terms)
    fsa_u_array:fsa_u_array_put(Index,List,Thing,NewThing,List0),
    update_key_list(List0,Term,Val,List).

fsa_u_hash_put(Term,OldVal,Default,Val,hash(Thing,Size),hash(NewThing,Size)):-
    term_hash(Term,-1,Size,Index), % library(terms)
    fsa_u_array:fsa_u_array_put(Index,List,Thing,NewThing,List0),
    update_key_list(List0,Term,OldVal,Default,Val,List).

update_key_list(List0,Key,Val,List):-
    (	var(List0)
    ->	List=[Key-Val]
    ;   List0==[]
    ->	List=[Key-Val]
    ;   List0=[Key1-Val1|Tail],
	(   Key==Key1
	->  List=[Key-Val|Tail]
	;   List=[Key1-Val1|List1],
	    update_key_list(Tail,Key,Val,List1)
	)
    ).

update_key_list(List0,Key,Old,Default,Val,List):-
    (	var(List0)
    ->	List=[Key-Val],
	Old=Default
    ;   List0==[]
    ->	List=[Key-Val],
	Old=Default
    ;   List0=[Key1-Val1|Tail],
	(   Key==Key1
	->  List=[Key-Val|Tail],
	    Old=Val1
	;   List=[Key1-Val1|List1],
	    update_key_list(Tail,Key,Old,Default,Val,List1)
	)
    ).

