:- module(ordsets, [ ord_member/2,
		     ord_union/2,
		     ord_union/3,
		     ord_subtract/3,
		     ord_intersect/2,
		     ord_intersection/3
		   ]).

ord_member(El,[H|T]):-
    compare(Op,El,H),
    ord_member(Op,El,T).

ord_member(=,_,_).
ord_member(>,El,[H|T]) :-
    compare(Op,El,H),
    ord_member(Op,El,T).

ord_union(A,B,C):-
    my_merge(A,B,C).

ord_union([], []).
ord_union([Set|Sets], Union) :-
    length([Set|Sets], NumberOfSets),
    merge_all(NumberOfSets, [Set|Sets], Union, []).

merge_all(N,Sets0,Union,Sets) :-
    (  N=:=1  -> Sets0=[Union|Sets]
    ;  N=:=2  -> Sets0=[Set1,Set2|Sets], 
                 my_merge(Set1,Set2,Union)
    ;  A is N>>1,
       Z is N-A,
       merge_all(A, Sets0, X, Sets1),
       merge_all(Z, Sets1, Y, Sets),
       my_merge(X, Y, Union)
    ).

my_merge([], Set, Set).
my_merge([O|Os], Ns, Set) :- 
    my_merge(Ns, O, Os, Set).

my_merge([], O, Os, [O|Os]).
my_merge([N|Ns], O, Os, Set) :-
    compare(C, O, N), 
    my_merge(C, O, Os, N, Ns, Set).

my_merge(<, O1, Os, N, Ns, [O1|Set]) :- 
    my_merge(Os, N, Ns, Set).
my_merge(=, _, Os, N, Ns, [N|Set]) :- 
    my_merge(Os, Ns, Set).
my_merge(>, O, Os, N1, Ns, [N1|Set]) :- 
    my_merge(Ns, O, Os, Set).

%% ord_intersect(+Set1, +Set2)
%% is true when the two ordered sets have at least one element in common.
ord_intersect([H1|T1],[H2|T2]) :-
    compare(Order,H1,H2),
    ord_intersect(Order,H1,T1,H2,T2).

ord_intersect(<,_,[H1|T1],H2,T2) :-
    compare(Order,H1,H2),
    ord_intersect(Order,H1,T1,H2,T2).    
ord_intersect(=,_,_,_,_).
ord_intersect(>,H1,T1,_,[H2|T2]) :-
    compare(Order,H1,H2),
    ord_intersect(Order,H1,T1,H2,T2).


ord_subtract([],_,[]).
ord_subtract([H1|T1],Set2,D) :-
    ord_subtract0(Set2,H1,T1,D).

ord_subtract0(<,H,Set1,H2,T2,[H|D]) :-
    ord_subtract1(Set1,H2,T2,D).
ord_subtract0(=,_,T1,_,T2,D) :-
    ord_subtract(T1,T2,D).
ord_subtract0(>,H1,T1,_,Set2,D) :-
    ord_subtract0(Set2,H1,T1,D).

ord_subtract1([],_,_,[]).
ord_subtract1([H1|T1],H2,T2,D):-
    compare(Order,H1,H2),
    ord_subtract0(Order,H1,T1,H2,T2,D).

ord_subtract0([],H1,T1,[H1|T1]).
ord_subtract0([H2|T2],H1,T1,D) :-
    compare(Order,H1,H2),
    ord_subtract0(Order,H1,T1,H2,T2,D).

ord_intersection([],_,[]).
ord_intersection([Head1|Tail1],Set2,Intersection) :-
    ord_intersection(Set2,Head1,Tail1,Intersection).

ord_intersection(<,_,Set1,Head2,Tail2,Intersection) :-
    ord_intersection(Set1,Head2,Tail2,Intersection).
ord_intersection(=,Head,Tail1,_,Tail2,[Head|Intersection]) :-
    ord_intersection(Tail1,Tail2,Intersection).
ord_intersection(>,Head1,Tail1,_,Set2,Intersection) :-
    ord_intersection(Set2,Head1,Tail1,Intersection).

ord_intersection([],_,_,[]).
ord_intersection([Head2|Tail2],Head1,Tail1,Intersection) :-
    compare(Order,Head1,Head2),
    ord_intersection(Order,Head1,Tail1,Head2,Tail2,Intersection).
